#! /bin/sh
################################################################################
#
# Package autolink.sh script.
# Copyright (c) 2002-2006 Oak Ridge National Laboratory.
#
# For more information see the following files in the source distribution top-
# level directory or package data directory (usually /usr/local/share/package):
#
# - README    for general package information.
# - INSTALL   for package install information.
# - COPYING   for package license information and copying conditions.
# - AUTHORS   for package authors information.
# - ChangeLog for package changes information.
#
# Run `autolink.sh` to automatically create, remove and modifiy links to
# default binaries/directories when using versioning in the form: "<name>.
# <number of first supported API version>.<number of supported API versions>.
# <revision number of implementation>".
#
################################################################################

# Set defaults for command line options.
SUFFIX=""
VERBOSE=0

# Parse command line options.
until test -z "$1"; do
  if test "$1" = "-a"; then
    if test -z "$2"; then
      echo "Error: Missing value for option \`$1\`."
      echo "Try \`$0 --help\` for more information."
      exit 1
    fi
    shift
    AGE=$1
  elif echo "$1" | grep -e ^--age= >/dev/null; then
    if test "$1" = "--age="; then
      echo "Error: Missing value for option \`$1\`."
      echo "Try \`$0 --help\` for more information."
      exit 1
    fi
    AGE=`echo $1 | sed s/--age=//`
  elif test "$1" = "-c"; then
    if test -z "$2"; then
      echo "Error: Missing value for option \`$1\`."
      echo "Try \`$0 --help\` for more information."
      exit 1
    fi
    shift
    CURRENT=$1
  elif echo "$1" | grep -e ^--current= >/dev/null; then
    if test "$1" = "--current="; then
      echo "Error: Missing value for option \`$1\`."
      echo "Try \`$0 --help\` for more information."
      exit 1
    fi
    CURRENT=`echo $1 | sed s/--current=//`
  elif test "$1" = "-f"; then
    if test -z "$2"; then
      echo "Error: Missing value for option \`$1\`."
      echo "Try \`$0 --help\` for more information."
      exit 1
    fi
    shift
    FILE=$1
  elif echo "$1" | grep -e ^--file= >/dev/null; then
    if test "$1" = "--file="; then
      echo "Error: Missing value for option \`$1\`."
      echo "Try \`$0 --help\` for more information."
      exit 1
    fi
    FILE=`echo $1 | sed s/--file=//`
  elif test "$1" = "--help" || test "$1" = "-h"; then
    echo "Usage:"
    echo " autolink.sh [options]"
    echo "Options:"
    echo " -a VALUE, --age=VALUE      Specify the version age."
    echo " -c VALUE, --current=VALUE  Specify the current version."
    echo " -f VALUE, --file=VALUE     Specify the target file."
    echo " -h,       --help           Print out command line help."
    echo " -s VALUE, --suffix=VALUE   Specify a target file suffix."
    echo " -v,       --version        Print out version information."
    echo " -V,       --verbose        Print out debug information."
    exit 0
  elif test "$1" = "-s"; then
    if test -z "$2"; then
      echo "Error: Missing value for option \`$1\`."
      echo "Try \`$0 --help\` for more information."
      exit 1
    fi
    shift
    SUFFIX=$1
  elif echo "$1" | grep -e ^--suffix= >/dev/null; then
    if test "$1" = "--suffix="; then
      echo "Error: Missing value for option \`$1\`."
      echo "Try \`$0 --help\` for more information."
      exit 1
    fi
    SUFFIX=`echo $1 | sed s/--suffix=//`
  elif test "$1" = "--version" || test "$1" = "-v"; then
    echo "autolink.sh 1.0"
    echo "Copyright (c) 2002-2006 Oak Ridge National Laboratory."
    echo "See COPYING in the source distribution top-level directory"
    echo "or package data directory (usually /usr/local/share/package)"
    echo "for license information and copying conditions."
    echo "Report bugs to engelmannc@ornl.gov."
    exit 0
  elif test "$1" = "--verbose" || test "$1" = "-V"; then
    VERBOSE=1
  else
    echo "Unrecognized option \`$1\`."
    echo "Try \`$0 --help\` for more information."
    exit 1
  fi
  shift
done

# Check command line options.
if test -z "$AGE"; then
  echo "Error: Missing version age."
  echo "Try \`$0 --help\` for more information."
  exit 1
fi
if test -z "$CURRENT"; then
  echo "Error: Missing current version."
  echo "Try \`$0 --help\` for more information."
  exit 1
fi
if test -z "$FILE"; then
  echo "Error: Missing target file."
  echo "Try \`$0 --help\` for more information."
  exit 1
fi

# Derive directory and base.
DIR=`dirname $FILE`
BASE=`basename $FILE`

# Remove any conflicting links.
if test "$VERBOSE" = "1"; then
  echo rm -f $FILE$SUFFIX
fi
rm -f $FILE$SUFFIX
FIRST=`expr $CURRENT - $AGE`
until test $FIRST -gt $CURRENT; do
  if test "$VERBOSE" = "1"; then
    echo rm -f $FILE.$FIRST$SUFFIX
  fi
  rm -f $FILE.$FIRST$SUFFIX
  FIRST=`expr $FIRST + 1`
done

# Create link to default.
LINK1=`find $DIR | sort -r |\
       grep -m 1 -e "/$BASE\\.[0-9]*\\.[0-9]*\\.[0-9]*$SUFFIX$"`
LINK1=`basename $LINK1 2>/dev/null`
if ! test -z "$LINK1"; then
  if test "$VERBOSE" = "1"; then
    echo ln -s $LINK1 $FILE$SUFFIX
  fi
  ln -s $LINK1 $FILE$SUFFIX
fi

# Create links to defaults depending on API version.
FIRST=`expr $CURRENT - $AGE`
LINK2=`find $DIR | sort -r |\
       grep -m 1 -e "/$BASE\\.$FIRST*\\.[0-9]*\\.[0-9]*$SUFFIX$"`
LINK2=`basename $LINK1 2>/dev/null`
if ! test -z "$LINK2"; then
  AGE=`echo $LINK2 | sed s/\\\\.$FIRST\\\\.[0-9]*\\\\.[0-9]*/----------\&/ |\
       sed s/.*----------\\\\.$FIRST\\\\.// | sed s/\\\\.[0-9]*$SUFFIX//`
  CURRENT=`expr $FIRST + $AGE`
  until test $FIRST -gt $CURRENT; do
    if test "$VERBOSE" = "1"; then
      echo ln -s $LINK2 $FILE.$FIRST$SUFFIX
    fi
    ln -s $LINK2 $FILE.$FIRST$SUFFIX
    FIRST=`expr $FIRST + 1`
  done
fi

################################################################################
#
# End of file.
#
################################################################################
