/******************************************************************************
 *
 * Source file for harness-test (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* Command line version output. */
#ifdef DEBUG
#define HARNESS_TEST_OPTIONS_VERSION HARNESS_STRING(\
harness-test.HARNESS_TEST_VERSION.debug (HARNESS daemon in debug mode)\
 HARNESS_TEST_PACKAGE_NAME-HARNESS_TEST_PACKAGE_VERSION.\
HARNESS_TEST_PACKAGE_RELEASE\n\
Copyright (c) 2002-2006 Oak Ridge National Laboratory.\n\
See COPYING in the source distribution top-level directory or package\n\
data directory (DATA_DIR) for license information.\n\
Report bugs to <HARNESS_TEST_PACKAGE_BUGREPORT>.\n)
#else /* DEBUG */
#define HARNESS_TEST_OPTIONS_VERSION HARNESS_STRING(\
harness-test.HARNESS_TEST_VERSION (HARNESS daemon)\
 HARNESS_TEST_PACKAGE_NAME-HARNESS_TEST_PACKAGE_VERSION.\
HARNESS_TEST_PACKAGE_RELEASE\n\
Copyright (c) 2002-2006 Oak Ridge National Laboratory.\n\
See COPYING in the source distribution top-level directory or package\n\
data directory (DATA_DIR) for license information.\n\
Report bugs to <HARNESS_TEST_PACKAGE_BUGREPORT>.\n)
#endif /* DEBUG */

/* Command line help output. */
#define HARNESS_TEST_OPTIONS_HELP "\
Usage:\n\
  harness-test[.debug] [option]\n\
General options:\n\
  -e, --echo     Write all following arguments to stdout.\n\
  -E, --env      Write all values of environment variables that\n\
                 correspond to following arguments to stdout.\n\
  -h, --help     Print out command line help.\n\
  -p, --pipe     Pipe all data from stdin to stdout.\n\
  -s, --signal   Write some received signals to stdout.\n\
  -v, --version  Print out version information.\n\
"

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include <harness-test.h>

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Harness-test mode type. */
typedef enum {
  HARNESS_TEST_MODE_NONE = 0, 
  HARNESS_TEST_MODE_ECHO,
  HARNESS_TEST_MODE_ENV,
  HARNESS_TEST_MODE_PIPE,
  HARNESS_TEST_MODE_SIGNAL
} harness_test_mode_t;

/* Harness-test module data type. */
typedef struct {
  harness_test_mode_t  mode;      /* test mode    */
  int                  wakeup[2]; /* wake-up pipe */
  unsigned int         size;      /* buffer size  */
  char                *buffer;    /* buffer       */
  sigset_t             sigmask;   /* signal mask  */
} harness_test_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Parses the command-line options.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
int harness_test_options_argv_parse (int          argc,
                                     const char **argv);

/*
 * Prints out a message to stdout and exits with an exit status.
 *
 * argv    = The argument vector (argument[0] is the program name).
 * format  = The message format.
 * message = The message.
 * status  = The exit status.
 */
void harness_test_options_exit (const char **argv,
                                const char  *format,
                                const char  *message,
                                int          status);

/*
 * Writes all arguments out to stdout with separating spaces.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_test_echo (int          argc,
                       const char **argv);

/*
 * Writes all environment variables that relate to the arguments out to stdout
 * with separating spaces.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_test_env (int          argc,
                      const char **argv);

/*
 * Pipes stdin to stdout.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_test_pipe ();

/*
 * The signal handler of the pipe test.
 *
 * signum = The number of the received signal.
 */
void harness_test_handler (int signum);

/*
 * Writes SIGPIPE, SIGTERM and SIGCHLD received signals to to stdout.
 *
 * return = Does not return. Waits in an endless loop instead.
 */
int harness_test_signal ();

/*
 * The signal handler of the signal test.
 *
 * signum = The number of the received signal.
 */
void harness_test_handler2 (int signum);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Harness-test module data. */
harness_test_data_t harness_test_data = {
  HARNESS_TEST_MODE_NONE,      /* test mode    */
  { -1, -1 },                  /* wake-up pipe */
  HARNESS_FORKER_BUFFER_SIZE, /* buffer size  */
  NULL                         /* buffer       */
  /* not initialized */        /* signal mask  */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Parses the command line options and runs the appropiate test mode.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = EXIT_SUCCESS on success or EXIT_FAILURE on any error.
 */
#undef  __FUNC__
#define __FUNC__ "main"
int main (int    argc,
          char **argv,
          char **envp) {
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    return EXIT_FAILURE;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    return EXIT_FAILURE;
  }
  /* Check envp parameter. */
  if (NULL == envp) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(envp parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    return EXIT_FAILURE;
  }
#endif /* DEBUG */
  /* Parse command-line options. */
  if (0 != harness_test_options_argv_parse(argc, (const char**)argv)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to parse command-line options)))
    return EXIT_FAILURE;
  }
  /* Check test mode. */
  switch (harness_test_data.mode) {
    case HARNESS_TEST_MODE_NONE: {
      break;
    }
    case HARNESS_TEST_MODE_ECHO: {
      /* Run echo test mode. */
      if (0 != harness_test_echo(argc - 2, (const char**)(argv + 2))) {
        HARNESS_LOG((HARNESS_WARN(unable to run echo test mode)))
        return EXIT_FAILURE;
      }
      break;
    }
    case HARNESS_TEST_MODE_ENV: {
      /* Run env test mode. */
      if (0 != harness_test_env(argc - 2, (const char**)(argv + 2))) {
        HARNESS_LOG((HARNESS_WARN(unable to run env test mode)))
        return EXIT_FAILURE;
      }
      break;
    }
    case HARNESS_TEST_MODE_PIPE: {
      /* Run pipe test mode. */
      if (0 != harness_test_pipe()) {
        HARNESS_LOG((HARNESS_WARN(unable to run pipe test mode)))
        return EXIT_FAILURE;
      }
      break;
    }
    case HARNESS_TEST_MODE_SIGNAL: {
      /* Run pipe test mode. */
      if (0 != harness_test_signal()) {
        HARNESS_LOG((HARNESS_WARN(unable to run signal test mode)))
        return EXIT_FAILURE;
      }
      break;
    }
    default: {
      HARNESS_LOG((HARNESS_WARN(unknown test mode)))
      return EXIT_FAILURE;
    }
  }
  return EXIT_SUCCESS;
}

/*
 * Parses the command-line options.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
#undef  __FUNC__
#define __FUNC__ "harness_test_options_argv_parse"
int harness_test_options_argv_parse (int          argc,
                                     const char **argv) {
  unsigned int index;
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  for (index = 1; index < argc; index++) {
    if ((0 == strcmp (argv[index],  "-e"   ))||
        (0 == strcmp (argv[index], "--echo"))) {
      /* Set echo test mode. */
      harness_test_data.mode = HARNESS_TEST_MODE_ECHO;
      break;
    } else if ((0 == strcmp (argv[index],  "-E"  ))||
               (0 == strcmp (argv[index], "--env"))) {
      /* Set env test mode. */
      harness_test_data.mode = HARNESS_TEST_MODE_ENV;
      break;
    } else if ((0 == strcmp (argv[index],  "-h"   ))||
               (0 == strcmp (argv[index], "--help"))) {
      /* Exit with help message. */
      harness_test_options_exit(argv, "%s",
                                HARNESS_TEST_OPTIONS_HELP, EXIT_SUCCESS);
    } else if ((0 == strcmp (argv[index],  "-p"   ))||
               (0 == strcmp (argv[index], "--pipe"))) {
      /* Set pipe test mode. */
      harness_test_data.mode = HARNESS_TEST_MODE_PIPE;
      break;
    } else if ((0 == strcmp (argv[index],  "-s"     ))||
               (0 == strcmp (argv[index], "--signal"))) {
      /* Set signal test mode. */
      harness_test_data.mode = HARNESS_TEST_MODE_SIGNAL;
      break;
    } else if ((0 == strcmp (argv[index],  "-v"      ))||
               (0 == strcmp (argv[index], "--version"))) {
      /* Exit with version message. */
      harness_test_options_exit(argv, "%s",
                                HARNESS_TEST_OPTIONS_VERSION, EXIT_SUCCESS);
    } else {
      /* Exit with error message. */
      harness_test_options_exit(argv, "ERROR: Unknown option '%s'.\n",
                                argv[index], EXIT_FAILURE);
    }
  }
  return 0;
}

/*
 * Prints out a message to stdout and exits with an exit status.
 *
 * argv    = The argument vector (argument[0] is the program name).
 * format  = The message format.
 * message = The message.
 * status  = The exit status.
 */
#undef  __FUNC__
#define __FUNC__ "harness_test_options_exit"
void harness_test_options_exit (const char **argv,
                                const char  *format,
                                const char  *message,
                                int          status) {
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.");
    exit(EXIT_FAILURE);
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.");
    exit(EXIT_FAILURE);
  }
  /* Check format parameter. */
  if (NULL == format) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(format parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.");
    exit(EXIT_FAILURE);
  }
  /* Check message parameter. */
  if (NULL == message) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(message parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.");
    exit(EXIT_FAILURE);
  }
#endif /* DEBUG */
  fprintf(stdout, format, message);
  if (EXIT_FAILURE == status) {
    fprintf(stdout, "Try `%s --help` for more information.\n", argv[0]);
  }
  exit(status);
}

/*
 * Writes all arguments out to stdout with separating spaces.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_test_echo"
int harness_test_echo (int          argc,
                       const char **argv) {
  int          bytes;
  unsigned int index1, index2, length;
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  for (index1 = 0; index1 < argc; index1++) {
    length = strlen(argv[index1]);
    /* Check for the need of writing space to stdout. */
    if ((0 < index1)&&(0 != length)) {
      /* Write space to stdout. */
      while (0 == (bytes = write(1, " ", 1)));
      /* Ckeck for error. */
      if (-1 == bytes) {
        HARNESS_LOG((HARNESS_WARN(unable to write to stdout)))
        return -1;
      }
    }
    /* Write argument to stdout. */
    for (index2  = 0, bytes = 1;
         index2  < length && 0 <= bytes;
         index2 += (0 < bytes)?bytes:0) {
      bytes = write(1, argv[index1] + index2, length - index2);
    }
    /* Ckeck for error. */
    if (-1 == bytes) {
      HARNESS_LOG((HARNESS_WARN(unable to write to stdout)))
      return -1;
    }
  }
  return 0;
}

/*
 * Writes all environment variables that relate to the arguments out to stdout
 * with separating spaces.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_test_env"
int harness_test_env (int          argc,
                      const char **argv) {
  int           bytes;
  char         *value;
  unsigned int  index1, index2, length;
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  for (index1 = 0; index1 < argc; index1++) {
    if (NULL != (value = getenv(argv[index1]))) {
      length = strlen(value);
      /* Check for the need of writing space to stdout. */
      if ((0 < index1)&&(0 != length)) {
        /* Write space to stdout. */
        while (0 == (bytes = write(1, " ", 1)));
        /* Ckeck for error. */
        if (-1 == bytes) {
          HARNESS_LOG((HARNESS_WARN(unable to write to stdout)))
          return -1;
        }
      }
      /* Write argument to stdout. */
      for (index2  = 0, bytes = 1;
           index2  < length && 0 <= bytes;
           index2 += (0 < bytes)?bytes:0) {
        bytes = write(1, value + index2, length - index2);
      }
      /* Ckeck for error. */
      if (-1 == bytes) {
        HARNESS_LOG((HARNESS_WARN(unable to write to stdout)))
        return -1;
      }
    }
  }
  return 0;
}

/*
 * Pipes stdin to stdout.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_test_pipe"
int harness_test_pipe () {
  int          fds, result, bytes1, bytes2;
  fd_set       fdset;
  sigset_t     sigmask;
  unsigned int index1, index2;
  result = 0;
  /* Initialize signal masks. */
  if ((-1 == sigemptyset(&sigmask                  ))||
      (-1 == sigemptyset(&harness_test_data.sigmask))) {
    HARNESS_LOG((HARNESS_WARN(unable to empty signal mask)))
    return -1;
  }
  if ((-1 == sigaddset(&sigmask                  , SIGTERM))||
      (-1 == sigaddset(&sigmask                  , SIGPIPE))||
      (-1 == sigaddset(&harness_test_data.sigmask, SIGPIPE))) {
    HARNESS_LOG((HARNESS_WARN(unable to add signal to signal mask)))
    return -1;
  }
  /* Block SIGCLD, SIGTERM and SIGPIPE signals. */
  if (-1 == sigprocmask(SIG_BLOCK, &sigmask, NULL)) {
    HARNESS_LOG((HARNESS_WARN(unable to set thread signal mask)))
    return -1;
  }
  /* Set stdin non-blocking. */
  if (0 != fcntl(0, F_SETFL, O_NONBLOCK)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to set stdin to non-blocking)))\
    return -1;
  }
  /* Initialize wake-up pipe. */
  if (0 != pipe(harness_test_data.wakeup)) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize wake-up pipe)))
    return -1;
  }
  /* Set wake-up pipe non-blocking. */
  if (0 != fcntl(harness_test_data.wakeup[0], F_SETFL, O_NONBLOCK)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to set wake-up pipe to non-blocking)))
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_test_data.wakeup[0]))||
        (0 != close(harness_test_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    }
    return -1;
  }
  /* Allocate buffer. */
  if (NULL == (harness_test_data.buffer = malloc(harness_test_data.size))) {
    HARNESS_LOG((HARNESS_WARN(unable to allocate buffer)))
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_test_data.wakeup[0]))||
        (0 != close(harness_test_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    }
    return -1;
  }
  /* Install signal handler for SIGCHLD and SIGTERM. */
  if ((SIG_ERR == signal(SIGTERM, harness_test_handler))||
      (SIG_ERR == signal(SIGCHLD, harness_test_handler))) {
    HARNESS_LOG((HARNESS_WARN(unable to install signal handler)))
    /* Deallocate buffer. */
    free(harness_test_data.buffer);
    harness_test_data.buffer = NULL;
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_test_data.wakeup[0]))||
        (0 != close(harness_test_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    }
    return -1;
  }
  while (1) {
    /* Empty file descriptor set. */
    FD_ZERO(&fdset);
    /* Add stdin and wake-up pipe to file descriptor set. */
    FD_SET(0,  &fdset);
    FD_SET(harness_test_data.wakeup[0], &fdset);
    /* Wait on file descriptor set. */
    fds = pselect(harness_test_data.wakeup[0] + 1, &fdset, NULL, NULL, NULL,
                  &harness_test_data.sigmask);
    /* Ckeck for error. */
    if ((-1 == fds)&&(EINTR != errno)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to wait on file descriptor set)))
      result = -1;
      break;
    }
    /* Ckeck ready wake-up pipe. */
    if ((0 < fds)&&(0 != FD_ISSET(harness_test_data.wakeup[0], &fdset))) {
      break;
    }
    /* Ckeck for ready stdin. */
    if ((0 < fds)&&(0 != FD_ISSET(0, &fdset))) {
      /* Read some data from stdin. */
      for (index1  = 0, bytes1 = 1;
           index1  < harness_test_data.size && 0 < bytes1;
           index1 += (0 < bytes1)?bytes1:0) {
        bytes1 = read(0, harness_test_data.buffer + index1,
                         harness_test_data.size   - index1);
      }
      /* Ckeck for error. */
      if ((-1 == bytes1)&&(EAGAIN != errno)) {
        HARNESS_LOG((HARNESS_WARN(unable to read from stdin)))
        result = -1;
        break;
      }
      /* Write data to stdout. */
      for (index2  = 0, bytes2 = 1;
           index2  < index1 && 0 <= bytes2;
           index2 += (0 < bytes2)?bytes2:0) {
        bytes2 = write(1, harness_test_data.buffer + index2, index1 - index2);
      }
      /* Ckeck for closed stdin. */
      if (0 == bytes1) {
        HARNESS_LOG((
          HARNESS_INFO(process shutdown by closed stdin)))
        break;
      }
      /* Ckeck for error. */
      if (-1 == bytes2) {
        HARNESS_LOG((HARNESS_WARN(unable to write to stdout)))
        result = -1;
        break;
      }
    }
  }
  /* Reset signal handler for SIGCHLD and SIGTERM. */
  if ((SIG_ERR == signal(SIGTERM, SIG_IGN))||
      (SIG_ERR == signal(SIGCHLD, SIG_IGN))) {
    HARNESS_LOG((HARNESS_WARN(unable to reset signal handler)))
    result = -1;
  }
  /* Deallocate buffer. */
  free(harness_test_data.buffer);
  harness_test_data.buffer = NULL;
  /* Finalize wake-up pipe. */
  if ((0 != close(harness_test_data.wakeup[0]))||
      (0 != close(harness_test_data.wakeup[1]))) {
    HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    result = -1;
  }
  return result;
}

/*
 * The signal handler of the forker process.
 *
 * signum = The number of the received signal.
 */
#undef  __FUNC__
#define __FUNC__ "harness_test_handler"
void harness_test_handler (int signum) {
  /* Check the number of the received signal. */
  switch (signum) {
    /* Check for SIGTERM signal. */
    case SIGTERM: {
      HARNESS_LOG((
        HARNESS_INFO(process shutdown by signal %s), "SIGTERM"))
      /* Send byte to wake-up pipe. */
      write(harness_test_data.wakeup[1], "", 1);
      break;
    }
    /* Check for any other signal. */
    default: {
      HARNESS_LOG((
        HARNESS_INFO(process shutdown by signal %d), signum))
      /* Send byte to wake-up pipe. */
      write(harness_test_data.wakeup[1], "", 1);
      break;
    }
  }
}

/*
 * Writes SIGPIPE, SIGTERM and SIGCHLD received signals to to stdout.
 *
 * return = Does not return. Waits in an endless loop instead.
 */
#undef  __FUNC__
#define __FUNC__ "harness_test_signal"
int harness_test_signal () {
  sigset_t sigmask;
  /* Initialize signal mask. */
  if (-1 == sigemptyset(&sigmask)) {
    HARNESS_LOG((HARNESS_WARN(unable to empty signal mask)))
    return -1;
  }
  if ((-1 == sigaddset(&sigmask, SIGUSR1))||
      (-1 == sigaddset(&sigmask, SIGUSR2))||
      (-1 == sigaddset(&sigmask, SIGTERM))) {
    HARNESS_LOG((HARNESS_WARN(unable to add signal to signal mask)))
    return -1;
  }
  /* Allow SIGUSR1, SIGUSR2 and SIGTERM signals. */
  if (-1 == sigprocmask(SIG_UNBLOCK, &sigmask, NULL)) {
    HARNESS_LOG((HARNESS_WARN(unable to set thread signal mask)))
    return -1;
  }
  /* Install signal handler for SIGUSR1, SIGUSR2 and SIGTERM. */
  if ((SIG_ERR == signal(SIGUSR1, harness_test_handler2))||
      (SIG_ERR == signal(SIGUSR2, harness_test_handler2))||
      (SIG_ERR == signal(SIGTERM, harness_test_handler2))) {
    HARNESS_LOG((HARNESS_WARN(unable to install signal handler)))
    return -1;
  }
  /* Wait in an endless loop. */
  while (1) pause();
}

/*
 * The signal handler of the signal test.
 *
 * signum = The number of the received signal.
 */
#undef  __FUNC__
#define __FUNC__ "harness_test_handler2"
void harness_test_handler2 (int signum) {
  /* Check for SIGTERM. */
  if (SIGTERM == signum) {
    /* Terminate. */
    exit(EXIT_SUCCESS);
  }
  /* Write signum to stdout. */
  write(1, &signum, sizeof(int));
  /* Reinstall signal handler for signum. */
  if (SIG_ERR == signal(signum, harness_test_handler2)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to reinstall signal handler)))
    /* Terminate. */
    exit(EXIT_FAILURE);
  }
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
