/******************************************************************************
 *
 * Source file for harness kernel (kernel process) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "harness.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Kernel process data type. */
typedef struct {
  pid_t             pid;       /* pid                                   */
  unsigned int      exit;      /* exit flag                             */
  unsigned int      forker;    /* forker flag                           */
  unsigned int      block;     /* block flag                            */
  int               pipes[2];  /* I/O pipe ends                         */
  int               wakeup[2]; /* wake-up pipe                          */
  pthread_mutex_t   mutex;     /* mutex                                 */
  unsigned int      size;      /* buffer size                           */
  char             *buffer;    /* buffer                                */
  char            **options;   /* init options to be applied at startup */
  sigset_t          sigmask;   /* signal mask                           */
  pthread_t         thread;    /* kernel thread                         */
  pthread_attr_t    attribs;   /* kernel thread attributes              */
} harness_kernel_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Initializes the plug-in loader.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_plugins_init ();

/*
 * Finalizes the plug-in loader.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_plugins_fini ();

/*
 * Initializes the process launcher.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_processes_init ();

/*
 * Closes the process launcher.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_processes_close ();

/*
 * Finalizes the process launcher.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_processes_fini ();

/*
 * Processes an execution notification.
 *
 * handle = The process handle.
 * pid    = The process system pid.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_processes_executed (unsigned int handle,
                                pid_t        pid);

/*
 * Processes a pipe closure notification.
 *
 * handle = The process handle.
 * fd     = The closed pipe.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_processes_closed (unsigned int handle,
                              int          fd);

/*
 * Processes a process output notification.
 *
 * handle = The process handle.
 * length = The process output length.
 * buffer = The process output buffer.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_processes_output (unsigned int  handle,
                              unsigned int  length,
                              const void   *buffer);

/*
 * Processes a process exit notification.
 *
 * handle = The process handle.
 * status = The process exit status.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_processes_exited (unsigned int handle,
                              int          status);

/*
 * Sets the thread pool configuration and initializes the thread pool.
 *
 * pool_minimum = The new minimum thread count.
 * pool_maximum = The new maximum thread count.
 * pool_timeout = The new exiting thread timeout.
 * jobs_maximum = The new maximum job count.
 * jobs_timeout = The new waiting thread timeout.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_threads_init (unsigned int pool_minimum,
                          unsigned int pool_maximum,
                          unsigned int pool_timeout,
                          unsigned int jobs_maximum,
                          unsigned int jobs_timeout);

/*
 * Initializes the thread pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_threads_init2 ();

/*
 * Finalizes the thread pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_threads_fini ();

/*
 * The kernel process thread (if needed).
 *
 * argument = NULL
 * return   = 0 on success or -1 on any system error with errno set
 *            appropriately.
 */
void* harness_kernel_thread (void *argument);

/*
 * Waits for the kernel thread to shut down if started in non-blocking mode.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_kernel_join ();

/*
 * Sets the kernel process configuration and runs the kernel process.
 *
 * pool_minimum = The new minimum thread pool thread count.
 * pool_maximum = The new maximum thread pool thread count.
 * pool_timeout = The new exiting thread pool thread timeout.
 * jobs_maximum = The new maximum thread pool job count.
 * jobs_timeout = The new waiting thread pool thread timeout.
 * init_options = The new init options to be applied at startup.
 *                A NULL terminated array of strings containing option/value
 *                pairs. Possible option/value pairs are:
 *                'load' and 'VALUE' to load a plug-in with the name of VALUE at
 *                    start-up. Allows plug-in to load more than once and does 
 *                    not export its symbols. VALUE cannot be ommitted.
 *                'load-unique' and 'VALUE' to load a plug-in with the name of
 *                    VALUE at start-up. Allows plug-in to load only once and
 *                    does not export its symbols. VALUE cannot be ommitted.
 *                'load-export' and 'VALUE' to load a plug-in with the name of
 *                    VALUE at start-up. Allows plug-in to load only once and
 *                    exports its symbols. VALUE cannot be ommitted.
 *                'exec' and 'VALUE' to execute a program with the name of VALUE
 *                    in a child process at start-up. VALUE cannot be ommitted.
 *                All options are applied in the order they appear.
 * pipes        = The pipe ends for kernel-forker communication.
 * block        = The block flag to indicate wether to run the kernel in this
 *                thread and block until finished, or to create a new kernel
 *                thread and return after kernel startup.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_kernel_main (unsigned int   pool_minimum,
                         unsigned int   pool_maximum,
                         unsigned int   pool_timeout,
                         unsigned int   jobs_maximum,
                         unsigned int   jobs_timeout,
                         char         **init_options,
                         int            pipes[2],
                         unsigned int   block);

/*
 * Runs the kernel process.
 *
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_kernel_main2 (int pipes[2]);

/*
 * Sets the kernel process configuration and initializes the kernel process.
 *
 * pool_minimum = The new minimum thread pool thread count.
 * pool_maximum = The new maximum thread pool thread count.
 * pool_timeout = The new exiting thread pool thread timeout.
 * jobs_maximum = The new maximum thread pool job count.
 * jobs_timeout = The new waiting thread pool thread timeout.
 * init_options = The new init options to be applied at startup.
 *                A NULL terminated array of strings containing option/value
 *                pairs. Possible option/value pairs are:
 *                'load' and 'VALUE' to load a plug-in with the name of VALUE at
 *                    start-up. Allows plug-in to load more than once and does 
 *                    not export its symbols. VALUE cannot be ommitted.
 *                'load-unique' and 'VALUE' to load a plug-in with the name of
 *                    VALUE at start-up. Allows plug-in to load only once and
 *                    does not export its symbols. VALUE cannot be ommitted.
 *                'load-export' and 'VALUE' to load a plug-in with the name of
 *                    VALUE at start-up. Allows plug-in to load only once and
 *                    exports its symbols. VALUE cannot be ommitted.
 *                'exec' and 'VALUE' to execute a program with the name of VALUE
 *                    in a child process at start-up. VALUE cannot be ommitted.
 *                All options are applied in the order they appear.
 * pipes        = The pipe ends for kernel-forker communication.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_kernel_init (unsigned int   pool_minimum,
                         unsigned int   pool_maximum,
                         unsigned int   pool_timeout,
                         unsigned int   jobs_maximum,
                         unsigned int   jobs_timeout,
                         char         **init_options,
                         int            pipes[2]);

/*
 * Initializes the kernel process.
 *
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_kernel_init2 (int pipes[2]);

/*
 * Applies init options, such as loading plug-ins and executing programs.
 *
 * argument = NULL
 * return   = 0 on success or -1 on any system error with errno set
 *            appropriately.
 */
HARNESS_THREADS_JOB(harness_kernel_options_apply);

/*
 * Finalizes the kernel process.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_kernel_fini ();

/*
 * The signal handler of the kernel process.
 *
 * signum = The number of the received signal.
 */
void harness_kernel_handler (int signum);

/*
 * Runs the kernel process tasks.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_kernel_run ();

/*
 * Processes forker messages.
 *
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_kernel_process ();

/*
 * Reads an execution notification.
 *
 * handle = The process handle (return).
 * pid    = The pid (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_kernel_read_executed (unsigned int *handle,
                                  pid_t        *pid);

/*
 * Reads a pipe closure notification.
 *
 * handle = The process handle (return).
 * fd     = The closed pipe (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_kernel_read_closed (unsigned int *handle,
                                int          *fd);

/*
 * Reads a process output notification.
 *
 * handle = The process handle (return).
 * status = The process exit status (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_kernel_read_output (unsigned int *handle,
                                unsigned int *length);

/*
 * Reads a process exit notification.
 *
 * handle = The process handle (return).
 * status = The process exit status (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_kernel_read_exited (unsigned int *handle,
                                int          *status);

/*
 * Writes a process execution request.
 *
 * handle = The process handle.
 * argc   = The argument vector count.
 * envc   = The environment vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envv   = The environment vector.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_kernel_execute (unsigned int   handle,
                            unsigned int   argc,
                            unsigned int   envc,
                            const char   **argv,
                            const char   **envv);

/*
 * Writes one or more process input requests.
 *
 * handle = The process handle.
 * length = The process input length.
 * buffer = The process input.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_kernel_forward (unsigned int  handle,
                            unsigned int  length,
                            const void   *buffer);

/*
 * Reads a buffer from a file descriptor (non-/blocking).
 *
 * fd     = The file descriptor to read from.
 * length = The maximum buffer length to read.
 * buffer = The buffer to read into.
 * block  = The (non-)blocking flag (0 = non-blocking, 1 = blocking).
 * return = number of bytes read on success, -2 on closed file descriptor or -1
 *          on any other error with errno set appropriately.
 */
int harness_kernel_read (int           fd,
                         unsigned int  length,
                         void         *buffer,
                         unsigned int  block);

/*
 * Writes a buffer into a file descriptor (blocking).
 *
 * fd     = The file descriptor to write to.
 * length = The buffer length to write.
 * buffer = The buffer to write from.
 * return = 0 on success, -2 on closed file descriptor or -1 on any other error
 *          with errno set appropriately.
 */
int harness_kernel_write (int           fd,
                          unsigned int  length,
                          const void   *buffer);

/*
 * The kernel write mutex unlock cleanup.
 *
 * argument = NULL.
 */
void harness_kernel_unlock (void *argument);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Kernel process data. */
harness_kernel_data_t harness_kernel_data = {
  0,                          /* pid                                   */
  0,                          /* exit flag                             */
  0,                          /* forker flag                           */
  1,                          /* block flag                            */
  { -1, -1 },                 /* I/O pipe ends                         */
  { -1, -1 },                 /* wake-up pipe                          */
  PTHREAD_MUTEX_INITIALIZER,  /* mutex                                 */
  HARNESS_FORKER_BUFFER_SIZE, /* buffer size                           */
  NULL,                       /* buffer                                */
  NULL                        /* init options to be applied at startup */
  /* not initialized */       /* signal mask                           */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Sets the block flag.
 * This function may only be called before initializing.
 *
 * block   = The block flag to indicate wether to run the kernel in this
 *           thread and block until finished, or to create a new kernel
 *           thread and return after kernel startup.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_block_set"
int harness_kernel_block_set (unsigned int block) {
  harness_kernel_data.block = block;
  return 0;
}

/*
 * Sets the init options to be applied at startup.
 * This function may only be called before initializing.
 *
 * options = The init options to be applied at startup.
 *           A NULL terminated array of strings containing option/value
 *           pairs. Possible option/value pairs are:
 *           'load' and 'VALUE' to load a plug-in with the name of VALUE at
 *               start-up. Allows plug-in to load more than once and does 
 *               not export its symbols. VALUE cannot be ommitted.
 *           'load-unique' and 'VALUE' to load a plug-in with the name of
 *               VALUE at start-up. Allows plug-in to load only once and
 *               does not export its symbols. VALUE cannot be ommitted.
 *           'load-export' and 'VALUE' to load a plug-in with the name of
 *               VALUE at start-up. Allows plug-in to load only once and
 *               exports its symbols. VALUE cannot be ommitted.
 *           'exec' and 'VALUE' to execute a program with the name of VALUE
 *               in a child process at start-up. VALUE cannot be ommitted.
 *           All options are applied in the order they appear.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_options_set"
int harness_kernel_options_set (char **options) {
  harness_kernel_data.options = options;
  return 0;
}

/*
 * Gets the block flag.
 *
 * block   = The block flag to indicate wether to run the kernel in this
 *           thread and block until finished, or to create a new kernel
 *           thread and return after kernel startup.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_block_get"
int harness_kernel_block_get (unsigned int *block) {
  *block = harness_kernel_data.block;
  return 0;
}

/*
 * Gets the forker process buffer size.
 *
 * options = The init options to be applied at startup (return).
 *           A NULL terminated array of strings containing option/value
 *           pairs. Possible option/value pairs are:
 *           'load' and 'VALUE' to load a plug-in with the name of VALUE at
 *               start-up. Allows plug-in to load more than once and does 
 *               not export its symbols. VALUE cannot be ommitted.
 *           'load-unique' and 'VALUE' to load a plug-in with the name of
 *               VALUE at start-up. Allows plug-in to load only once and
 *               does not export its symbols. VALUE cannot be ommitted.
 *           'load-export' and 'VALUE' to load a plug-in with the name of
 *               VALUE at start-up. Allows plug-in to load only once and
 *               exports its symbols. VALUE cannot be ommitted.
 *           'exec' and 'VALUE' to execute a program with the name of VALUE
 *               in a child process at start-up. VALUE cannot be ommitted.
 *           All options are applied in the order they appear.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_options_get"
int harness_kernel_options_get (char ***options) {
  *options = harness_kernel_data.options;
  return 0;
}

/*
 * Sets the kernel process configuration and runs the kernel process.
 *
 * pool_minimum = The new minimum thread pool thread count.
 * pool_maximum = The new maximum thread pool thread count.
 * pool_timeout = The new exiting thread pool thread timeout.
 * jobs_maximum = The new maximum thread pool job count.
 * jobs_timeout = The new waiting thread pool thread timeout.
 * init_options = The new init options to be applied at startup.
 *                A NULL terminated array of strings containing option/value
 *                pairs. Possible option/value pairs are:
 *                'load' and 'VALUE' to load a plug-in with the name of VALUE at
 *                    start-up. Allows plug-in to load more than once and does 
 *                    not export its symbols. VALUE cannot be ommitted.
 *                'load-unique' and 'VALUE' to load a plug-in with the name of
 *                    VALUE at start-up. Allows plug-in to load only once and
 *                    does not export its symbols. VALUE cannot be ommitted.
 *                'load-export' and 'VALUE' to load a plug-in with the name of
 *                    VALUE at start-up. Allows plug-in to load only once and
 *                    exports its symbols. VALUE cannot be ommitted.
 *                'exec' and 'VALUE' to execute a program with the name of VALUE
 *                    in a child process at start-up. VALUE cannot be ommitted.
 *                All options are applied in the order they appear.
 * pipes        = The pipe ends for kernel-forker communication.
 * block        = The block flag to indicate wether to run the kernel in this
 *                thread and block until finished, or to create a new kernel
 *                thread and return after kernel startup.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_main"
int harness_kernel_main (unsigned int   pool_minimum,
                         unsigned int   pool_maximum,
                         unsigned int   pool_timeout,
                         unsigned int   jobs_maximum,
                         unsigned int   jobs_timeout,
                         char         **init_options,
                         int            pipes[2],
                         unsigned int   block) {
  int error;
  /* Initialize kernel process. */
  if (0 != harness_kernel_init(pool_minimum, pool_maximum, pool_timeout,
                               jobs_maximum, jobs_timeout, init_options,
                               pipes)) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize kernel process)))
    return -1;
  }
  /* Check for blocking kernel startup. */
  harness_kernel_data.block = block;
  if (1 == block) {
    /* Run kernel process. */
    if (0 != harness_kernel_run()) {
      HARNESS_LOG((HARNESS_WARN(unable to run kernel process)))
      if (0 != harness_kernel_fini()) {
        HARNESS_LOG((HARNESS_WARN(unable to finalize kernel process)))
      }
      return -1;
    }
    /* Finalize kernel process. */
    if (0 != harness_kernel_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize kernel process)))
      return -1;
    }
  /* Perform non-blocking kernel startup. */
  } else {
    /* Initialize thread attributes. */
    if (0 != (error = pthread_attr_init(&harness_kernel_data.attribs))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to initialize thread attributes)))
      return -1;
    }
    /* Set thread attributes detach state. */
    if (0 != (error =
              pthread_attr_setdetachstate(&harness_kernel_data.attribs,
                                          PTHREAD_CREATE_DETACHED))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to set thread attributes detach state)))
      /* Finalize thread attributes. */
      if (0 != (error = pthread_attr_destroy(&harness_kernel_data.attribs))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to finalize thread attributes)))
      }
      return -1;
    }
    if (0 != (error = pthread_create(&harness_kernel_data.thread,
                                     &harness_kernel_data.attribs,
                                      harness_kernel_thread, NULL))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to create thread)))
      /* Finalize thread attributes. */
      if (0 != (error = pthread_attr_destroy(&harness_kernel_data.attribs))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to finalize thread attributes)))
      }
      return -1;
    }
  }
  return 0;
}

/*
 * Runs the kernel process.
 *
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_main2"
int harness_kernel_main2 (int pipes[2]) {
  int      error;
  /* Initialize kernel process. */
  if (0 != harness_kernel_init2(pipes)) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize kernel process)))
    return -1;
  }
  /* Check for blocking kernel startup. */
  if (1 == harness_kernel_data.block) {
    /* Run kernel process. */
    if (0 != harness_kernel_run()) {
      HARNESS_LOG((HARNESS_WARN(unable to run kernel process)))
      if (0 != harness_kernel_fini()) {
        HARNESS_LOG((HARNESS_WARN(unable to finalize kernel process)))
      }
      return -1;
    }
    /* Finalize kernel process. */
    if (0 != harness_kernel_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize kernel process)))
      return -1;
    }
  /* Perform non-blocking kernel startup. */
  } else {
    /* Initialize thread attributes. */
    if (0 != (error = pthread_attr_init(&harness_kernel_data.attribs))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to initialize thread attributes)))
      return -1;
    }
    /* Set thread attributes joinable state. */
    if (0 != (error =
              pthread_attr_setdetachstate(&harness_kernel_data.attribs,
                                          PTHREAD_CREATE_JOINABLE))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to set thread attributes detach state)))
      /* Finalize thread attributes. */
      if (0 != (error = pthread_attr_destroy(&harness_kernel_data.attribs))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to finalize thread attributes)))
      }
      return -1;
    }
    if (0 != (error = pthread_create(&harness_kernel_data.thread,
                                     &harness_kernel_data.attribs,
                                      harness_kernel_thread, NULL))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to create thread)))
      /* Finalize thread attributes. */
      if (0 != (error = pthread_attr_destroy(&harness_kernel_data.attribs))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to finalize thread attributes)))
      }
      return -1;
    }
  }
  return 0;
}

/*
 * The kernel process thread (if needed).
 *
 * argument = NULL
 * return   = 0 on success or -1 on any system error with errno set
 *            appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_thread"
void* harness_kernel_thread (void *argument) {
  /* Run kernel process. */
  if (0 != harness_kernel_run()) {
    fprintf(stderr, "ERROR: Unable to run kernel process.\n");
    if (0 != harness_kernel_fini()) {
      fprintf(stderr, "ERROR: Unable to finalize kernel process.\n");
    }
    return NULL;
  }
  /* Finalize kernel process. */
  if (0 != harness_kernel_fini()) {
    fprintf(stderr, "ERROR: Unable to finalize kernel process.\n");
    return NULL;
  }
  return NULL;
}

/*
 * Waits for the kernel thread to shut down if started in non-blocking mode.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_join"
int harness_kernel_join () {
  int error;
  /* Check blocking flag. */
  if (1 == harness_kernel_data.block) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(kernel was started in blocking mode)))
    return -1;
  }
  /* Join the kernel thread. */
  if (0 != (error = pthread_join(harness_kernel_data.thread, NULL))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to join thread)))
    return -1;
  }
  /* Finalize thread attributes. */
  if (0 != (error = pthread_attr_destroy(&harness_kernel_data.attribs))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to finalize thread attributes)))
    return -1;
  }
  return 0;
}

/*
 * Sets the kernel process configuration and initializes the kernel process.
 *
 * pool_minimum = The new minimum thread pool thread count.
 * pool_maximum = The new maximum thread pool thread count.
 * pool_timeout = The new exiting thread pool thread timeout.
 * jobs_maximum = The new maximum thread pool job count.
 * jobs_timeout = The new waiting thread pool thread timeout.
 * init_options = The new init options to be applied at startup.
 *                A NULL terminated array of strings containing option/value
 *                pairs. Possible option/value pairs are:
 *                'load' and 'VALUE' to load a plug-in with the name of VALUE at
 *                    start-up. Allows plug-in to load more than once and does 
 *                    not export its symbols. VALUE cannot be ommitted.
 *                'load-unique' and 'VALUE' to load a plug-in with the name of
 *                    VALUE at start-up. Allows plug-in to load only once and
 *                    does not export its symbols. VALUE cannot be ommitted.
 *                'load-export' and 'VALUE' to load a plug-in with the name of
 *                    VALUE at start-up. Allows plug-in to load only once and
 *                    exports its symbols. VALUE cannot be ommitted.
 *                'exec' and 'VALUE' to execute a program with the name of VALUE
 *                    in a child process at start-up. VALUE cannot be ommitted.
 *                All options are applied in the order they appear.
 * pipes        = The pipe ends for kernel-forker communication.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_init"
int harness_kernel_init (unsigned int   pool_minimum,
                         unsigned int   pool_maximum,
                         unsigned int   pool_timeout,
                         unsigned int   jobs_maximum,
                         unsigned int   jobs_timeout,
                         char         **init_options,
                         int            pipes[2]) {
  sigset_t sigmask;
  HARNESS_LOG((HARNESS_INFO(kernel process startup)))
  /* Initialize init options. */
  harness_kernel_data.options = init_options;
  /* Initialize I/O pipe ends. */
  harness_kernel_data.pipes[0] = pipes[0];
  harness_kernel_data.pipes[1] = pipes[1];
  /* Save pid. */
  harness_kernel_data.pid = getpid();
  /* Initialize exit flag. */
  harness_kernel_data.exit = 0;
  /* Get forker flag. */
  if (0 != harness_forker_get(&harness_kernel_data.forker)) {
    HARNESS_LOG((HARNESS_WARN(unable to get forker flag)))
    harness_syserr();
    return -1;
  }
  /* Initialize signal masks. */
  if ((-1 == sigemptyset(&sigmask                     ))||
      (-1 == sigemptyset(&harness_kernel_data.sigmask))) {
    HARNESS_LOG((HARNESS_WARN(unable to empty signal mask)))
    harness_syserr();
    return -1;
  }
  if ((-1 == sigaddset(&sigmask                    , SIGCLD ))||
      (-1 == sigaddset(&sigmask                    , SIGTERM))||
      (-1 == sigaddset(&sigmask                    , SIGPIPE))||
      (-1 == sigaddset(&harness_kernel_data.sigmask, SIGPIPE))) {
    HARNESS_LOG((HARNESS_WARN(unable to add signal to signal mask)))
    harness_syserr();
    return -1;
  }
  /* Block SIGCLD, SIGTERM and SIGPIPE signals. */
  if (-1 == pthread_sigmask(SIG_BLOCK, &sigmask, NULL)) {
    HARNESS_LOG((HARNESS_WARN(unable to set thread signal mask)))
    harness_syserr();
    return -1;
  }
  /* Initialize wake-up pipe. */
  if (0 != pipe(harness_kernel_data.wakeup)) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize wake-up pipe)))
    harness_syserr();
    return -1;
  }
  /* Set wake-up pipe non-blocking. */
  if (0 != fcntl(harness_kernel_data.wakeup[0], F_SETFL, O_NONBLOCK)) {
    HARNESS_LOG((HARNESS_WARN(unable to set wake-up pipe to non-blocking)))
    harness_syserr();
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    }
    return -1;
  }
  /* Get forker process buffer size. */
  if (0 != harness_forker_size_get(&harness_kernel_data.size)) {
    HARNESS_LOG((HARNESS_WARN(unable to get forker process buffer size)))
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    }
    return -1;
  }
  /* Allocate buffer. */
  if (NULL == (harness_kernel_data.buffer =
               malloc(harness_kernel_data.size))) {
    HARNESS_LOG((HARNESS_WARN(unable to allocate buffer)))
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    }
    return -1;
  }
  /* Initialize process launcher. */
  if ((1 == harness_kernel_data.forker)&&
      (0 != harness_processes_init()  )) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize process launcher)))
    return -1;
  }
  /* Initialize plug-in loader. */
  if (0 != harness_plugins_init()) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize plug-in loader)))
    /* Finalize process launcher. */
    if ((1 == harness_kernel_data.forker)&&
        (0 != harness_processes_fini()  )) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize process launcher)))
    }
    /* Deallocate buffer. */
    free(harness_kernel_data.buffer);
    harness_kernel_data.buffer = NULL;
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
      harness_syserr();
    }
    return -1;
  }
  /* Initialize thread pool. */
  if (0 != harness_threads_init(pool_minimum, pool_maximum, pool_timeout,
                                jobs_maximum, jobs_timeout)) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize thread pool)))
    /* Finalize plug-in loader. */
    if (0 != harness_plugins_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize plug-in loader)))
    }
    /* Finalize process launcher. */
    if ((1 == harness_kernel_data.forker)&&
        (0 != harness_processes_fini()  )) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize process launcher)))
    }
    /* Deallocate buffer. */
    free(harness_kernel_data.buffer);
    harness_kernel_data.buffer = NULL;
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
      harness_syserr();
    }
    return -1;
  }
  /* Install signal handler for SIGCHLD and SIGTERM. */
  if ((SIG_ERR == signal(SIGTERM, harness_kernel_handler))||
      (SIG_ERR == signal(SIGCHLD, harness_kernel_handler))) {
    HARNESS_LOG((HARNESS_WARN(unable to install signal handler)))
    harness_syserr();
    /* Finalize thread pool. */
    if (0 != harness_threads_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize thread pool)))
    }
    /* Finalize plug-in loader. */
    if (0 != harness_plugins_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize plug-in loader)))
    }
    /* Finalize process launcher. */
    if ((1 == harness_kernel_data.forker)&&
        (0 != harness_processes_fini()  )) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize process launcher)))
    }
    /* Deallocate buffer. */
    free(harness_kernel_data.buffer);
    harness_kernel_data.buffer = NULL;
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
      harness_syserr();
    }
    return -1;
  }
  /* Submit job to apply init options. */
  if (0 != harness_threads_submit(harness_kernel_options_apply,
                                   NULL, NULL)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to submit job to apply init options)))
    /* Reset signal handler for SIGCHLD and SIGTERM. */
    if ((SIG_ERR == signal(SIGTERM, SIG_IGN))||
        (SIG_ERR == signal(SIGCHLD, SIG_IGN))) {
      HARNESS_LOG((HARNESS_WARN(unable to reset signal handler)))
      harness_syserr();
    }
    /* Finalize thread pool. */
    if (0 != harness_threads_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize thread pool)))
    }
    /* Finalize plug-in loader. */
    if (0 != harness_plugins_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize plug-in loader)))
    }
    /* Finalize process launcher. */
    if ((1 == harness_kernel_data.forker)&&
        (0 != harness_processes_fini())) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize process launcher)))
    }
    /* Deallocate buffer. */
    free(harness_kernel_data.buffer);
    harness_kernel_data.buffer = NULL;
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
      harness_syserr();
    }
    return -1;
  }
  return 0;
}

/*
 * Initializes the kernel process.
 *
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_init2"
int harness_kernel_init2 (int pipes[2]) {
  sigset_t sigmask;
  HARNESS_LOG((HARNESS_INFO(kernel process startup)))
  /* Initialize I/O pipe ends. */
  harness_kernel_data.pipes[0] = pipes[0];
  harness_kernel_data.pipes[1] = pipes[1];
  /* Save pid. */
  harness_kernel_data.pid = getpid();
  /* Initialize exit flag. */
  harness_kernel_data.exit = 0;
  /* Get forker flag. */
  if (0 != harness_forker_get(&harness_kernel_data.forker)) {
    HARNESS_LOG((HARNESS_WARN(unable to get forker flag)))
    harness_syserr();
    return -1;
  }
  /* Initialize signal masks. */
  if ((-1 == sigemptyset(&sigmask                     ))||
      (-1 == sigemptyset(&harness_kernel_data.sigmask))) {
    HARNESS_LOG((HARNESS_WARN(unable to empty signal mask)))
    harness_syserr();
    return -1;
  }
  if ((-1 == sigaddset(&sigmask                    , SIGCLD ))||
      (-1 == sigaddset(&sigmask                    , SIGTERM))||
      (-1 == sigaddset(&sigmask                    , SIGPIPE))||
      (-1 == sigaddset(&harness_kernel_data.sigmask, SIGPIPE))) {
    HARNESS_LOG((HARNESS_WARN(unable to add signal to signal mask)))
    harness_syserr();
    return -1;
  }
  /* Block SIGCLD, SIGTERM and SIGPIPE signals. */
  if (-1 == pthread_sigmask(SIG_BLOCK, &sigmask, NULL)) {
    HARNESS_LOG((HARNESS_WARN(unable to set thread signal mask)))
    harness_syserr();
    return -1;
  }
  /* Initialize wake-up pipe. */
  if (0 != pipe(harness_kernel_data.wakeup)) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize wake-up pipe)))
    harness_syserr();
    return -1;
  }
  /* Set wake-up pipe non-blocking. */
  if (0 != fcntl(harness_kernel_data.wakeup[0], F_SETFL, O_NONBLOCK)) {
    HARNESS_LOG((HARNESS_WARN(unable to set wake-up pipe to non-blocking)))
    harness_syserr();
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    }
    return -1;
  }
  /* Get forker process buffer size. */
  if (0 != harness_forker_size_get(&harness_kernel_data.size)) {
    HARNESS_LOG((HARNESS_WARN(unable to get forker process buffer size)))
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    }
    return -1;
  }
  /* Allocate buffer. */
  if (NULL == (harness_kernel_data.buffer =
               malloc(harness_kernel_data.size))) {
    HARNESS_LOG((HARNESS_WARN(unable to allocate buffer)))
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    }
    return -1;
  }
  /* Initialize process launcher. */
  if ((1 == harness_kernel_data.forker)&&
      (0 != harness_processes_init()  )) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize process launcher)))
    return -1;
  }
  /* Initialize plug-in loader. */
  if (0 != harness_plugins_init()) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize plug-in loader)))
    /* Finalize process launcher. */
    if ((1 == harness_kernel_data.forker)&&
        (0 != harness_processes_fini()  )) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize process launcher)))
    }
    /* Deallocate buffer. */
    free(harness_kernel_data.buffer);
    harness_kernel_data.buffer = NULL;
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
      harness_syserr();
    }
    return -1;
  }
  /* Initialize thread pool. */
  if (0 != harness_threads_init2()) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize thread pool)))
    /* Finalize plug-in loader. */
    if (0 != harness_plugins_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize plug-in loader)))
    }
    /* Finalize process launcher. */
    if ((1 == harness_kernel_data.forker)&&
        (0 != harness_processes_fini()  )) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize process launcher)))
    }
    /* Deallocate buffer. */
    free(harness_kernel_data.buffer);
    harness_kernel_data.buffer = NULL;
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
      harness_syserr();
    }
    return -1;
  }
  /* Install signal handler for SIGCHLD and SIGTERM. */
  if ((SIG_ERR == signal(SIGTERM, harness_kernel_handler))||
      (SIG_ERR == signal(SIGCHLD, harness_kernel_handler))) {
    HARNESS_LOG((HARNESS_WARN(unable to install signal handler)))
    harness_syserr();
    /* Finalize thread pool. */
    if (0 != harness_threads_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize thread pool)))
    }
    /* Finalize plug-in loader. */
    if (0 != harness_plugins_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize plug-in loader)))
    }
    /* Finalize process launcher. */
    if ((1 == harness_kernel_data.forker)&&
        (0 != harness_processes_fini()  )) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize process launcher)))
    }
    /* Deallocate buffer. */
    free(harness_kernel_data.buffer);
    harness_kernel_data.buffer = NULL;
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
      harness_syserr();
    }
    return -1;
  }
  /* Submit job to apply init options. */
  if (0 != harness_threads_submit(harness_kernel_options_apply,
                                   NULL, NULL)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to submit job to apply init options)))
    /* Reset signal handler for SIGCHLD and SIGTERM. */
    if ((SIG_ERR == signal(SIGTERM, SIG_IGN))||
        (SIG_ERR == signal(SIGCHLD, SIG_IGN))) {
      HARNESS_LOG((HARNESS_WARN(unable to reset signal handler)))
      harness_syserr();
    }
    /* Finalize thread pool. */
    if (0 != harness_threads_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize thread pool)))
    }
    /* Finalize plug-in loader. */
    if (0 != harness_plugins_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize plug-in loader)))
    }
    /* Finalize process launcher. */
    if ((1 == harness_kernel_data.forker)&&
        (0 != harness_processes_fini())) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize process launcher)))
    }
    /* Deallocate buffer. */
    free(harness_kernel_data.buffer);
    harness_kernel_data.buffer = NULL;
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_kernel_data.wakeup[0]))||
        (0 != close(harness_kernel_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
      harness_syserr();
    }
    return -1;
  }
  return 0;
}

/*
 * Applies init options, such as loading plug-ins and executing programs.
 *
 * argument = NULL
 * return   = 0 on success or -1 on any system error with errno set
 *            appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_options_apply"
HARNESS_THREADS_JOB(harness_kernel_options_apply) {
  unsigned int index, handle;
  /* Test for no options. */
  if (NULL == harness_kernel_data.options) {
    return 0;
  }
  for (index = 0; NULL != harness_kernel_data.options[index]; index += 2) {
    /* Test for no option value. */
    if (NULL == harness_kernel_data.options[index + 1]) {
      fprintf(stderr, "ERROR: Missing kernel init option '%s' value.\n",
              harness_kernel_data.options[index]);
      harness_kernel_shutdown();
      return 0;
    }
    if (0 == strcmp(harness_kernel_data.options[index], "load")) {
      if (0 != harness_plugins_load(&handle,
                 harness_kernel_data.options[index + 1], 0)) {
        fprintf(stderr, "ERROR: Unable to load '%s'.\n",
                harness_kernel_data.options[index + 1]);
        harness_kernel_shutdown();
        return 0;
      }
    } else if (0 == strcmp(harness_kernel_data.options[index],
                           "load-export")) {
      if (0 != harness_plugins_load(&handle,
                 harness_kernel_data.options[index + 1], 
                 HARNESS_PLUGINS_EXPORT)) {
        fprintf(stderr, "ERROR: Unable to load '%s'.\n",
                harness_kernel_data.options[index + 1]);
        harness_kernel_shutdown();
        return 0;
      }
    } else if (0 == strcmp(harness_kernel_data.options[index], "load-unique")) {
      if (0 != harness_plugins_load(&handle,
                  harness_kernel_data.options[index + 1],
                  HARNESS_PLUGINS_UNIQUE)) {
        fprintf(stderr, "ERROR: Unable to load '%s'.\n",
                harness_kernel_data.options[index + 1]);
        harness_kernel_shutdown();
        return 0;
      }
    } else if (0 == strcmp(harness_kernel_data.options[index],
                           "load-export-unique")) {
      if (0 != harness_plugins_load(&handle,
                 harness_kernel_data.options[index + 1], 
                 HARNESS_PLUGINS_EXPORT | HARNESS_PLUGINS_UNIQUE)) {
        fprintf(stderr, "ERROR: Unable to load '%s'.\n",
                harness_kernel_data.options[index + 1]);
        harness_kernel_shutdown();
        return 0;
      }
    } else if (0 == strcmp(harness_kernel_data.options[index], "exec")) {
      if (0 != harness_processes_execute(&handle,
                 harness_kernel_data.options[index + 1], NULL)) {
        fprintf(stderr, "ERROR: Unable to execute '%s'.\n",
                harness_kernel_data.options[index + 1]);
        harness_kernel_shutdown();
        return 0;
      }
    } else {
      fprintf(stderr, "ERROR: Unknown kernel init option '%s'.\n",
              harness_kernel_data.options[index + 1]);
      return -1;
    }
  }
  return 0;
}

/*
 * Finalizes the kernel process.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_fini"
int harness_kernel_fini () {
  int result;
  result = 0;
  HARNESS_LOG((HARNESS_INFO(kernel process shutdown)))
  /* Reset signal handler for SIGCHLD and SIGTERM. */
  if ((SIG_ERR == signal(SIGTERM, SIG_IGN))||
      (SIG_ERR == signal(SIGCHLD, SIG_IGN))) {
    HARNESS_LOG((HARNESS_WARN(unable to reset signal handler)))
    harness_syserr();
    result = -1;
  }
  /* Close process launcher. */
  if ((1 == harness_kernel_data.forker)&&
      (0 != harness_processes_close()  )) {
    HARNESS_LOG((HARNESS_WARN(unable to close process launcher)))
    result = -1;
  }
  /* Finalize thread pool. */
  if (0 != harness_threads_fini()) {
    HARNESS_LOG((HARNESS_WARN(unable to finalize thread pool)))
    result = -1;
  }
  /* Finalize plug-in loader. */
  if (0 != harness_plugins_fini()) {
    HARNESS_LOG((HARNESS_WARN(unable to finalize plug-in loader)))
    result = -1;
  }
  /* Finalize process launcher. */
  if ((1 == harness_kernel_data.forker)&&
      (0 != harness_processes_fini()  )) {
    HARNESS_LOG((HARNESS_WARN(unable to finalize process launcher)))
    result = -1;
  }
  /* Deallocate buffer. */
  free(harness_kernel_data.buffer);
  harness_kernel_data.buffer = NULL;
  /* Finalize wake-up pipe. */
  if ((0 != close(harness_kernel_data.wakeup[0]))||
      (0 != close(harness_kernel_data.wakeup[1]))) {
    HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    harness_syserr();
    result = -1;
  }
  harness_kernel_data.wakeup[0] = -1;
  harness_kernel_data.wakeup[1] = -1;
  /* Reset exit flag. */
  harness_kernel_data.exit = 0;
  /* Reset pid. */
  harness_kernel_data.pid = 0;
  /* Close I/O pipe ends. */
  if ((-1 != close(harness_kernel_data.pipes[0]))&&(0 == result)) {
    close(harness_kernel_data.pipes[0]);
  }
  if ((-1 != close(harness_kernel_data.pipes[1]))&&(0 == result)) {
    close(harness_kernel_data.pipes[1]);
  }
  /* Reset I/O pipe ends. */
  harness_kernel_data.pipes[0] = -1;
  harness_kernel_data.pipes[1] = -1;
  return result;
}

/*
 * The signal handler of the kernel process.
 *
 * signum = The number of the received signal.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_handler"
void harness_kernel_handler (int signum) {
  /* Check the number of the received signal. */
  switch (signum) {
    /* Check for SIGCHLD signal. */
    case SIGCHLD: {
      HARNESS_LOG((
        HARNESS_INFO(kernel process shutdown by signal %s), "SIGCHLD"))
      /* Ckeck exit flag. */
      if (1 != harness_kernel_data.exit) {
        /* Set exit flag. */
        harness_kernel_data.exit = 1;
        /* Send byte to wake-up pipe. */
        write(harness_kernel_data.wakeup[1], "", 1);
      }
      /* Reinstall signal handler for SIGCHLD. */
      if (SIG_ERR == signal(SIGCHLD, harness_kernel_handler)) {
        HARNESS_LOG((HARNESS_WARN(unable to reinstall signal handler)))
        harness_syserr();
      }
      break;
    }
    /* Check for SIGTERM signal. */
    case SIGTERM: {
      HARNESS_LOG((
        HARNESS_INFO(kernel process shutdown by signal %s), "SIGTERM"))
      /* Set exit flag. */
      harness_kernel_data.exit = 1;
      /* Send byte to wake-up pipe. */
      write(harness_kernel_data.wakeup[1], "", 1);
      /* Reinstall signal handler for SIGTERM. */
      if (SIG_ERR == signal(SIGTERM, harness_kernel_handler)) {
        HARNESS_LOG((HARNESS_WARN(unable to reinstall signal handler)))
        harness_syserr();
      }
      break;
    }
    /* Check for any other signal. */
    default: {
      HARNESS_LOG((
        HARNESS_INFO(kernel process shutdown by signal %d), signum))
      harness_syserr();
      /* Set exit flag. */
      harness_kernel_data.exit = 1;
      /* Send byte to wake-up pipe. */
      write(harness_kernel_data.wakeup[1], "", 1);
      break;
    }
  }
}

/*
 * Asynchonously shuts down the kernel and subsequently the forker.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_shutdown"
void harness_kernel_shutdown () {
  /* Check for thread cancellation. */
  pthread_testcancel();
  /* Send SIGTERM to kernel process. */
  if (0 != kill(harness_kernel_data.pid, SIGTERM)) {
    HARNESS_LOG((HARNESS_WARN(unable to send SIGTERM to kernel process)))
    exit(EXIT_FAILURE);
  }
}

/*
 * Runs the kernel process tasks.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_run"
int harness_kernel_run () {
  int    maxfd, fds;
  fd_set fdset;
  HARNESS_LOG((HARNESS_INFO(kernel process running)))
  /* Calculate the greatest file descriptor in the set. */
  maxfd = harness_kernel_data.wakeup[0];
  if ((1     == harness_kernel_data.forker  )&&
      (maxfd <  harness_kernel_data.pipes[0])) {
    maxfd = harness_kernel_data.pipes[0];
  }
  while (1) {
    /* Empty file descriptor set. */
    FD_ZERO(&fdset);
    /* Add incoming I/O pipe and wake-up pipe to file descriptor set. */
    FD_SET(harness_kernel_data.wakeup[0], &fdset);
    if (1 == harness_kernel_data.forker) {
      FD_SET(harness_kernel_data.pipes[0],  &fdset);
    }
    /* Wait on file descriptor set. */
    fds = pselect(maxfd + 1, &fdset, NULL, NULL, NULL,
                  &harness_kernel_data.sigmask);
    /* Ckeck for error. */
    if ((-1 == fds)&&(EINTR != errno)) {
      HARNESS_LOG((HARNESS_WARN(unable to wait on file descriptor set)))
      harness_syserr();
      return -1;
    }
    /* Ckeck exit flag. */
    if (1 == harness_kernel_data.exit) {
      return 0;
    }
    /* Ckeck ready wake-up pipe. */
    if ((0 < fds)&&(0 != FD_ISSET(harness_kernel_data.wakeup[0], &fdset))) {
      /* Reduce the number of ready file descriptors. */
      fds--;
      /* Try to empty wake-up pipe. */
      if (0 > harness_kernel_read(harness_kernel_data.wakeup[0],
                                   harness_kernel_data.size,
                                   harness_kernel_data.buffer, 0)) {
        HARNESS_LOG((HARNESS_WARN(unable to empty wake-up pipe)))
        return -1;
      }
    }
    /* Ckeck for ready incoming I/O pipe. */
    if ((0  < fds                                            )&&
        (1 == harness_kernel_data.forker                    )&&
        (0 != FD_ISSET(harness_kernel_data.pipes[0], &fdset))) {
      /* Reduce the number of ready file descriptors. */
      fds--;
      /* Process kernel message. */
      if (0 != harness_kernel_process()) {
        HARNESS_LOG((HARNESS_WARN(unable to process kernel message)))
        return -1;
      }
    }
  }
  return 0;
}

/*
 * Processes forker messages.
 *
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_process"
int harness_kernel_process () {
  char command;
  /* Read command from incoming I/O pipe. */
  if (0 > harness_kernel_read(harness_kernel_data.pipes[0],
                               sizeof(char), &command, 1)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to read length from incoming I/O pipe)))
    return -1;
  }
  /* Check command char. */
  switch (command) {
    /* Check for execution notification. */
    case 'e': {
      pid_t        pid;
      unsigned int handle;
      /* Read execution notification. */
      if (0 != harness_kernel_read_executed(&handle, &pid)) {
        HARNESS_LOG((HARNESS_WARN(unable to read execution notification)))
        return -1;
      }
      /* Process execution notification. */
      if (0 != harness_processes_executed(handle, pid)) {
        HARNESS_LOG((
          HARNESS_WARN(unable to process execution notification)))
        return -1;
      }
      return 0;
    }
    /* Check for pipe closure notification. */
    case 'p': {
      int          fd;
      unsigned int handle;
      /* Read pipe closure notification. */
      if (0 != harness_kernel_read_closed(&handle, &fd)) {
        HARNESS_LOG((
          HARNESS_WARN(unable to read pipe closure notification)))
        return -1;
      }
      /* Process pipe closure notification. */
      if (0 != harness_processes_closed(handle, fd)) {
        HARNESS_LOG((
          HARNESS_WARN(unable to process pipe closure notification)))
        return -1;
      }
      return 0;
    }
    /* Check for process output notification. */
    case 'o': {
      unsigned int  handle, length;
      /* Read process output notification. */
      if (0 != harness_kernel_read_output(&handle, &length)) {
        HARNESS_LOG((
          HARNESS_WARN(unable to read process output notification)))
        return -1;
      }
      /* Process process output notification. */
      if (0 != harness_processes_output(handle, length,
                                         harness_kernel_data.buffer)) {
        HARNESS_LOG((
          HARNESS_WARN(unable to process process output notification)))
        return -1;
      }
      return 0;
    }
    /* Check for process exit notification. */
    case 'q': {
      int          status;
      unsigned int handle;
      /* Read process exit notification. */
      if (0 != harness_kernel_read_exited(&handle, &status)) {
        HARNESS_LOG((
          HARNESS_WARN(unable to read process exit notification)))
        return -1;
      }
      /* Process process exit notification. */
      if (0 != harness_processes_exited(handle, status)) {
        HARNESS_LOG((
          HARNESS_WARN(unable to process process exit notification)))
        return -1;
      }
      return 0;
    }
    /* Check for unknown command */
    default: {
      HARNESS_LOG((HARNESS_WARN(unknown command from incoming I/O pipe)))
      return -1;
    }
  }
}

/*
 * Reads an execution notification.
 *
 * handle = The process handle (return).
 * pid    = The pid (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_read_executed"
int harness_kernel_read_executed (unsigned int *handle,
                                  pid_t        *pid) {
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    return -1;
  }
  /* Check pid parameter. */
  if (NULL == pid) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(pid parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Read handle from incoming I/O pipe. */
  if (0 > harness_kernel_read(harness_kernel_data.pipes[0],
                              sizeof(unsigned int), handle, 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read handle from incoming I/O pipe)))
    return -1;
  }
  /* Read pid from incoming I/O pipe. */
  if (0 > harness_kernel_read(harness_kernel_data.pipes[0],
                              sizeof(pid_t), pid, 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read pid from incoming I/O pipe)))
    return -1;
  }
  return 0;
}

/*
 * Reads a pipe closure notification.
 *
 * handle = The process handle (return).
 * fd     = The closed pipe (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_read_closed"
int harness_kernel_read_closed (unsigned int *handle,
                                int          *fd) {
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    return -1;
  }
  /* Check fd parameter. */
  if (NULL == fd) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(fd parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Read handle from incoming I/O pipe. */
  if (0 > harness_kernel_read(harness_kernel_data.pipes[0],
                              sizeof(unsigned int), handle, 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read handle from incoming I/O pipe)))
    return -1;
  }
  /* Read fd from incoming I/O pipe. */
  if (0 > harness_kernel_read(harness_kernel_data.pipes[0],
                              sizeof(int), fd, 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read fd from incoming I/O pipe)))
    return -1;
  }
  return 0;
}

/*
 * Reads a process output notification.
 *
 * handle = The process handle (return).
 * status = The process exit status (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_read_output"
int harness_kernel_read_output (unsigned int *handle,
                                 unsigned int *length) {
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    return -1;
  }
  /* Check length parameter. */
  if (NULL == length) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(length parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Read handle from incoming I/O pipe. */
  if (0 > harness_kernel_read(harness_kernel_data.pipes[0],
                              sizeof(unsigned int), handle, 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read handle from incoming I/O pipe)))
    return -1;
  }
  /* Read length from incoming I/O pipe. */
  if (0 > harness_kernel_read(harness_kernel_data.pipes[0],
                              sizeof(unsigned int), length, 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read length from incoming I/O pipe)))
    return -1;
  }
  /* Check length. */
  if (*length > harness_kernel_data.size) {
    HARNESS_LOG((HARNESS_WARN(unable to read more than buffer size)))
    harness_syserr();
    return -1;
  }
  /* Read buffer from incoming I/O pipe. */
  if (0 > harness_kernel_read(harness_kernel_data.pipes[0],
                              *length, harness_kernel_data.buffer, 1)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to read buffer from incoming I/O pipe)))
    return -1;
  }
  return 0;
}

/*
 * Reads a process exit notification.
 *
 * handle = The process handle (return).
 * status = The process exit status (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_read_exited"
int harness_kernel_read_exited (unsigned int *handle,
                                int          *status) {
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    return -1;
  }
  /* Check status parameter. */
  if (NULL == status) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(status parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Read handle from incoming I/O pipe. */
  if (0 > harness_kernel_read(harness_kernel_data.pipes[0],
                               sizeof(unsigned int), handle, 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read handle from incoming I/O pipe)))
    return -1;
  }
  /* Read status from incoming I/O pipe. */
  if (0 > harness_kernel_read(harness_kernel_data.pipes[0],
                               sizeof(int), status, 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read status from incoming I/O pipe)))
    return -1;
  }
  return 0;
}

/*
 * Writes a process execution request.
 *
 * handle = The process handle.
 * argc   = The argument vector count.
 * envc   = The environment vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envv   = The environment vector.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_execute"
int harness_kernel_execute (unsigned int   handle,
                            unsigned int   argc,
                            unsigned int   envc,
                            const char   **argv,
                            const char   **envv) {
  int          error;
  unsigned int index, length;
#ifdef DEBUG
  /* Check length and buffer parameters. */
  if ((0 != argc)&&(NULL == argv)) {
    errno = EINVAL;
    HARNESS_LOG((
      HARNESS_WARN(buffer parameter is NULL and length parameter is not 0)))
    return -1;
  }
  /* Check length and buffer parameters. */
  if ((0 != envc)&&(NULL == envv)) {
    errno = EINVAL;
    HARNESS_LOG((
      HARNESS_WARN(buffer parameter is NULL and length parameter is not 0)))
    return -1;
  }
#endif /* DEBUG */
  /* Lock kernel write mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_kernel_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock kernel write mutex)))
    harness_syserr();
    return -1;
  }
  /* Push unlock kernel write mutex to pthread cleanup stack. */
  pthread_cleanup_push(harness_kernel_unlock, NULL);
  /* Write command to outgoing I/O pipe. */
  if (0 != harness_kernel_write(harness_kernel_data.pipes[1], 1, "E")) {
    HARNESS_LOG((
      HARNESS_WARN(unable to write command to outgoing I/O pipe)))
    harness_syserr();
    /* Unlock kernel write mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Write handle to outgoing I/O pipe. */
  if (0 != harness_kernel_write(harness_kernel_data.pipes[1],
                                sizeof(unsigned int), &handle)) {
    HARNESS_LOG((HARNESS_WARN(unable to write handle to outgoing I/O pipe)))
    harness_syserr();
    /* Unlock kernel write mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Write argc to outgoing I/O pipe. */
  if (0 != harness_kernel_write(harness_kernel_data.pipes[1],
                                sizeof(unsigned int), &argc)) {
    HARNESS_LOG((HARNESS_WARN(unable to write argc to outgoing I/O pipe)))
    harness_syserr();
    /* Unlock kernel write mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Write envc to outgoing I/O pipe. */
  if (0 != harness_kernel_write(harness_kernel_data.pipes[1],
                                sizeof(unsigned int), &envc)) {
    HARNESS_LOG((HARNESS_WARN(unable to write envc to outgoing I/O pipe)))
    harness_syserr();
    /* Unlock kernel write mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Write length of complete argv[0] to outgoing I/O pipe. */
  length = (strlen(HARNESS_STRING(HARNESS_EXECDIR/)) + strlen(argv[0]) + 1) *
           sizeof(char);
  if (0 != harness_kernel_write(harness_kernel_data.pipes[1],
                                sizeof(unsigned int), &length)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to write length to outgoing I/O pipe)))
    harness_syserr();
    /* Unlock kernel write mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Write argv[0] prefix to outgoing I/O pipe. */
  if (0 != harness_kernel_write(harness_kernel_data.pipes[1],
                                strlen(HARNESS_STRING(HARNESS_EXECDIR/)) *
                                sizeof(char),
                                HARNESS_STRING(HARNESS_EXECDIR/))) {
    HARNESS_LOG((
      HARNESS_WARN(unable to write argv string to outgoing I/O pipe)))
    harness_syserr();
    /* Unlock kernel write mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Write argv[0] string to outgoing I/O pipe. */
  if (0 != harness_kernel_write(harness_kernel_data.pipes[1],
                                (strlen(argv[0]) + 1) * sizeof(char),
                                argv[0])) {
    HARNESS_LOG((
      HARNESS_WARN(unable to write argv string to outgoing I/O pipe)))
    harness_syserr();
    /* Unlock kernel write mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Write argv to outgoing I/O pipe. */
  for (index = 1; index < argc; index++) {
    /* Write length to outgoing I/O pipe. */
    length = (strlen(argv[index]) + 1) * sizeof(char);
    if (0 != harness_kernel_write(harness_kernel_data.pipes[1],
                                  sizeof(unsigned int), &length)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write length to outgoing I/O pipe)))
      harness_syserr();
      /* Unlock kernel write mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Write argv string to outgoing I/O pipe. */
    if (0 != harness_kernel_write(harness_kernel_data.pipes[1],
                                  length, argv[index])) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write argv string to outgoing I/O pipe)))
      harness_syserr();
      /* Unlock kernel write mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
        harness_syserr();
      }
      return -1;
    }
  }
  /* Write envv to outgoing I/O pipe. */
  for (index = 0; index < envc; index++) {
    /* Write length to outgoing I/O pipe. */
    length = (strlen(envv[index]) + 1) * sizeof(char);
    if (0 != harness_kernel_write(harness_kernel_data.pipes[1],
                                  sizeof(unsigned int), &length)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write length to outgoing I/O pipe)))
      harness_syserr();
      /* Unlock kernel write mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Write envv string to outgoing I/O pipe. */
    if (0 != harness_kernel_write(harness_kernel_data.pipes[1],
                                  length, envv[index])) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write envv string to outgoing I/O pipe)))
      harness_syserr();
      /* Unlock kernel write mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
        harness_syserr();
      }
      return -1;
    }
  }
  /* Pop unlock kernel write mutex from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Unlock kernel write mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
    harness_syserr();
    return -1;
  }
  return 0;
}

/*
 * Writes one or more process input requests.
 *
 * handle = The process handle.
 * length = The process input length.
 * buffer = The process input.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_forward"
int harness_kernel_forward (unsigned int  handle,
                            unsigned int  length,
                            const void   *buffer) {
  int          error;
  unsigned int index, bytes;
#ifdef DEBUG
  /* Check length and buffer parameters. */
  if ((0 != length)&&(NULL == buffer)) {
    errno = EINVAL;
    HARNESS_LOG((
      HARNESS_WARN(buffer parameter is NULL and length parameter is not 0)))
    return -1;
  }
#endif /* DEBUG */
  for (index = 0; index < length; index += harness_kernel_data.size) {
    /* Lock kernel write mutex. */
    if (0 != (error = pthread_mutex_lock(&harness_kernel_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to lock kernel write mutex)))
      harness_syserr();
      return -1;
    }
    /* Push unlock kernel write mutex to pthread cleanup stack. */
    pthread_cleanup_push(harness_kernel_unlock, NULL);
    /* Write command to outgoing I/O pipe. */
    if (0 != harness_kernel_write(harness_kernel_data.pipes[1], 1, "I")) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write command to outgoing I/O pipe)))
      harness_syserr();
      /* Unlock kernel write mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Write handle to outgoing I/O pipe. */
    if (0 != harness_kernel_write(harness_kernel_data.pipes[1],
                                  sizeof(unsigned int), &handle)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write handle to outgoing I/O pipe)))
      harness_syserr();
      /* Unlock kernel write mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Calculate bytes to write. */
    bytes = length - index;
    if (bytes > harness_kernel_data.size) {
      bytes = harness_kernel_data.size;
    }
    /* Write bytes to outgoing I/O pipe. */
    if (0 != harness_kernel_write(harness_kernel_data.pipes[1],
                                  sizeof(unsigned int), &bytes)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write bytes to outgoing I/O pipe)))
      harness_syserr();
      /* Unlock kernel write mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Write buffer to outgoing I/O pipe. */
    if (0 != harness_kernel_write(harness_kernel_data.pipes[1],
                                  bytes, buffer + index)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write buffer to outgoing I/O pipe)))
      harness_syserr();
      /* Unlock kernel write mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Pop unlock kernel write mutex from pthread cleanup stack. */
    pthread_cleanup_pop(0);
    /* Unlock kernel write mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
      harness_syserr();
      return -1;
    }
  }
  return 0;
}

/*
 * Reads a buffer from a file descriptor (non-/blocking).
 *
 * fd     = The file descriptor to read from.
 * length = The maximum buffer length to read.
 * buffer = The buffer to read into.
 * block  = The (non-)blocking flag (0 = non-blocking, 1 = blocking).
 * return = number of bytes read on success, -2 on closed file descriptor or -1
 *          on any other error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_read"
int harness_kernel_read (int           fd,
                         unsigned int  length,
                         void         *buffer,
                         unsigned int  block) {
  int          bytes;
  unsigned int index;
#ifdef DEBUG
  /* Check length and buffer parameters. */
  if ((0 != length)&&(NULL == buffer)) {
    errno = EINVAL;
    HARNESS_LOG((
      HARNESS_WARN(buffer parameter is NULL and length parameter is not 0)))
    return -1;
  }
#endif /* DEBUG */
  for (index = 0; index < length;) {
    /* Read some data. */
    switch (bytes = read(fd, buffer + index, length - index)) {
      case -1: {
        switch (errno) {
          case EINTR: {
            break;
          }
          case EAGAIN: {
            if (0 == block) {
              return index;
            }
            break;
          }
          default: {
            HARNESS_LOG((
              HARNESS_WARN(unable to read from file descriptor)))
            return -1;
          }
        }
        break;
      }
      case 0: {
        errno = EPIPE;
        if (0 != index) {
          HARNESS_LOG((
            HARNESS_WARN(unable to read from closed file descriptor)))
        }
        return -2;
      }
      default: {
        index += bytes;
        if (0 == block) {
          return index;
        }
      }
    }
  }
  return index;
}

/*
 * Writes a buffer into a file descriptor (blocking).
 *
 * fd     = The file descriptor to write to.
 * length = The buffer length to write.
 * buffer = The buffer to write from.
 * return = 0 on success, -2 on closed file descriptor or -1 on any other error
 *          with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_write"
int harness_kernel_write (int           fd,
                          unsigned int  length,
                          const void   *buffer) {
  int          bytes;
  unsigned int index;
#ifdef DEBUG
  /* Check length and buffer parameters. */
  if ((0 != length)&&(NULL == buffer)) {
    errno = EINVAL;
    HARNESS_LOG((
      HARNESS_WARN(buffer parameter is NULL and length parameter is not 0)))
    return -1;
  }
#endif /* DEBUG */
  for (index = 0; index < length;) {
    /* Check for thread cancellation. */
    pthread_testcancel();
    /* Write some data. */
    switch (bytes = write(fd, buffer + index, length - index)) {
      case -1: {
        switch (errno) {
          case EINTR:
          case EAGAIN: {
            break;
          }
          case EPIPE: {
            if (0 != index) {
              HARNESS_LOG((
                HARNESS_WARN(unable to write to closed file descriptor)))
            }
            return -2;
          }
          default: {
            HARNESS_LOG((
              HARNESS_WARN(unable to write to file descriptor)))
            return -1;
          }
        }
        break;
      }
      default: {
        index += bytes;
      }
    }
  }
  return 0;
}

/*
 * The kernel write mutex unlock cleanup.
 *
 * argument = NULL.
 */
#undef  __FUNC__
#define __FUNC__ "harness_kernel_unlock"
void harness_kernel_unlock (void *argument) {
  int error;
  /* Unlock kernel write mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_kernel_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock kernel write mutex)))
    harness_syserr();
  }
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
