/******************************************************************************
 *
 * Source file for harness processes (process launcher) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

#define HARNESS_PROCESSES_PID_STARTED  0 /* PID of a  started process */
#define HARNESS_PROCESSES_PID_KILLED  -1 /* PID of a  killed  process */
#define HARNESS_PROCESSES_PID_EXITED  -2 /* PID of an exited  process */

#define HARNESS_PROCESSES_PIPE_OPEN    0 /* open   pipe flag */
#define HARNESS_PROCESSES_PIPE_CLOSED -1 /* closed pipe flag */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "harness.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Process launcher data type. */
typedef struct {
  pthread_mutex_t   mutex;     /* mutex                     */
  unsigned int      ready;     /* initialized flag          */
  unsigned int      closing;   /* closing flag              */
  unsigned int      count;     /* process table count       */
  struct {
    unsigned int    handle;    /* process handle            */
    pid_t           pid;       /* process system pid        */
    int             pipes[2];  /* process I/O pipes         */
    int             status;    /* process exit status       */
    pthread_cond_t *changed;   /* process changed condition */
    unsigned int    length;    /* process output length     */
    void           *output;    /* process output            */
  }                *procs;     /* process table array       */
} harness_processes_data_t;

/* Process launcher vector data type (for argv/envv cleanup). */
typedef struct {
  unsigned int   count;
  char         **array;
} harness_processes_vector_t;

/* Plug-in loader directory. */
const char *harness_execdir = HARNESS_STRING(HARNESS_EXECDIR);

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Writes a process execution request.
 *
 * handle = The process handle.
 * argc   = The argument vector count.
 * envc   = The environment vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envv   = The environment vector.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_kernel_execute (unsigned int   handle,
                            unsigned int   argc,
                            unsigned int   envc,
                            const char   **argv,
                            const char   **envv);

/*
 * Writes one or more process input requests.
 *
 * handle = The process handle.
 * length = The process input length.
 * buffer = The process input.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_kernel_forward (unsigned int  handle,
                            unsigned int  length,
                            const void   *buffer);

/*
 * Initializes the process launcher.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_processes_init ();

/*
 * Finalizes the process launcher.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_processes_fini ();

/*
 * Closes the process launcher.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_processes_close ();

/*
 * Processes an execution notification.
 *
 * handle = The process handle.
 * pid    = The process system pid.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_processes_executed (unsigned int handle,
                                pid_t        pid);

/*
 * Processes a pipe closure notification.
 *
 * handle = The process handle.
 * fd     = The closed pipe.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_processes_closed (unsigned int handle,
                              int          fd);

/*
 * Processes a process output notification.
 *
 * handle = The process handle.
 * length = The process output length.
 * buffer = The process output buffer.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_processes_output (unsigned int  handle,
                              unsigned int  length,
                              const void   *buffer);

/*
 * Processes a process exit notification.
 *
 * handle = The process handle.
 * status = The process exit status.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_processes_exited (unsigned int handle,
                              int          status);

/*
 * The process launcher mutex unlock cleanup.
 *
 * argument = NULL.
 */
void harness_processes_unlock (void *argument);

/*
 * The process launcher argc/envv vector cleanup.
 *
 * argument = The vector to cleanup..
 */
void harness_processes_vector_free (void *argument) ;

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Process launcher data. */
harness_processes_data_t harness_processes_data = {
  PTHREAD_MUTEX_INITIALIZER, /* mutex               */
  0,                         /* initialized flag    */
  0,                         /* closing flag        */
  0,                         /* process table count */
  NULL                       /* process table array */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes the process launcher.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_init"
int harness_processes_init () {
  int error;
  HARNESS_LOG((HARNESS_INFO(process launcher startup)))
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Check for initialized process launcher. */
  if (1 == harness_processes_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(process launcher already initialized)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Initialize process table. */
  harness_processes_data.count = 0;
  harness_processes_data.procs = NULL;
  /* Mark process launcher initialized. */
  harness_processes_data.closing = 0;
  harness_processes_data.ready   = 1;
  /* Unlock process launcher mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
    harness_syserr();
    /* Mark process launcher finalized. */
    harness_processes_data.ready = 0;
    return -1;
  }
  return 0;
}

/*
 * Closes the process launcher.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_close"
int harness_processes_close () {
  int error;
  HARNESS_LOG((HARNESS_INFO(process launcher closing)))
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Check for finalized process launcher. */
  if (0 == harness_processes_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(process launcher already finalized)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Mark process launcher closing. */
  harness_processes_data.closing = 1;
  /* Unlock process launcher mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  return 0;
}

/*
 * Finalizes the process launcher.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_fini"
int harness_processes_fini () {
  int          error, result;
  unsigned int index;
  result = 0;
  HARNESS_LOG((HARNESS_INFO(process launcher shutdown)))
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Check for finalized process launcher. */
  if (0 == harness_processes_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(process launcher already finalized)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Mark process launcher closing. */
  harness_processes_data.closing = 1;
  /* Finalize process table entries. */
  for (index = 0; index < harness_processes_data.count; index++) {
    /* Terminate process if still running. */
    if (0 < harness_processes_data.procs[index].pid) {
      kill(harness_processes_data.procs[index].pid, SIGTERM);
    }
    /* Broadcast process changed condition. */
    if (0 !=
        pthread_cond_broadcast(harness_processes_data.procs[index].changed)) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to broadcast process changed condition)))
      harness_syserr();
      result = -1;
    }
    if (0 !=
        pthread_cond_destroy(harness_processes_data.procs[index].changed)) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to finalize process changed condition)))
      harness_syserr();
      result = -1;
    }
    /* Deallocate process changed condition. */
    free(harness_processes_data.procs[index].changed);
    /* Deallocate process output. */
    if (NULL != harness_processes_data.procs[index].output) {
      free(harness_processes_data.procs[index].output);
    }
  }
  /* Finalize process table. */
  harness_processes_data.count = 0;
  if (NULL != harness_processes_data.procs) {
    free(harness_processes_data.procs);
    harness_processes_data.procs = NULL;
  }
  /* Mark process launcher finalized. */
  harness_processes_data.closing = 0;
  harness_processes_data.ready   = 0;
  /* Unlock process launcher mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  return result;
}

/*
 * Processes an execution notification.
 *
 * handle = The process handle.
 * pid    = The process system pid.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_executed"
int harness_processes_executed (unsigned int handle,
                                pid_t        pid) {
  int          error;
  unsigned int index;
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Check for finalized process launcher. */
  if (0 == harness_processes_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(process launcher finalized)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Search for process table entry. */
  for (index = 0; index < harness_processes_data.count; index++) {
    if (handle == harness_processes_data.procs[index].handle) {
      break;
    }
  }
  /* Check for process table entry. */
  if (index == harness_processes_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_INFO(unable to find process table entry)))
    harness_syserr();
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Save process system pid. */
  harness_processes_data.procs[index].pid = pid;
  /* Broadcast process change. */
  if (0 !=
      pthread_cond_broadcast(harness_processes_data.procs[index].changed)) {
    HARNESS_LOG((HARNESS_INFO(unable to broadcast process change)))
    harness_syserr();
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Unlock process launcher mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  return 0;
}

/*
 * Processes a pipe closure notification.
 *
 * handle = The process handle.
 * fd     = The closed pipe.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_closed"
int harness_processes_closed (unsigned int handle,
                              int          fd) {
  int          error;
  unsigned int index;
#ifdef DEBUG
  /* Check fd parameter. */
  if ((0 != fd)&&(1 != fd)) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(fd parameter is not 0 and not 1)))
    return -1;
  }
#endif /* DEBUG */
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Check for finalized process launcher. */
  if (0 == harness_processes_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(process launcher finalized)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Search for process table entry. */
  for (index = 0; index < harness_processes_data.count; index++) {
    if (handle == harness_processes_data.procs[index].handle) {
      break;
    }
  }
  /* Check for process table entry. */
  if (index == harness_processes_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_INFO(unable to find process table entry)))
    harness_syserr();
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Close process I/O pipe. */
    harness_processes_data.procs[index].pipes[fd] =
      HARNESS_PROCESSES_PIPE_CLOSED;
  /* Broadcast process change. */
  if (0 !=
      pthread_cond_broadcast(harness_processes_data.procs[index].changed)) {
    HARNESS_LOG((HARNESS_INFO(unable to broadcast process change)))
    harness_syserr();
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Unlock process launcher mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  return 0;
}

/*
 * Processes a process output notification.
 *
 * handle = The process handle.
 * length = The process output length.
 * buffer = The process output buffer.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_output"
int harness_processes_output (unsigned int  handle,
                              unsigned int  length,
                              const void   *buffer) {
  int           error;
  void         *output;
  unsigned int  index, oldlength;
#ifdef DEBUG
  /* Check length and buffer parameters. */
  if ((0 != length)&&(NULL == buffer)) {
    errno = EINVAL;
    HARNESS_LOG((
      HARNESS_WARN(buffer parameter is NULL and length parameter is not 0)))
    return -1;
  }
#endif /* DEBUG */
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Check for finalized process launcher. */
  if (0 == harness_processes_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(process launcher finalized)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Search for process table entry. */
  for (index = 0; index < harness_processes_data.count; index++) {
    if (handle == harness_processes_data.procs[index].handle) {
      break;
    }
  }
  /* Check for process table entry. */
  if (index == harness_processes_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_INFO(unable to find process table entry)))
    harness_syserr();
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }

  /* Increase process output. */
  oldlength = harness_processes_data.procs[index].length;
  harness_processes_data.procs[index].length += length;
  /* Reallocate process output. */
  if (NULL == (output = realloc(harness_processes_data.procs[index].output,
                                harness_processes_data.procs[index].length))) {
    HARNESS_LOG((HARNESS_WARN(unable to reallocate process output)))
    harness_syserr();
    /* Decrease process output. */
    harness_processes_data.procs[index].length = oldlength;
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  harness_processes_data.procs[index].output = output;
  /* Copy process output. */
  memcpy(harness_processes_data.procs[index].output + oldlength,
         buffer, length);
  /* Broadcast process change. */
  if (0 !=
      pthread_cond_broadcast(harness_processes_data.procs[index].changed)) {
    HARNESS_LOG((HARNESS_INFO(unable to broadcast process change)))
    harness_syserr();
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Unlock process launcher mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  return 0;
}

/*
 * Processes a process exit notification.
 *
 * handle = The process handle.
 * status = The process exit status.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_exited"
int harness_processes_exited (unsigned int handle,
                              int          status) {
  int          error;
  unsigned int index;
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Check for finalized process launcher. */
  if (0 == harness_processes_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(process launcher finalized)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Search for process table entry. */
  for (index = 0; index < harness_processes_data.count; index++) {
    if (handle == harness_processes_data.procs[index].handle) {
      break;
    }
  }
  /* Check for process table entry. */
  if (index == harness_processes_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_INFO(unable to find process table entry)))
    harness_syserr();
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Set process system pid. */
  harness_processes_data.procs[index].pid = HARNESS_PROCESSES_PID_EXITED;
  /* Close process I/O pipes. */
  harness_processes_data.procs[index].pipes[0] =
    HARNESS_PROCESSES_PIPE_CLOSED;
  harness_processes_data.procs[index].pipes[1] =
    HARNESS_PROCESSES_PIPE_CLOSED;
  /* Set process exit status. */
  harness_processes_data.procs[index].status = status;
  /* Broadcast process change. */
  if (0 !=
      pthread_cond_broadcast(harness_processes_data.procs[index].changed)) {
    HARNESS_LOG((HARNESS_INFO(unable to broadcast process change)))
    harness_syserr();
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Unlock process launcher mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  return 0;
}

/*
 * Starts a new process and executes a program.
 *
 * handle    = The process handle (return).
 * arguments = The program arguments (first argument is program name).
 * variables = The program variables (may be NULL).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_execute"
int harness_processes_execute (unsigned int *handle,
                               const char   *arguments,
                               const char   *variables) {
  char                        **argv1, **argv2, **envv1, **envv2;
  const char                   *token, *token1, *token2;
  unsigned int                  argc1, argc2, envc1, envc2;
  harness_processes_vector_t   args, envs;
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    return -1;
  }
  /* Check arguments parameter. */
  if (NULL == arguments) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(arguments parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Prepare arguments. */
  argc1 = 0;
  argv1 = NULL;
  /* Remove all leading spaces from arguments. */
  for (; '\0' != arguments[0] && ' ' == arguments[0]; arguments++);
  while ('\0' != arguments[0]) {
    /* Check for ' quoting in arguments. */
    if ('\'' == arguments[0]) {
      /* Search for next ' quote that is not protected by \ in arguments. */
      for (token = arguments + 1;
           ( '\0' != token[0])&&
           ((' '  != token[1])||('\'' != token[0])||('\\' == token[-1]))&&
           (('\0' != token[1])||('\'' != token[0])||('\\' == token[-1]));
           token++);
      /* Check for missing ' quote in arguments. */
      if ('\0' == token[0]) {
        errno = EINVAL;
        HARNESS_LOG((HARNESS_WARN(missing closing quote in arguments)))
        /* Deallocate arguments. */
        while (0 != argc1) {
          argc1--;
          free(argv1[argc1]);
        }
        /* Deallocate argument vector. */
        if (NULL != argv1) {
          free(argv1);
        }
        return -1;
      }
    /* Check for " quoting in arguments. */
    } else if ('\"' == arguments[0]) {
      /* Search for next " quote that is not protected by \ in arguments. */
      for (token = arguments + 1;
           ( '\0' != token[0])&&
           ((' '  != token[1])||('\"' != token[0])||('\\' == token[-1]))&&
           (('\0' != token[1])||('\"' != token[0])||('\\' == token[-1]));
           token++);
      /* Check for missing " quote in arguments. */
      if ('\0' == token[0]) {
        errno = EINVAL;
        HARNESS_LOG((HARNESS_WARN(missing closing quote in arguments)))
        /* Deallocate arguments. */
        while (0 != argc1) {
          argc1--;
          free(argv1[argc1]);
        }
        /* Deallocate argument vector. */
        if (NULL != argv1) {
          free(argv1);
        }
        return 1;
      }
    } else {
      /* Search for next space in arguments. */
      for (token = arguments; '\0' != token[0] && ' ' != token[0]; token++);
    }
    /* Increase argument vector. */
    argc2 = argc1;
    argc1++;
    /* Reallocate argument vector. */
    if (NULL == (argv2 = (char**)realloc(argv1, argc1 * sizeof(char*)))) {
      HARNESS_LOG((HARNESS_WARN(unable to reallocate argument vector)))
      /* Deallocate arguments. */
      while (0 != argc2) {
        argc2--;
        free(argv1[argc2]);
      }
      /* Deallocate argument vector. */
      if (NULL != argv1) {
        free(argv1);
      }
      return -1;
    }
    argv1 = argv2;
    /* Duplicate argument. */
    if (NULL == (argv1[argc2] = strndup(arguments, token - arguments))) {
      HARNESS_LOG((HARNESS_WARN(unable to duplicate argument)))
      /* Deallocate arguments. */
      while (0 != argc2) {
        argc2--;
        if (NULL != argv1[argc2]) {
          free(argv1[argc2]);
        }
      }
      /* Deallocate argument vector. */
      if (NULL != argv1) {
        free(argv1);
      }
      return -1;
    }
    arguments = token;
    if ('\0' != arguments[0]) {
      arguments++;
    }
    /* Remove all leading spaces in arguments. */
    for (; '\0' != arguments[0] && ' ' == arguments[0]; arguments++);
  }
  /* Prepare variables. */
  envc1 = 0;
  envv1 = NULL;
  if (NULL == variables) {
    variables = "";
  }
  /* Remove all leading spaces from variables. */
  for (; '\0' != variables[0] && ' ' == variables[0]; variables++);
  while ('\0' != variables[0]) {
    /* Search for next '=' in variables. */
    for (token = variables; '\0' != token[0] && '=' != token[0]; token++);
    /* Check for '='. */
    if ('=' == token[0]) {
      token1 = token;
      token++;
      /* Check for ' quoting in variables. */
      if ('\'' == token[0]) {
        /* Search for next ' quote that is not protected by \ in variables. */
        for (token++, token2 = token;
             ( '\0' != token[0])&&
             ((' '  != token[1])||('\'' != token[0])||('\\' == token[-1]))&&
             (('\0' != token[1])||('\'' != token[0])||('\\' == token[-1]));
             token++);
        /* Check for missing ' quote in variables. */
        if ('\0' == token[0]) {
          errno = EINVAL;
          HARNESS_LOG((HARNESS_WARN(missing closing quote in variables)))
          /* Deallocate variables. */
          while (0 != envc1) {
            envc1--;
            free(envv1[envc1]);
          }
          /* Deallocate variable vector. */
          if (NULL != envv1) {
            free(envv1);
          }
          /* Deallocate arguments. */
          while (0 != argc1) {
            argc1--;
            free(argv1[argc1]);
          }
          /* Deallocate argument vector. */
          if (NULL != argv1) {
            free(argv1);
          }
          return 1;
        }
      /* Check for " quoting in variables. */
      } else  if ('\"' == token[0]) {
        /* Search for next " quote that is not protected by \ in variables. */
        for (token++, token2 = token;
             ( '\0' != token[0])&&
             ((' '  != token[1])||('\"' != token[0])||('\\' == token[-1]))&&
             (('\0' != token[1])||('\"' != token[0])||('\\' == token[-1]));
             token++);
        /* Check for missing ' quote in variables. */
        if ('\0' == token[0]) {
          errno = EINVAL;
          HARNESS_LOG((HARNESS_WARN(missing closing quote in variables)))
          /* Deallocate variables. */
          while (0 != envc1) {
            envc1--;
            free(envv1[envc1]);
          }
          /* Deallocate variable vector. */
          if (NULL != envv1) {
            free(envv1);
          }
          /* Deallocate arguments. */
          while (0 != argc1) {
            argc1--;
            free(argv1[argc1]);
          }
          /* Deallocate argument vector. */
          if (NULL != argv1) {
            free(argv1);
          }
          return 1;
        }
      } else {
        /* Search for next space in arguments. */
        for (token2 = token; '\0' != token[0] && ' ' != token[0]; token++);
      }
      /* Increase variable vector. */
      envc2 = envc1;
      envc1++;
      /* Reallocate variable vector. */
      if (NULL == (envv2 = (char**)realloc(envv1, envc1 * sizeof(char*)))) {
        HARNESS_LOG((HARNESS_WARN(unable to reallocate variable vector)))
        /* Deallocate variables. */
        while (0 != envc2) {
          envc2--;
          free(envv1[envc2]);
        }
        /* Deallocate variable vector. */
        if (NULL != envv1) {
          free(envv1);
        }
        /* Deallocate arguments. */
        while (0 != argc1) {
          argc1--;
          free(argv1[argc1]);
        }
        /* Deallocate argument vector. */
        if (NULL != argv1) {
          free(argv1);
        }
        return -1;
      }
      envv1 = envv2;
      /* Duplicate variable. */
      if (NULL == (envv1[envc2] = strndup(variables, token - variables))) {
        HARNESS_LOG((HARNESS_WARN(unable to duplicate argument)))
        /* Deallocate variables. */
        while (0 != envc2) {
          envc2--;
          if (NULL != envv1[envc2]) {
            free(envv1[envc2]);
          }
        }
        /* Deallocate variable vector. */
        if (NULL != envv1) {
          free(envv1);
        }
        /* Deallocate arguments. */
        while (0 != argc1) {
          argc1--;
          free(argv1[argc1]);
        }
        /* Deallocate argument vector. */
        if (NULL != argv1) {
          free(argv1);
        }
        return -1;
      }
      /* Check if quotes where used. */
      if (token1 + 2 == token2) {
        /* Remove leading quote. */
        memmove(&envv1[envc2][token1 - variables + 1],
                &envv1[envc2][token2 - variables], token - token2 + 1);
      }
    }
    /* Remove all leading spaces in variables. */
    for (variables = token;
         '\0' != variables[0] && ' ' == variables[0]; variables++);
  }
  /* Push deallocate arguments and argument vector to pthread cleanup stack. */
  args.count = argc1;
  args.array = argv1;
  pthread_cleanup_push(harness_processes_vector_free, &args);
  /* Push deallocate variables and variable vector to pthread cleanup stack. */
  envs.count = envc1;
  envs.array = envv1;
  pthread_cleanup_push(harness_processes_vector_free, &envs);
  /* Start new process and execute program. */
  if (0 != harness_processes_execute2(handle, argc1, envc1,
                                       (const char**)argv1,
                                       (const char**)envv1)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to start new process and execute program)))
    /* Deallocate variables. */
    while (0 != envc1) {
      envc1--;
      free(envv1[envc1]);
    }
    /* Deallocate variable vector. */
    if (NULL != envv1) {
      free(envv1);
    }
    /* Deallocate arguments. */
    while (0 != argc1) {
      argc1--;
      free(argv1[argc1]);
    }
    /* Deallocate argument vector. */
    if (NULL != argv1) {
      free(argv1);
    }
    return -1;
  }
  /* Pop deallocate variables and variable vector from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate arguments and argument vector from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Deallocate variables. */
  while (0 != envc1) {
    envc1--;
    free(envv1[envc1]);
  }
  /* Deallocate variable vector. */
  if (NULL != envv1) {
    free(envv1);
  }
  /* Deallocate arguments. */
  while (0 != argc1) {
    argc1--;
    free(argv1[argc1]);
  }
  /* Deallocate argument vector. */
  if (NULL != argv1) {
    free(argv1);
  }
  return 0;
}

/*
 * Starts a new process and executes a program.
 *
 * handle = The process handle (return).
 * argc   = The number of arguments.
 * envc   = The number of variables.
 * argv   = The array of arguments (argv[0] is program name).
 * envv   = The array of variables.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_execute2"
int harness_processes_execute2 (unsigned int  *handle,
                                unsigned int   argc,
                                unsigned int   envc,
                                const char   **argv,
                                const char   **envv) {
  int           error;
  void         *procs;
  unsigned int  index1, index2;
  /* Check argc parameter. */
  if (0 == argc) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argc parameter is 0)))
    return -1;
  }
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    return -1;
  }
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter is null)))
    return -1;
  }
  /* Check envc and envv parameters. */
  if ((0 != envc)&&(NULL == envv)) {
    errno = EINVAL;
    HARNESS_LOG((
      HARNESS_WARN(envv parameter is NULL and envc parameter is not 0)))
    return -1;
  }
#endif /* DEBUG */
  if (NULL != strstr(argv[0], "/..")) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter contains "/..")))
    return -1;
  }
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Check for finalized process launcher. */
  if (0 == harness_processes_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(process launcher finalized)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Check for closing process launcher. */
  if (1 == harness_processes_data.closing) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(process launcher is closing)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Increase process table. */
  index1 = harness_processes_data.count;
  harness_processes_data.count++;
  /* Reallocate process table. */
  if (NULL == (procs = realloc(harness_processes_data.procs,
                               harness_processes_data.count *
                        sizeof(harness_processes_data.procs[0])))) {
    HARNESS_LOG((HARNESS_WARN(unable to reallocate process table)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  harness_processes_data.procs = procs;
  /* Allocate process changed condition. */
  if (NULL == (harness_processes_data.procs[index1].changed =
               (pthread_cond_t*)malloc(sizeof(pthread_cond_t)))) {
    HARNESS_LOG((
      HARNESS_WARN(unable to allocate process changed condition)))
    /* Decrease process table. */
    harness_processes_data.count--;
    /* Reallocate process table. */
    if (0 == harness_processes_data.count) {
      free(harness_processes_data.procs);
      harness_processes_data.procs = NULL;
    } else if (NULL == (procs =
                        realloc(harness_processes_data.procs,
                                harness_processes_data.count *
                         sizeof(harness_processes_data.procs[0])))) {
      HARNESS_LOG((HARNESS_WARN(unable to reallocate process table)))
    } else {
      harness_processes_data.procs = procs;
    }
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Initialize process changed condition. */
  if (0 != (error =
            pthread_cond_init(harness_processes_data.procs[index1].changed,
                              NULL))) {
    errno = error;
    HARNESS_LOG((
      HARNESS_WARN(unable to initialize process changed condition)))
    harness_syserr();
    /* Deallocate process changed condition. */
    free(harness_processes_data.procs[index1].changed);
    /* Decrease process table. */
    harness_processes_data.count--;
    /* Reallocate process table. */
    if (0 == harness_processes_data.count) {
      free(harness_processes_data.procs);
      harness_processes_data.procs = NULL;
    } else if (NULL == (procs =
                        realloc(harness_processes_data.procs,
                                harness_processes_data.count *
                         sizeof(harness_processes_data.procs[0])))) {
      HARNESS_LOG((HARNESS_WARN(unable to reallocate process table)))
    } else {
      harness_processes_data.procs = procs;
    }
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Randomly generate a unique non-zero 32-bit handle. */
  harness_processes_data.procs[index1].handle = 0;
  do {
    *handle = random() & 0xFFFFFFFF;
    for (index2 = 0; index2 < harness_processes_data.count; index2++) {
      if (*handle == harness_processes_data.procs[index2].handle) {
        break;
      }
    }
  } while (index2 < harness_processes_data.count);
  /* Initialize process table entry. */
  harness_processes_data.procs[index1].pid = HARNESS_PROCESSES_PID_STARTED;
  harness_processes_data.procs[index1].pipes[0] = HARNESS_PROCESSES_PIPE_OPEN;
  harness_processes_data.procs[index1].pipes[1] = HARNESS_PROCESSES_PIPE_OPEN;
  harness_processes_data.procs[index1].status   = 0;
  harness_processes_data.procs[index1].length   = 0;
  harness_processes_data.procs[index1].output   = NULL;
  harness_processes_data.procs[index1].handle   = *handle;
  /* Unlock process launcher mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Write process execution request. */
  if (0 != harness_kernel_execute(*handle, argc, envc, argv, envv)) {
    HARNESS_LOG((HARNESS_WARN(unable to write process execution request)))
    return -1;
  }
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Push unlock process launcher mutex to pthread cleanup stack. */
  pthread_cleanup_push(harness_processes_unlock, NULL);
  /* Check for finalized process launcher. */
  if (0 == harness_processes_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(process launcher finalized)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Search for process table entry. */
  for (index1 = 0; index1 < harness_processes_data.count; index1++) {
    if (*handle == harness_processes_data.procs[index1].handle) {
      break;
    }
  }
  /* Check for process table entry. */
  if (index1 == harness_processes_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_INFO(unable to find process table entry)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Ckeck process system pid. */
  if (HARNESS_PROCESSES_PID_STARTED ==
      harness_processes_data.procs[index1].pid) {
    /* Wait for process changed condition. */
    if (0 != (error =
              pthread_cond_wait(harness_processes_data.procs[index1].changed,
                                &harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to wait for process changed condition)))
      harness_syserr();
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -1;
    }
  }
  /* Pop unlock process launcher mutex from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Unlock process launcher mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  return 0;
}

/*
 * Reads the standard output of a process into a buffer (non-/blocking).
 *
 * handle = The process handle.
 * length = The maximum buffer length to read.
 * buffer = The buffer to read into.
 * block  = The (non-)blocking flag (0 = non-blocking, 1 = blocking).
 * return = number of bytes read on success, -2 on closed pipe or -1 on any
 *          other error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_read"
int harness_processes_read (unsigned int  handle,
                            unsigned int  length,
                            void         *buffer,
                            unsigned int  block) {
  int           error;
  void         *output;
  unsigned int  index;
#ifdef DEBUG
  /* Check length and buffer parameters. */
  if ((0 != length)&&(NULL == buffer)) {
    errno = EINVAL;
    HARNESS_LOG((
      HARNESS_WARN(buffer parameter is NULL and length parameter is not 0)))
    return -1;
  }
#endif /* DEBUG */
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Push unlock process launcher mutex to pthread cleanup stack. */
  pthread_cleanup_push(harness_processes_unlock, NULL);
  while (1) {
    /* Check for finalized process launcher. */
    if (0 == harness_processes_data.ready) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(process launcher finalized)))
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Search for process table entry. */
    for (index = 0; index < harness_processes_data.count; index++) {
      if (handle == harness_processes_data.procs[index].handle) {
        break;
      }
    }
    /* Check for process table entry. */
    if (index == harness_processes_data.count) {
      errno = ESRCH;
      HARNESS_LOG((HARNESS_INFO(unable to find process table entry)))
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Ckeck for no data and closed pipe. */
    if ((0  == harness_processes_data.procs[index].length)&&
        (HARNESS_PROCESSES_PIPE_CLOSED ==
         harness_processes_data.procs[index].pipes[0])) {
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -2;
    }
    /* Ckeck for enough data. */
    if (length <= harness_processes_data.procs[index].length) {
      /* Copy output. */
      memcpy(buffer, harness_processes_data.procs[index].output, length);
      /* Remove process output. */
      harness_processes_data.procs[index].length -= length;
      memmove(harness_processes_data.procs[index].output,
              harness_processes_data.procs[index].output + length,
              harness_processes_data.procs[index].length);
      /* Reallocate process output. */
      if (0 == harness_processes_data.procs[index].length) {
        free(harness_processes_data.procs[index].output);
        harness_processes_data.procs[index].output = NULL;
      } else if (NULL ==
                 (output =
                  realloc(harness_processes_data.procs[index].output,
                          harness_processes_data.procs[index].length))) {
        HARNESS_LOG((
          HARNESS_WARN(unable to reallocate process output)))
      } else {
        harness_processes_data.procs[index].output = output;
      }
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return length;
    }
    /* Ckeck for some data and closed pipe or non-blocking read. */
    if ((HARNESS_PROCESSES_PIPE_CLOSED ==
         harness_processes_data.procs[index].pipes[0])||(0 == block)) {
      /* Adjust length. */
      length = harness_processes_data.procs[index].length;
      /* Copy output. */
      memcpy(buffer, harness_processes_data.procs[index].output, length);
      /* Remove process output. */
      harness_processes_data.procs[index].length -= length;
      memmove(harness_processes_data.procs[index].output,
              harness_processes_data.procs[index].output + length,
              harness_processes_data.procs[index].length);
      /* Reallocate process output. */
      if (0 == harness_processes_data.procs[index].length) {
        free(harness_processes_data.procs[index].output);
        harness_processes_data.procs[index].output = NULL;
      } else if (NULL ==
                 (output =
                  realloc(harness_processes_data.procs[index].output,
                          harness_processes_data.procs[index].length))) {
        HARNESS_LOG((
          HARNESS_WARN(unable to reallocate process output)))
      } else {
        harness_processes_data.procs[index].output = output;
      }
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return length;
    }
    /* Check for closing process launcher. */
    if (1 == harness_processes_data.closing) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(process launcher is closing)))
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Wait for process changed condition. */
    if (0 != (error =
              pthread_cond_wait(harness_processes_data.procs[index].changed,
                                &harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to wait for process changed condition)))
      harness_syserr();
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -1;
    }
  }
  /* Pop unlock process launcher mutex from pthread cleanup stack. */
  pthread_cleanup_pop(0);
}

/*
 * Writes a buffer into the standard input of a process (blocking).
 *
 * handle = The process handle.
 * length = The length to write.
 * buffer = The buffer to write.
 * return = 0 on success, -2 on closed pipe or -1 on any other error with errno
 *          set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_write"
int harness_processes_write (unsigned int  handle,
                             unsigned int  length,
                             const void   *buffer) {
  int          error;
  unsigned int index;
#ifdef DEBUG
  /* Check length and buffer parameters. */
  if ((0 != length)&&(NULL == buffer)) {
    errno = EINVAL;
    HARNESS_LOG((
      HARNESS_WARN(buffer parameter is NULL and length parameter is not 0)))
    return -1;
  }
#endif /* DEBUG */
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Check for finalized process launcher. */
  if (0 == harness_processes_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(process launcher finalized)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Search for process table entry. */
  for (index = 0; index < harness_processes_data.count; index++) {
    if (handle == harness_processes_data.procs[index].handle) {
      break;
    }
  }
  /* Check for process table entry. */
  if (index == harness_processes_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_INFO(unable to find process table entry)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Ckeck for closed pipe. */
  if (HARNESS_PROCESSES_PIPE_CLOSED ==
      harness_processes_data.procs[index].pipes[1]) {
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -2;
  }
  /* Unlock process launcher mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Write process input request. */
  if (0 != harness_kernel_forward(handle, length, buffer)) {
    HARNESS_LOG((HARNESS_WARN(unable to write process input request)))
    return -1;
  }
  return 0;
}

/*
 * Sends a signal to a process
 *
 * handle = The process handle.
 * signum = The signal to send.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_signal"
int harness_processes_signal (unsigned int handle,
                              int          signum) {
  int          error;
  unsigned int index;
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Check for finalized process launcher. */
  if (0 == harness_processes_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(process launcher finalized)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Search for process table entry. */
  for (index = 0; index < harness_processes_data.count; index++) {
    if (handle == harness_processes_data.procs[index].handle) {
      break;
    }
  }
  /* Check for process table entry. */
  if (index == harness_processes_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_INFO(unable to find process table entry)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Send signal to process. */
  if ((0  <  harness_processes_data.procs[index].pid              )&&
      (-1 == kill(harness_processes_data.procs[index].pid, signum))) {
    HARNESS_LOG((HARNESS_INFO(unable to send signal to process)))
    /* Unlock process launcher mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Unlock process launcher mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  return 0;
}

/*
 * Waits for the exit of a process, but does not remove the entry in the kernel
 * process table.
 *
 * handle = The process handle.
 * status = The process exit status (return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_wait"
int harness_processes_wait (unsigned int  handle,
                            int          *status) {
  int          error;
  unsigned int index;
#ifdef DEBUG
  /* Check status parameter. */
  if (NULL == status) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(status parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Push unlock process launcher mutex to pthread cleanup stack. */
  pthread_cleanup_push(harness_processes_unlock, NULL);
  while (1) {
    /* Check for finalized process launcher. */
    if (0 == harness_processes_data.ready) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(process launcher finalized)))
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Search for process table entry. */
    for (index = 0; index < harness_processes_data.count; index++) {
      if (handle == harness_processes_data.procs[index].handle) {
        break;
      }
    }
    /* Check for process table entry. */
    if (index == harness_processes_data.count) {
      errno = ESRCH;
      HARNESS_LOG((HARNESS_INFO(unable to find process table entry)))
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Check if process exited. */
    if (HARNESS_PROCESSES_PID_EXITED ==
        harness_processes_data.procs[index].pid) {
      /* Save process exit status. */
      *status = harness_processes_data.procs[index].status;
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
        return -1;
      }
      return 0;
    }
    /* Check for closing process launcher. */
    if (1 == harness_processes_data.closing) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(process launcher is closing)))
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Wait for process changed condition. */
    if (0 != (error =
              pthread_cond_wait(harness_processes_data.procs[index].changed,
                                &harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to wait for process changed condition)))
      harness_syserr();
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -1;
    }
  }
  /* Pop unlock process launcher mutex from pthread cleanup stack. */
  pthread_cleanup_pop(0);
}

/*
 * Terminates a process, waits for its exit and removes the entry in the kernel
 * process table.
 *
 * handle = The process handle.
 * status = The process exit status (return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_terminate"
int harness_processes_terminate (unsigned int  handle,
                                 int          *status) {
  int           error;
  void         *procs;
  unsigned int  index;
#ifdef DEBUG
  /* Check status parameter. */
  if (NULL == status) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(status parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Lock process launcher mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock process launcher mutex)))
    harness_syserr();
    return -1;
  }
  /* Push unlock process launcher mutex to pthread cleanup stack. */
  pthread_cleanup_push(harness_processes_unlock, NULL);
  while (1) {
    /* Check for finalized process launcher. */
    if (0 == harness_processes_data.ready) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(process launcher finalized)))
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Search for process table entry. */
    for (index = 0; index < harness_processes_data.count; index++) {
      if (handle == harness_processes_data.procs[index].handle) {
        break;
      }
    }
    /* Check for process table entry. */
    if (index == harness_processes_data.count) {
      errno = ESRCH;
      HARNESS_LOG((HARNESS_INFO(unable to find process table entry)))
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Terminate process if still running. */
    if (0 < harness_processes_data.procs[index].pid) {
      kill(harness_processes_data.procs[index].pid, SIGTERM);
      harness_processes_data.procs[index].pid = HARNESS_PROCESSES_PID_KILLED;
    }
    /* Check if process exited. */
    if (HARNESS_PROCESSES_PID_EXITED ==
        harness_processes_data.procs[index].pid) {
      /* Save process exit status. */
      *status = harness_processes_data.procs[index].status;
      /* Broadcast process changed condition. */
      if (0 != pthread_cond_broadcast(
                 harness_processes_data.procs[index].changed)) {
        errno = error;
        HARNESS_LOG((
          HARNESS_WARN(unable to broadcast process changed condition)))
        harness_syserr();
        /* Unlock process launcher mutex. */
        if (0 != (error =
                  pthread_mutex_unlock(&harness_processes_data.mutex))) {
          errno = error;
          HARNESS_LOG((
            HARNESS_WARN(unable to unlock process launcher mutex)))
          harness_syserr();
        }
        return -1;
      }
      if (0 != pthread_cond_destroy(
                 harness_processes_data.procs[index].changed)) {
        HARNESS_LOG((
          HARNESS_WARN(unable to finalize process changed condition)))
        errno = error;
        harness_syserr();
        /* Unlock process launcher mutex. */
        if (0 != (error =
                  pthread_mutex_unlock(&harness_processes_data.mutex))) {
          errno = error;
          HARNESS_LOG((
            HARNESS_WARN(unable to unlock process launcher mutex)))
          harness_syserr();
        }
        return -1;
      }
      /* Deallocate process changed condition. */
      free(harness_processes_data.procs[index].changed);
      /* Deallocate process output. */
      if (NULL != harness_processes_data.procs[index].output) {
        free(harness_processes_data.procs[index].output);
      }
      /* Remove process table entry. */
      harness_processes_data.count--;
      memmove(harness_processes_data.procs + index,
              harness_processes_data.procs + index + 1,
             (harness_processes_data.count - index) *
       sizeof(harness_processes_data.procs[0]));
      /* Reallocate process table. */
      if (0 == harness_processes_data.count) {
        free(harness_processes_data.procs);
        harness_processes_data.procs = NULL;
      } else if (NULL == (procs = realloc(harness_processes_data.procs,
                                          harness_processes_data.count *
                                   sizeof(harness_processes_data.procs[0])))) {
        HARNESS_LOG((
          HARNESS_WARN(unable to reallocate process table entry)))
      } else {
        harness_processes_data.procs = procs;
      }
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
        return -1;
      }
      return 0;
    }
    /* Check for closing process launcher. */
    if (1 == harness_processes_data.closing) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(process launcher is closing)))
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -1;
    }
    /* Wait for process changed condition. */
    if (0 != (error =
              pthread_cond_wait(harness_processes_data.procs[index].changed,
                                &harness_processes_data.mutex))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to wait for process changed condition)))
      harness_syserr();
      /* Unlock process launcher mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
        harness_syserr();
      }
      return -1;
    }
  }
  /* Pop unlock process launcher mutex from pthread cleanup stack. */
  pthread_cleanup_pop(0);
}

/*
 * The process launcher mutex unlock cleanup.
 *
 * argument = NULL.
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_unlock"
void harness_processes_unlock (void *argument) {
  int error;
  /* Unlock process launcher mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_processes_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock process launcher mutex)))
    harness_syserr();
  }
}

/*
 * The process launcher argc/envv vector cleanup.
 *
 * argument = The vector to cleanup..
 */
#undef  __FUNC__
#define __FUNC__ "harness_processes_vector_free"
void harness_processes_vector_free (void *argument) {
  harness_processes_vector_t *vector;
#ifdef DEBUG
  /* Check argument parameter. */
  if (NULL == argument) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argument parameter is null)))
    return;
  }
#endif /* DEBUG */
  vector = argument;
  /* Deallocate variables. */
  while (0 != vector->count) {
    vector->count--;
    free(vector->array[vector->count]);
  }
  /* Deallocate variable vector. */
  if (NULL != vector->array) {
    free(vector->array);
  }
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
