/******************************************************************************
 *
 * Source file for harness-rmix (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "harness-rmix.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Harness-rmix plug-in module data type. */
typedef struct {
  pthread_mutex_t  mutex;     /* mutex              */
  struct {
    unsigned int   count;     /* handle array count */
    unsigned int  *array;     /* handle array       */
  }                instances; /* instances          */
} harness_rmix_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Initializes the harness-rmix plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
HARNESS_PLUGINS_INIT(harness_rmix_init);

/*
 * Finalizes the harness-rmix plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
HARNESS_PLUGINS_FINI(harness_rmix_fini);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Harness-rmix plug-in module data. */
harness_rmix_data_t harness_rmix_data = {
  PTHREAD_MUTEX_INITIALIZER, /* mutex              */
  {
    0,                       /* handle array count */
    NULL                     /* handle array       */
  }                          /* instances          */
};

/* Harness-rmix plug-in "object". */
harness_rmix_t harness_rmix = {
  {
    LIBHARNESS_RMIX_VERSION_CURRENT,  /* current version         */
    LIBHARNESS_RMIX_VERSION_REVISION, /* current revision        */
    LIBHARNESS_RMIX_VERSION_AGE,      /* version age             */
    LIBHARNESS_RMIX_VERSION_FIRST     /* first supported version */
  },                                     /* plug-in version         */
  /**********************************************/
  /* PUT YOUR PUBLIC PLUG-IN DATA POINTERS HERE */
  /**********************************************/
  /**************************************************/
  /* PUT YOUR PUBLIC PLUG-IN FUNCTION POINTERS HERE */
  /**************************************************/
  harness_rmix_instances
};

/* Server-side method descriptors for harness. */
const rmix_method_t harness_rmix_methods[HARNESS_RMIX_METHODS_COUNT] =
  HARNESS_RMIX_METHODS;

/* Server-side interface for harness. */
const rmix_interface_t harness_rmix_interface = {
  HARNESS_RMIX_METHODS_COUNT,           /* method descriptor count */
  (rmix_method_t*)harness_rmix_methods  /* method descriptor array */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes the harness-rmix plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_init"
HARNESS_PLUGINS_INIT(harness_rmix_init) {
  int           error;
  void         *array;
  unsigned int  index;
  HARNESS_LOG((HARNESS_INFO(harness-rmix startup)))
  /* Lock harness-rmix plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_rmix_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock harness-rmix plug-in mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Search for handle in handle array. */
  for (index = 0; index < harness_rmix_data.instances.count; index++) {
    if (harness_rmix_data.instances.array[index] == handle) {
      errno = EALREADY;
      HARNESS_LOG((HARNESS_WARN(handle is already in handle array)))
      errno = EALREADY;
      harness_syserr();
      /* Unlock harness-rmix plug-in mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_rmix_data.mutex))) {
        errno = error;
        HARNESS_LOG((
          HARNESS_WARN(unable to unlock harness-rmix plug-in mutex)))
        errno = error;
        harness_syserr();
      }
      errno = EALREADY;
      return -1;
    }
  }
  /* Increase handle array. */
  index = harness_rmix_data.instances.count;
  harness_rmix_data.instances.count++;
  /* Reallocate handle array. */
  if (NULL == (array = realloc(harness_rmix_data.instances.array,
                               harness_rmix_data.instances.count *
                        sizeof(harness_rmix_data.instances.array[0])))) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to reallocate handle array)))
    /* Decrease handle array. */
    harness_rmix_data.instances.count--;
    /* Unlock harness-rmix plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_rmix_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock harness-rmix plug-in mutex)))
      errno = error;
      harness_syserr();
    }
    errno2 = errno;
    return -1;
  }
  harness_rmix_data.instances.array = array;
  /* Save handle array entry. */
  harness_rmix_data.instances.array[index] = handle;
  /* Check for first initialization. */
  if (1 == harness_rmix_data.instances.count) {
    /* Initialize rmix with default configuration. */
    if (0 != rmix_init2()) {
      int errno2 = errno;
      HARNESS_LOG((
        HARNESS_WARN(unable to initialize rmix with default configuration)))
      /* Deeallocate handle array. */
      free(harness_rmix_data.instances.array);
      harness_rmix_data.instances.array = NULL;
      harness_rmix_data.instances.count   = 0;
      /* Unlock harness-rmix plug-in mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_rmix_data.mutex))) {
        errno = error;
        HARNESS_LOG((
          HARNESS_WARN(unable to unlock harness-rmix plug-in mutex)))
        errno = error;
        harness_syserr();
      }
      errno = errno2;
    }
  }
  /* Unlock harness-rmix plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_rmix_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock harness-rmix plug-in mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Finalizes the harness-rmix plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_fini"
HARNESS_PLUGINS_FINI(harness_rmix_fini) {
  int           error, result, errno2;
  void         *array;
  unsigned int  index;
  result = 0;
  errno2 = 0;
  HARNESS_LOG((HARNESS_INFO(harness-rmix shutdown)))
  /* Lock harness-rmix plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_rmix_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock harness-rmix plug-in mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Search for handle in handle array. */
  for (index = 0; index < harness_rmix_data.instances.count; index++) {
    if (harness_rmix_data.instances.array[index] == handle) {
      break;
    }
  }
  /* Check if handle is not in handle array. */
  if (index == harness_rmix_data.instances.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_WARN(handle is not in handle array)))
    errno = ESRCH;
    harness_syserr();
    /* Unlock harness-rmix plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_rmix_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock harness-rmix plug-in mutex)))
      errno = error;
      harness_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Remove handle from handle array. */
  harness_rmix_data.instances.count--;
  memmove(harness_rmix_data.instances.array + index,
          harness_rmix_data.instances.array + index + 1,
         (harness_rmix_data.instances.count   - index) *
   sizeof(harness_rmix_data.instances.array[0]));
  /* Reallocate handle array. */
  if (0 == harness_rmix_data.instances.count) {
    free(harness_rmix_data.instances.array);
    harness_rmix_data.instances.array = NULL;
  } else if (NULL ==
             (array = realloc(harness_rmix_data.instances.array,
                              harness_rmix_data.instances.count *
                       sizeof(harness_rmix_data.instances.array[0])))) {
    HARNESS_LOG((HARNESS_WARN(unable to reallocate handle array)))
    errno = 0;
  } else {
    harness_rmix_data.instances.array = array;
  }
  /* Check for last finalization. */
  if (0 == harness_rmix_data.instances.count) {
    /* Finalize rmix. */
    if (0 != rmix_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize rmix)))
      result = -1;
    }
  }
  /* Unlock harness-rmix plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_rmix_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock harness-rmix plug-in mutex)))
    errno = error;
    harness_syserr();
    if (0 != errno2) {
      errno = errno2;
    } else {
      errno = error;
    }
    return -1;
  }
  if (0 != errno2) {
    errno = errno2;
  }
  return result;
}

/*
 * Returns the number of loaded instances of this harness-rmix plug-in
 * without locking its plug-in mutex.
 *
 * return = The number of loaded instances of this harness-rmix plug-in.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_instances"
unsigned int harness_rmix_instances () {
  return harness_rmix_data.instances.count;
}

/*
 * Exports harness without forcing a specific object handle.
 *
 * localref     = The local object reference of the exported harness (return).
 * protocol     = The communication protocol. Overrides any additional PROTOCOL
 *                parameter. May only be NULL or "" if additional parameters
 *                contains PROTOCOL.
 * binding      = The binding  (protocol specific) (may be NULL, ""). Overrides
 *                any additional BINDING parameter.
 * parameters   = Additional parameters (protocol specific) (may be NULL, "").
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_export"
int harness_rmix_export (rmix_localref_t         **localref,
                         const char               *protocol,
                         const char               *binding,
                         const rmix_parameters_t  *parameters) {
  /* Export harness. */
  if (0 != rmix_export(localref, protocol, binding, parameters, NULL,
                       &harness_rmix_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export harness)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports harness without forcing a specific object handle and without
 * loading the provider plug-in.
 *
 * localref     = The local object reference of the exported harness (return).
 * provider     = The preloaded RMIX provider plug-in handle.
 * protocol     = The communication protocol. Overrides any additional PROTOCOL
 *                parameter. May only be NULL or "" if additional parameters
 *                contains PROTOCOL.
 * binding      = The binding  (protocol specific) (may be NULL, ""). Overrides
 *                any additional BINDING parameter.
 * parameters   = Additional parameters (protocol specific) (may be NULL, "").
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_export_p"
int harness_rmix_export_p (rmix_localref_t         **localref,
                           unsigned int              provider,
                           const char               *protocol,
                           const char               *binding,
                           const rmix_parameters_t  *parameters) {
  /* Export harness. */
  if (0 != rmix_export_p(localref, provider, protocol, binding, parameters,
                         NULL, &harness_rmix_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export harness)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports harness while forcing a specific object handle.
 *
 * localref     = The local object reference of the exported harness (return).
 * protocol     = The communication protocol. Overrides any additional PROTOCOL
 *                parameter. May only be NULL or "" if additional parameters
 *                contains PROTOCOL.
 * binding      = The binding  (protocol specific) (may be NULL, ""). Overrides
 *                any additional BINDING parameter.
 * parameters   = Additional parameters (protocol specific) (may be NULL, "").
 * handle       = The object handle.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_export2"
int harness_rmix_export2 (rmix_localref_t         **localref,
                          const char               *protocol,
                          const char               *binding,
                          const rmix_parameters_t  *parameters,
                          unsigned int              handle) {
  /* Export harness. */
  if (0 != rmix_export2(localref, protocol, binding, parameters, handle,
                        NULL, &harness_rmix_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export harness)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports harness while forcing a specific object handle and without
 * loading the provider plug-in.
 *
 * localref     = The local object reference of the exported harness (return).
 * provider     = The preloaded RMIX provider plug-in handle.
 * protocol     = The communication protocol. Overrides any additional PROTOCOL
 *                parameter. May only be NULL or "" if additional parameters
 *                contains PROTOCOL.
 * binding      = The binding  (protocol specific) (may be NULL, ""). Overrides
 *                any additional BINDING parameter.
 * parameters   = Additional parameters (protocol specific) (may be NULL, "").
 * handle       = The object handle.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_export2_p"
int harness_rmix_export2_p (rmix_localref_t         **localref,
                            unsigned int              provider,
                            const char               *protocol,
                            const char               *binding,
                            const rmix_parameters_t  *parameters,
                            unsigned int              handle) {
  /* Export harness. */
  if (0 != rmix_export2_p(localref, provider, protocol, binding, parameters,
                          handle, NULL, &harness_rmix_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export harness)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports harness while forcing a specific object handle. This is very
 * useful for standard object exports.
 *
 * localref   = The local object reference of the exported object (return).
 * parameters = The parameters must include PROTOCOL (for protocol) and
 *              OBJECTID (for handle), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_export3"
int harness_rmix_export3 (rmix_localref_t         **localref,
                          const rmix_parameters_t  *parameters) {
  /* Export harness. */
  if (0 != rmix_export3(localref, parameters,
                        NULL, &harness_rmix_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export harness)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports harness while forcing a specific object handle and without
 * loading the provider plug-in. This is very useful for standard object
 * exports.
 *
 * localref   = The local object reference of the exported object (return).
 * provider   = The preloaded RMIX provider plug-in handle.
 * parameters = The parameters must include PROTOCOL (for protocol) and
 *              OBJECTID (for handle), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_export3_p"
int harness_rmix_export3_p (rmix_localref_t         **localref,
                            unsigned int              provider,
                            const rmix_parameters_t  *parameters) {
  /* Export harness. */
  if (0 != rmix_export3_p(localref, provider, parameters,
                          NULL, &harness_rmix_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export harness)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports harness while forcing a specific object handle. This is very
 * useful for standard object exports.
 *
 * localref   = The local object reference of the exported object (return).
 * parameters = The parameter pairs must include PROTOCOL (for protocol) and
 *              OBJECTID (for handle), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_export4"
int harness_rmix_export4 (rmix_localref_t **localref,
                          const char       *parameters) {
  /* Export harness. */
  if (0 != rmix_export4(localref, parameters, NULL, &harness_rmix_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export harness)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports harness while forcing a specific object handle and without
 * loading the provider plug-in. This is very useful for standard object
 * exports.
 *
 * localref   = The local object reference of the exported object (return).
 * provider   = The preloaded RMIX provider plug-in handle.
 * parameters = The parameter pairs must include PROTOCOL (for protocol) and
 *              OBJECTID (for handle), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_export4_p"
int harness_rmix_export4_p (rmix_localref_t **localref,
                            unsigned int      provider,
                            const char       *parameters) {
  /* Export harness. */
  if (0 != rmix_export4_p(localref, provider, parameters,
                          NULL, &harness_rmix_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export harness)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Unexports harness.
 *
 * localref = The local object reference of the exported object (also return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_unexport"
int harness_rmix_unexport (rmix_localref_t **localref) {
  /* Unexport harness. */
  if (0 != rmix_unexport(localref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to unexport harness)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Unexports harness without unloading the provider plug-in.
 *
 * localref = The local object reference of the exported harness (also return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_unexport_p"
int harness_rmix_unexport_p (rmix_localref_t **localref) {
  /* Unexport harness. */
  if (0 != rmix_unexport_p(localref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to unexport harness)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load (server-side method stub).
 * Loads a plug-in.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_plugins_load_call"
RMIX_METHOD_CALL(harness_rmix_plugins_load_call) {
  int           result;
  const char   *name;
  unsigned int  flags, handle;
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  name  =                  inary[0];
  flags = *(unsigned int*)(inary[1]);
  /* Call method function. */
  result = harness_plugins_load(&handle, name, flags);
  /* Prepare output. */
  *(int*         )(outary[0]) = result;
  *(unsigned int*)(outary[1]) = handle;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load2 (server-side method stub).
 * Loads a plug-in.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_plugins_load2_call"
RMIX_METHOD_CALL(harness_rmix_plugins_load2_call) {
  int           result;
  const char   *name;
  unsigned int  version, flags, handle;
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  name    =                  inary[0];
  version = *(unsigned int*)(inary[1]);
  flags   = *(unsigned int*)(inary[2]);
  /* Call method function. */
  result = harness_plugins_load2(&handle, name, version, flags);
  /* Prepare output. */
  *(int*         )(outary[0]) = result;
  *(unsigned int*)(outary[1]) = handle;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load3 (server-side method stub).
 * Loads a plug-in.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_plugins_load3_call"
RMIX_METHOD_CALL(harness_rmix_plugins_load3_call) {
  int           result;
  const char   *name;
  unsigned int  first, age, revision, flags, handle;
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  name    =                   inary[0];
  first    = *(unsigned int*)(inary[1]);
  age      = *(unsigned int*)(inary[2]);
  revision = *(unsigned int*)(inary[3]);
  flags    = *(unsigned int*)(inary[4]);
  /* Call method function. */
  result = harness_plugins_load3(&handle, name, first, age, revision, flags);
  /* Prepare output. */
  *(int*         )(outary[0]) = result;
  *(unsigned int*)(outary[1]) = handle;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes harness_plugins_load4 (server-side method stub).
 * Loads a plug-in.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_plugins_load4_call"
RMIX_METHOD_CALL(harness_rmix_plugins_load4_call) {
  int           result;
  const char   *name;
  unsigned int  current, age, revision, flags, handle;
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  name     =                  inary[0];
  current  = *(unsigned int*)(inary[1]);
  age      = *(unsigned int*)(inary[2]);
  revision = *(unsigned int*)(inary[3]);
  flags    = *(unsigned int*)(inary[4]);
  /* Call method function. */
  result = harness_plugins_load4(&handle, name, current, age, revision, flags);
  /* Prepare output. */
  *(int*         )(outary[0]) = result;
  *(unsigned int*)(outary[1]) = handle;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes harness_plugins_name (server-side method stub).
 * Loads a plug-in.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_plugins_name_call"
RMIX_METHOD_CALL(harness_rmix_plugins_name_call) {
  int           result;
  const char   *name;
  unsigned int  handle;
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  handle = *(unsigned int*)(inary[0]);
  /* Call method function. */
  result = harness_plugins_name(&name, handle);
  /* Prepare output. */
  *(int*)(outary[0]) = result;
  /* Duplicate name parameter. */
  if (NULL == (outary[1] = strdup(name))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate name parameter)))
    errno = errno2;
    return -1;
  }
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes harness_plugins_unload (server-side method stub).
 * Loads a plug-in.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_plugins_unload_call"
RMIX_METHOD_CALL(harness_rmix_plugins_unload_call) {
  int          result;
  unsigned int handle;
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  handle = *(unsigned int*)(inary[0]);
  /* Call method function. */
  result = harness_plugins_unload(handle);
  /* Prepare output. */
  *(int*)(outary[0]) = result;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes harness_processes_execute (server-side method stub).
 * Starts a new process and executes a program.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_processes_execute_call"
RMIX_METHOD_CALL(harness_rmix_processes_execute_call) {
  int           result;
  const char   *arguments, *variables;
  unsigned int  handle;
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  arguments = inary[0];
  variables = inary[1];
  /* Call method function. */
  result = harness_processes_execute(&handle, arguments, variables);
  /* Prepare output. */
  *(int*         )(outary[0]) = result;
  *(unsigned int*)(outary[1]) = handle;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes harness_processes_execute2 (server-side method stub).
 * Starts a new process and executes a program.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_processes_execute2_call"
RMIX_METHOD_CALL(harness_rmix_processes_execute2_call) {
  int            result;
  const char   **argv, **envv;
  unsigned int   handle, argc, envc;
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  argc = *(unsigned int*)(inary[0]);
  argv =  (const char** )(inary[1]);
  envc = *(unsigned int*)(inary[2]);
  envv =  (const char** )(inary[3]);
  /* Call method function. */
  result = harness_processes_execute2(&handle, argc, envc, argv, envv);
  /* Prepare output. */
  *(int*         )(outary[0]) = result;
  *(unsigned int*)(outary[1]) = handle;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes harness_processes_read (server-side method stub).
 * Reads the standard output of a process into a buffer (non-/blocking).
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_processes_read_call"
RMIX_METHOD_CALL(harness_rmix_processes_read_call) {
  int           result;
  void         *buffer;
  unsigned int  handle, length, block;
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  handle = *(unsigned int*)(inary[0]);
  length = *(unsigned int*)(inary[1]);
  block  = *(unsigned int*)(inary[2]);
  /* Allocate buffer. */
  if (NULL == (buffer = malloc(length))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate buffer)))
    errno = errno2;
    return -1;
  }
  /* Call method function. */
  result = harness_processes_read(handle, length, buffer, block);
  /* Prepare output. */
  *(int*         )(outary[0]) = result;
  *(unsigned int*)(outary[1]) = length;
                   outary[2]  = buffer;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes harness_processes_write (server-side method stub).
 * Writes a buffer into the standard input of a process (blocking).
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_processes_write_call"
RMIX_METHOD_CALL(harness_rmix_processes_write_call) {
  int           result;
  const void   *buffer;
  unsigned int  handle, length;
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  handle = *(unsigned int*)(inary[0]);
  length = *(unsigned int*)(inary[1]);
  buffer =                  inary[2];
  /* Call method function. */
  result = harness_processes_write(handle, length, buffer);
  /* Prepare output. */
  *(int*)(outary[0]) = result;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes harness_processes_signal (server-side method stub).
 * Sends a signal to a process.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_processes_signal_call"
RMIX_METHOD_CALL(harness_rmix_processes_signal_call) {
  int           signum, result;
  unsigned int  handle;
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  handle = *(unsigned int*)(inary[0]);
  signum = *(int*         )(inary[1]);
  /* Call method function. */
  result = harness_processes_signal(handle, signum);
  /* Prepare output. */
  *(int*)(outary[0]) = result;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes harness_processes_wait (server-side method stub).
 * Waits for the exit of a process, but does not
 * remove the entry in the kernel process table.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_processes_wait_call"
RMIX_METHOD_CALL(harness_rmix_processes_wait_call) {
  int          status, result;
  unsigned int handle;
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  handle = *(unsigned int*)(inary[0]);
  /* Call method function. */
  result = harness_processes_wait(handle, &status);
  /* Prepare output. */
  *(int*)(outary[0]) = result;
  *(int*)(outary[1]) = status;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes harness_processes_terminate (server-side method stub).
 * Terminates a process, waits for its exit and
 * removes the entry in the kernel process table.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_processes_terminate_call"
RMIX_METHOD_CALL(harness_rmix_processes_terminate_call) {
  int          status, result;
  unsigned int handle;
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  handle = *(unsigned int*)(inary[0]);
  /* Call method function. */
  result = harness_processes_terminate(handle, &status);
  /* Prepare output. */
  *(int*)(outary[0]) = result;
  *(int*)(outary[0]) = status;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes harness_kernel_shutdown_call (server-side method stub).
 * Asynchonously shuts down the kernel and subsequently the forker.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_rmix_kernel_shutdown_call"
RMIX_METHOD_CALL(harness_rmix_kernel_shutdown_call) {
#ifdef DEBUG
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Call method function. */
  harness_kernel_shutdown();
  /* Reset errno if needed. */
  errno = 0;
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
