/******************************************************************************
 *
 * Header file for rmix localref (local object reference) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef RMIX_LOCALREF_H
#define RMIX_LOCALREF_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Local object reference data type. */
typedef struct {
  rmix_remoteref_t *remoteref; /* remote object reference */
  unsigned int      provider;  /* provider plug-in handle */
  unsigned int      objectid;    /* object id               */
  void             *object;    /* object (may be NULL)    */
  rmix_interface_t *interface; /* object interface        */
} rmix_localref_t;             /* local object reference  */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Creates a local object reference.
 *
 * localref  = The local object reference to create (return).
 * remoteref = The remote object reference.
 * provider  = The provider plug-in handle.
 * objectid  = The object id.
 * object    = The object (may be NULL).
 * interface = The object interface.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_localref_create (rmix_localref_t        **localref,
                          const rmix_remoteref_t  *remoteref,
                          unsigned int             provider,
                          unsigned int             objectid,
                          void                    *object,
                          const rmix_interface_t  *interface);

/*
 * Creates a local object reference.
 *
 * localref   = The local object reference to create (return).
 * protocol   = The communication protocol.
 * binding    = The binding  (protocol specific) (may be NULL, "").
 * interfaces = The remote object interfaces list (NULL terminated or NULL).
 * location   = The remote object location (protocol specific).
 * provider   = The provider plug-in handle.
 * objectid   = The object id.
 * object     = The object (may be NULL).
 * interface  = The object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_localref_create2 (rmix_localref_t        **localref,
                           const char              *protocol,
                           const char              *binding,
                           const char             **interfaces,
                           const char              *location,
                           unsigned int             provider,
                           unsigned int             objectid,
                           void                    *object,
                           const rmix_interface_t  *interface);

/*
 * Creates a local object reference.
 *
 * localref   = The local object reference to create (return).
 * protocol   = The communication protocol.
 * binding    = The binding  (protocol specific) (may be NULL, "").
 * interfaces = The remote object interfaces list string (may be NULL).
 *              Individual interface names are delimited by ';'.
 * location   = The remote object location (protocol specific).
 * provider   = The provider plug-in handle.
 * objectid   = The object id.
 * object     = The object (may be NULL).
 * interface  = The object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_localref_create3 (rmix_localref_t        **localref,
                           const char              *protocol,
                           const char              *binding,
                           const char              *interfaces,
                           const char              *location,
                           unsigned int             provider,
                           unsigned int             objectid,
                           void                    *object,
                           const rmix_interface_t  *interface);
/*
 * Duplicates a local object reference.
 *
 * localref1 = The local object reference to create (return).
 * localref2 = The local object reference to duplicate.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_localref_duplicate (rmix_localref_t       **localref1,
                             const rmix_localref_t  *localref2);

/*
 * Destroys a local object reference.
 *
 * localref = The local object reference to destroy (also return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_localref_destroy (rmix_localref_t **localref);

/*
 * Compares two local object references.
 *
 * diff       = The difference between both local object references (return).
 *              diff > 0 if localref1 < localref2,
 *              diff < 0 if localref1 > localref2, and
 *              diff = 0 if localref1 = localref2
 * remoteref1 = The local object reference to compare against.
 * remoteref2 = The local object reference to compare with.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_localref_compare (int                   *diff,
                           const rmix_localref_t *localref1,
                           const rmix_localref_t *localref2);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

#endif /* RMIX_LOCALREF_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
