/******************************************************************************
 *
 * Source file for rmix providers (provider plug-in registry) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmix.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Looks up the provider plug-in name of a protocol.
 *
 * provider = The provider plug-in name (return).
 * protocol = The protocol to lookup.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_providers_lookup"
int rmix_providers_lookup (char       **provider,
                           const char  *protocol) {
  int             file, bytes;
  char           *home, *filename, *path, *content, *begin, *end, *delim;
  off_t           offset;
  unsigned int    homelen, filelen, length, index, mark;
  struct timeval  time1, time2;
#ifdef DEBUG
  /* Check provider parameter. */
  if (NULL == provider) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(provider parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check protocol parameter. */
  if (NULL == protocol) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(protocol parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check protocol parameter. */
  if ('\0' == protocol[0]) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(protocol parameter is an empty string)))
    errno = EINVAL;
    return -1;
  }
  if (NULL != (home = getenv("HOME"))) {
    filename = RMIX_STRING(/RMIX_USRCONFDIR/RMIX_CONFFILE);
    homelen  = strlen(home);
    filelen  = strlen(filename) + 1;
    /* Allocate user configuration file path. */
    if (NULL == (path = malloc((homelen + filelen) * sizeof(char)))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to allocate user configuration file path)))
      errno = errno2;
      return -1;
    }
    /* Compose user configuration file path. */
    memcpy(path          , home    , homelen);
    memcpy(path + homelen, filename, filelen);
  } else {
    /* Duplicate user configuration file path. */
    if (NULL == (path = strdup(RMIX_STRING(~/RMIX_USRCONFDIR/RMIX_CONFFILE)))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate user configuration file path)))
      errno = errno2;
      return -1;
    }
  }
  /* Get time of day for waiting file open timeout. */
  if (0 != gettimeofday(&time1, NULL)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get time of day)))
    errno = errno2;
    rmix_syserr();
    /* Deallocate user configuration file path. */
    free(path);
    errno = errno2;
    return -1;
  }
  /* Open user configuration file. */
  while (-1 == (file = open(path, O_RDONLY))) {
    if ((EACCES != errno)&&(EINTR != errno)) {
      break;
    }
    /* Get time of day for waiting file open timeout. */
    if (0 != gettimeofday(&time2, NULL)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to get time of day)))
      errno = errno2;
      rmix_syserr();
      /* Deallocate user configuration file path. */
      free(path);
      errno = errno2;
      return -1;
    }
    /* Check if waiting file open timeout is reached. */
    if (time1.tv_sec + RMIX_PROVIDERS_TIMEOUT < time2.tv_sec) {
      errno = ETIMEDOUT;
      RMIX_LOG((
        RMIX_INFO(unable to access user configuration file '%s' after timeout),
                  path))
      errno = ETIMEDOUT;
      rmix_syserr();
      /* Deallocate user configuration file path. */
      free(path);
      errno = ETIMEDOUT;
      return -1;
    }
  }
  /* Deallocate user configuration file path. */
  free(path);
  /* Check for open user configuration file. */
  if (-1 != file) {
    /* Seek to end of user configuration file. */
    if (-1 == (offset = lseek(file, 0, SEEK_END))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to seek to end of user configuration file)))
      rmix_syserr();
      /* Close configuration file. */
      if (0 != close(file)) {
        int errno3 = errno;
        RMIX_LOG((RMIX_WARN(unable to close user configuration file)))
        errno = errno3;
        rmix_syserr();
      }
      errno = errno2;
      return -1;
    }
    /* Calculate user configuration file length. */
    length = offset;
    /* Seek to start of user configuration file. */
    if (-1 == lseek(file, 0, SEEK_SET)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to seek to start of user configuration file)))
      rmix_syserr();
      /* Close configuration file. */
      if (0 != close(file)) {
        int errno3 = errno;
        RMIX_LOG((RMIX_WARN(unable to close user configuration file)))
        errno = errno3;
        rmix_syserr();
      }
      errno = errno2;
      return -1;
    }
    /* Check for empty user configuration file. */
    if (0 == length) {
      /* Close configuration file. */
      if (0 != close(file)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to close configuration file)))
        errno = errno2;
        rmix_syserr();
        errno = errno2;
        return -1;
      }
    } else {
      /* Allocate user configuration file content. */
      if (NULL == (content = malloc(length + 1))) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to allocate configuration file content)))
        /* Close user configuration file. */
        if (0 != close(file)) {
          int errno3 = errno;
          RMIX_LOG((RMIX_WARN(unable to close configuration file)))
          errno = errno3;
          rmix_syserr();
        }
        errno = errno2;
        return -1;
      }
      /* Read user configuration file content. */
      for (index = 0; index < length; index += bytes) {
        /* Read some data. */
        if (-1 == (bytes = read(file, content + index, length - index))) {
          /* Check for other than interruption. */
          if ((errno != EINTR)&&(errno != EAGAIN)) {
            int errno2 = errno;
            RMIX_LOG((RMIX_WARN(unable to read from user configuration file)))
            errno = errno2;
            rmix_syserr();
            /* Close user configuration file. */
            if (0 != close(file)) {
              int errno3 = errno;
              RMIX_LOG((RMIX_WARN(unable to close user configuration file)))
              errno = errno3;
              rmix_syserr();
            }
            /* Deallocate user configuration file content. */
            free(content);
            errno = errno2;
            return -1;
          }
        }
      }
      content[length] = '\0';
      /* Close user configuration file. */
      if (0 != close(file)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to close user configuration file)))
        errno = errno2;
        rmix_syserr();
        /* Deallocate configuration file content. */
        free(content);
        errno = errno2;
        return -1;
      }
      /* Remove any comments from user configuration file content. */
      for (index = 0; index < length; index++) {
        if ('#' == content[index]) {
          mark = index;
          for (index++; index < length && '\n' != content[index]; index++);
          memmove(content + mark,
                  content + index, (length - index + 1) * sizeof(char));
          length -= (index - mark);
          index   = mark;
        }
      }
      /* Search for begin of providers section. */
      if (NULL == (begin = strstr(content, "BEGIN PROVIDERS\n"))) {
        errno = EINVAL;
        RMIX_LOG((RMIX_WARN(configuration file format error)))
        errno = EINVAL;
        rmix_syserr();
        /* Deallocate configuration file content. */
        free(content);
        errno = EINVAL;
        return -1;
      }
      /* Search for end of providers section. */
      if (NULL == (end = strstr(begin, "END PROVIDERS\n"))) {
        errno = EINVAL;
        RMIX_LOG((RMIX_WARN(configuration file format error)))
        errno = EINVAL;
        rmix_syserr();
        /* Deallocate configuration file content. */
        free(content);
        errno = EINVAL;
        return -1;
      }
      /* Isolate providers section. */
      length = end - (begin + 15);
      memmove(content, begin + 15, length);
      content[length] = '\0';
      /* Isolate provider. */
      for (begin = content;
           NULL != (begin = strstr(begin, protocol)); begin++) {
        if (('\n'  == begin[-1]                   )&&
            (NULL  != (end   = strchr (begin, '\n')))&&
            (NULL  != (delim = strrchr(end,    ' ')))&&
            (delim == (begin + strlen (protocol   )))) {
          end[0] = '\0';
          begin  = delim + 1;
          /* Duplicate provider plug-in name. */
          if (NULL == (*provider = strdup(begin))) {
            int errno2 = errno;
            RMIX_LOG((RMIX_WARN(unable to duplicate provider plug-in name)))
            /* Deallocate user configuration file content. */
            free(content);
            errno = errno2;
            return -1;
          }
          /* Deallocate configuration file content. */
          free(content);
          return 0;
        }
      }
    }
  }
  /* Get time of day for waiting file open timeout. */
  if (0 != gettimeofday(&time1, NULL)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get time of day)))
    errno = errno2;
    rmix_syserr();
    errno = errno2;
    return -1;
  }
  /* Open system configuration file. */
  while (-1 == (file = open(RMIX_STRING(RMIX_SYSCONFDIR/RMIX_CONFFILE),
                            O_RDONLY))) {
    if ((EACCES != errno)&&(EINTR != errno)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_INFO(unable to open system configuration file '%s'),
                          RMIX_STRING(RMIX_SYSCONFDIR/RMIX_CONFFILE)))
      errno = errno2;
      rmix_syserr();
      errno = errno2;
      return -1;
    }
    /* Get time of day for waiting file open timeout. */
    if (0 != gettimeofday(&time2, NULL)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to get time of day)))
      errno = errno2;
      rmix_syserr();
      errno = errno2;
      return -1;
    }
    /* Check if waiting file open timeout is reached. */
    if (time1.tv_sec + RMIX_PROVIDERS_TIMEOUT < time2.tv_sec) {
      errno = ETIMEDOUT;
      RMIX_LOG((RMIX_INFO(unable to open configuration file '%s' after timeout),
                          RMIX_STRING(RMIX_SYSCONFDIR/RMIX_CONFFILE)))
      errno = ETIMEDOUT;
      rmix_syserr();
      errno = ETIMEDOUT;
      return -1;
    }
  }
  /* Seek to end of system configuration file. */
  if (-1 == (offset = lseek(file, 0, SEEK_END))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to seek to end of system configuration file)))
    errno = errno2;
    rmix_syserr();
    /* Close system configuration file. */
    if (0 != close(file)) {
      int errno3 = errno;
      RMIX_LOG((RMIX_WARN(unable to close system configuration file)))
      errno = errno3;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Calculate system configuration file length. */
  length = offset;
  /* Seek to start of system configuration file. */
  if (-1 == lseek(file, 0, SEEK_SET)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to seek to start of system configuration file)))
    errno = errno2;
    rmix_syserr();
    /* Close system configuration file. */
    if (0 != close(file)) {
      int errno3 = errno;
      RMIX_LOG((RMIX_WARN(unable to close system configuration file)))
      errno = errno3;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Check for empty system configuration file. */
  if (0 == length) {
    errno = ESRCH;
    RMIX_LOG((RMIX_WARN(unable to find provider)))
    /* Close system configuration file. */
    if (0 != close(file)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to close system configuration file)))
      errno = errno2;
      rmix_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Allocate system configuration file content. */
  if (NULL == (content = malloc(length + 1))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate system configuration file content)))
    rmix_syserr();
    /* Close configuration file. */
    if (0 != close(file)) {
      int errno3 = errno;
      RMIX_LOG((RMIX_WARN(unable to close system configuration file)))
      errno = errno3;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Read configuration file content. */
  for (index = 0; index < length; index += bytes) {
    /* Read some data. */
    if (-1 == (bytes = read(file, content + index, length - index))) {
      /* Check for other than interruption. */
      if ((errno != EINTR)&&(errno != EAGAIN)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to read from configuration file)))
        errno = errno2;
        rmix_syserr();
        /* Close system configuration file. */
        if (0 != close(file)) {
          int errno3 = errno;
          RMIX_LOG((RMIX_WARN(unable to close system configuration file)))
          errno = errno3;
          rmix_syserr();
        }
        /* Deallocate system configuration file content. */
        free(content);
        errno = errno2;
        return -1;
      }
    }
  }
  content[length] = '\0';
  /* Close system configuration file. */
  if (0 != close(file)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to close system configuration file)))
    errno = errno2;
    rmix_syserr();
    /* Deallocate system configuration file content. */
    free(content);
    errno = errno2;
    return -1;
  }
  /* Remove any comments from system configuration file content. */
  for (index = 0; index < length; index++) {
    if ('#' == content[index]) {
      mark = index;
      for (index++; index < length && '\n' != content[index]; index++);
      memmove(content + mark,
              content + index, (length - index + 1) * sizeof(char));
      length -= (index - mark);
      index   = mark;
    }
  }
  /* Search for begin of providers section. */
  if (NULL == (begin = strstr(content, "BEGIN PROVIDERS\n"))) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(configuration file format error)))
    errno = EINVAL;
    rmix_syserr();
    /* Deallocate system configuration file content. */
    free(content);
    errno = EINVAL;
    return -1;
  }
  /* Search for end of providers section. */
  if (NULL == (end = strstr(begin, "END PROVIDERS\n"))) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(configuration file format error)))
    errno = EINVAL;
    rmix_syserr();
    /* Deallocate system configuration file content. */
    free(content);
    errno = EINVAL;
    return -1;
  }
  /* Isolate providers section. */
  begin += 15;
  length = end - begin;
  memmove(content, begin, length);
  content[length] = '\0';
  /* Isolate provider. */
  for (begin = content + 1;
       NULL != (begin = strstr(begin, protocol)); begin++) {
    if (('\n'  == begin[-1]                     )&&
        (NULL  != (end   = strchr (begin, '\n')))) {
      end[0] = '\0';
      if ((NULL  != (delim = strrchr(begin, ' ')))&&
          (delim == (begin + strlen (protocol  )))) {
        begin  = delim + 1;
        /* Duplicate provider plug-in name. */
        if (NULL == (*provider = strdup(begin))) {
          int errno2 = errno;
          RMIX_LOG((RMIX_WARN(unable to duplicate provider plug-in name)))
          /* Deallocate system configuration file content. */
          free(content);
          errno = errno2;
          return -1;
        }
        /* Deallocate system configuration file content. */
        free(content);
        return 0;
      }
    }
  }
  errno = ESRCH;
  /* Unable to find provider. */
  RMIX_LOG((RMIX_WARN(unable to find provider)))
  /* Deallocate system configuration file content. */
  free(content);
  errno = ESRCH;
  return -1;
}

/*
 * Looks up the provider plug-in name of a protocol and loads the provider
 * plug-in.
 *
 * handle   = The provider plug-in handle (return).
 * protocol = The protocol to lookup and load.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_providers_load"
int rmix_providers_load (unsigned int *handle,
                         const char   *protocol) {
  char *provider;
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check protocol parameter. */
  if (NULL == protocol) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(protocol parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Look up provider plug-in name. */
  if (0 != rmix_providers_lookup(&provider, protocol)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to look up provider plug-in name)))
    errno = errno2;
    return -1;
  }
  /* Load provider plug-in. */
  if (0 != rmix_plugins_load(handle, provider, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to load provider plug-in)))
    /* Deallocate provider plug-in name. */
    free(provider);
    errno = errno2;
    return -1;
  }
  /* Deallocate provider plug-in name. */
  free(provider);
  return 0;
}

/*
 * Loads the provider plug-in.
 *
 * handle   = The provider plug-in handle (return).
 * provider = The protocol to lookup and load.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_providers_load2"
int rmix_providers_load2 (unsigned int *handle,
                          const char   *provider) {
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check provider parameter. */
  if (NULL == provider) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(provider parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Load provider plug-in. */
  if (0 != rmix_plugins_load(handle, provider, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to load provider plug-in)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Gets the symbol of a provider plug-in corresponding to a name (using dlsym).
 *
 * symbol = The plug-in symbol that corresponds to the name (return).
 * handle = The plug-in handle.
 * name   = The plug-in symbol name to resolve.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_providers_symbol"
int rmix_providers_symbol (void         **symbol,
                           unsigned int   handle,
                           const char    *name) {
#ifdef DEBUG
  /* Check symbol parameter. */
  if (NULL == symbol) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(symbol parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Get provider plug-in symbol. */
  if (0 != rmix_plugins_symbol(symbol, handle, name)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get provider plug-in symbol)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Gets the symbol of a provider plug-in corresponding to a name (using dlsym).
 * Prepends <provider name>_ (without any lib prefix) to the symbol name.
 *
 * symbol = The plug-in symbol that corresponds to the name (return).
 * handle = The plug-in handle.
 * name   = The plug-in symbol name to resolve.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_providers_symbol2"
int rmix_providers_symbol2 (void         **symbol,
                            unsigned int   handle,
                            const char    *name) {
#ifdef DEBUG
  /* Check symbol parameter. */
  if (NULL == symbol) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(symbol parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Get provider plug-in symbol. */
  if (0 != rmix_plugins_symbol2(symbol, handle, name)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get provider plug-in symbol)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Unloads a provider plug-in.
 *
 * provider = The provider plug-in handle.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_providers_unload"
int rmix_providers_unload (unsigned int handle) {
  /* Unload provider plug-in. */
  if (0 != rmix_plugins_unload(handle)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to unload provider plug-in)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
