/******************************************************************************
 *
 * Header file for rmix providers (provider plug-in registry) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef RMIX_PROVIDERS_H
#define RMIX_PROVIDERS_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* Rmix configuration file opening timeout for provider lookup. */
#define RMIX_PROVIDERS_TIMEOUT 120

/* Provider "invoke" function signature. */
/*
 * Invokes a method at a remote object.
 *
 * remoteref = The remote object reference.
 * interface = The remote object interface.
 * method    = The index of the method to invoke.
 * outary    = The output values array.
 * outcnt    = The output values count.
 * inary     = The input values array.
 * incnt     = The input values count.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#define RMIX_PROVIDERS_INVOKE(name)\
  int name (const rmix_remoteref_t  *remoteref,\
            const rmix_interface_t  *interface,\
            unsigned int             method,\
            void                   **outary,\
            unsigned int             outcnt,\
            const void             **inary,\
            unsigned int             incnt)

/* Provider "invoke one-way" function signature. */
/*
 * Invokes a method at a remote object one-way, i.e. there is no return value
 * and invocation returns immediately after sending message.
 *
 * remoteref = The remote object reference.
 * interface = The remote object interface.
 * method    = The index of the method to invoke.
 * inary     = The input values array.
 * incnt     = The input values count.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#define RMIX_PROVIDERS_ONEWAY(name)\
  int name (const rmix_remoteref_t  *remoteref,\
            const rmix_interface_t  *interface,\
            unsigned int             method,\
            const void             **inary,\
            unsigned int             incnt)

/* Provider "invoke asynchronous" function signature. */
/*
 * Asynchronously invokes a method at a remote object, i.e. invocation returns
 * immediately after sending message. The output may be retrieved later using
 * the invocation id. The provider gets not unloaded until the output was
 * retrieved.
 *
 * remoteref  = The remote object reference.
 * interface  = The remote object interface.
 * method     = The index of the method to invoke.
 * inary      = The input values array.
 * incnt      = The input values count.
 * invocation = The invocation id (return).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#define RMIX_PROVIDERS_SEND(name)\
  int name (const rmix_remoteref_t  *remoteref,\
            const rmix_interface_t  *interface,\
            unsigned int             method,\
            const void             **inary,\
            unsigned int             incnt,\
            unsigned int            *invocation)

/* Provider "retrieve asynchronous" function signature. */
/*
 * Retrieves the output (if any) of an asynchronous invocation using an
 * invokation id previously issued by the provider.
 *
 * remoteref  = The remote object reference.
 * interface  = The remote object interface.
 * method     = The index of the method to invoke.
 * outary     = The output values array.
 * outcnt     = The output values count.
 * invocation = The invocation id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#define RMIX_PROVIDERS_RETRIEVE(name)\
  int name (const rmix_remoteref_t  *remoteref,\
            const rmix_interface_t  *interface,\
            unsigned int             method,\
            void                   **outary,\
            unsigned int             outcnt,\
            unsigned int             invocation)

/* Provider "export" function signature. */
/*
 * Exports a local object.
 *
 * localref   = The local object reference of the exported object (return).
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may be NULL, "").
 *              Ignores any PROTOCOL or BINDING parameters.
 * provider   = The provider plug-in handle.
 * objectid   = The object id.
 * object     = The local object (may be NULL).
 * interface  = The local object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#define RMIX_PROVIDERS_EXPORT(name)\
  int name (rmix_localref_t         **localref,\
            const char               *protocol,\
            const char               *binding,\
            const rmix_parameters_t  *parameters,\
            unsigned int              provider,\
            unsigned int              objectid,\
            void                     *object,\
            const rmix_interface_t   *interface)

/* Provider "unexport" function signature. */
/*
 * Unxports a local object.
 *
 * localref = The local object reference of the exported object (also return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#define RMIX_PROVIDERS_UNEXPORT(name)\
  int name (rmix_localref_t **localref)

/*
 * Creates a remote object reference, while forcing a specific object id.
 * This is very useful for accessing standard object exports at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may be NULL, "").
 *              Ignores any PROTOCOL or BINDING parameters.
 * objectid   = The object id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#define RMIX_PROVIDERS_REMOTEREF_CREATE(name)\
  int name (rmix_remoteref_t        **remoteref,\
            const char               *protocol,\
            const char               *binding,\
            const rmix_parameters_t  *parameters,\
            unsigned int              objectid)

/*
 * Returs a remote object reference in parameters form.
 *
 * parameters = The parameters form to create.
 * remoteref  = The remote object reference to convert.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#define RMIX_PROVIDERS_REMOTEREF_PARAMETERS(name)\
  int name (rmix_parameters_t      **parameters,\
            const rmix_remoteref_t  *remoteref)

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Provider "invoke" function data type. */
typedef RMIX_PROVIDERS_INVOKE((*rmix_providers_invoke_t));

/* Provider "invoke one-way" function data type. */
typedef RMIX_PROVIDERS_ONEWAY((*rmix_providers_oneway_t));

/* Provider "invoke asynchronous" function data type. */
typedef RMIX_PROVIDERS_SEND((*rmix_providers_send_t));

/* Provider "retrieve asynchronous" function data type. */
typedef RMIX_PROVIDERS_RETRIEVE((*rmix_providers_retrieve_t));

/* Provider "export" function data type. */
typedef RMIX_PROVIDERS_EXPORT((*rmix_providers_export_t));

/* Provider "unexport" function data type. */
typedef RMIX_PROVIDERS_UNEXPORT((*rmix_providers_unexport_t));

/* Provider "remoteref_create" function data type. */
typedef RMIX_PROVIDERS_REMOTEREF_CREATE((*rmix_providers_remoteref_create_t));

/* Provider "remoteref_parameters" function data type. */
typedef RMIX_PROVIDERS_REMOTEREF_PARAMETERS(
          (*rmix_providers_remoteref_parameters_t));

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Looks up the provider plug-in name of a protocol.
 *
 * provider = The provider plug-in name (return).
 * protocol = The protocol to lookup.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_providers_lookup (char       **provider,
                           const char  *protocol);

/*
 * Looks up the provider plug-in name of a protocol and loads the provider
 * plug-in.
 *
 * handle   = The provider plug-in handle (return).
 * protocol = The protocol to lookup and load.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_providers_load (unsigned int *handle,
                         const char   *protocol);

/*
 * Loads the provider plug-in.
 *
 * handle   = The provider plug-in handle (return).
 * provider = The protocol to lookup and load.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_providers_load2 (unsigned int *handle,
                          const char   *provider);

/*
 * Gets the symbol of a provider plug-in corresponding to a name (using dlsym).
 *
 * symbol = The plug-in symbol that corresponds to the name (return).
 * handle = The plug-in handle.
 * name   = The plug-in symbol name to resolve.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_providers_symbol (void         **symbol,
                           unsigned int   handle,
                           const char    *name);

/*
 * Gets the symbol of a provider plug-in corresponding to a name (using dlsym).
 * Prepends <provider name>_ (without any lib prefix) to the symbol name.
 *
 * symbol = The plug-in symbol that corresponds to the name (return).
 * handle = The plug-in handle.
 * name   = The plug-in symbol name to resolve.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_providers_symbol2 (void         **symbol,
                            unsigned int   handle,
                            const char    *name);

/*
 * Unloads a provider plug-in.
 *
 * provider = The provider plug-in handle.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_providers_unload (unsigned int handle);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

#endif /* RMIX_PROVIDERS_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
