/******************************************************************************
 *
 * Header file for rmix remoteref (remote object reference) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef RMIX_REMOTEREF_H
#define RMIX_REMOTEREF_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Remote object reference data type. */
typedef struct {
  char            *protocol;   /* communication protocol       */
  char            *binding;    /* binding  (protocol specific) */
  struct {
    unsigned int   count;      /* interfaces list count        */
    char         **array;      /* interfaces list array        */
  }                interfaces; /* interfaces list              */
  char            *location;   /* location (protocol specific) */
  char            *object;     /* object identifier            */
} rmix_remoteref_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Creates a remote object reference.
 *
 * remoteref  = The remote object reference to create (return).
 * protocol   = The communication protocol.
 * binding    = The binding (protocol specific, may be NULL, "").
 * interfaces = The interfaces list (NULL terminated or NULL).
 * location   = The location (protocol specific).
 * object     = The object identifier.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_create (rmix_remoteref_t **remoteref,
                           const char        *protocol,
                           const char        *binding,
                           const char       **interfaces,
                           const char        *location,
                           const char        *object);

/*
 * Creates a remote object reference.
 *
 * remoteref  = The remote object reference to create (return).
 * protocol   = The communication protocol.
 * binding    = The binding (protocol specific, may be NULL, "").
 * interfaces = The interfaces list (NULL terminated or NULL).
 * location   = The location (protocol specific).
 * objectid   = The object id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_create2 (rmix_remoteref_t **remoteref,
                            const char        *protocol,
                            const char        *binding,
                            const char       **interfaces,
                            const char        *location,
                            unsigned int       objectid);

/*
 * Creates a remote object reference.
 *
 * remoteref  = The remote object reference to create (return).
 * protocol   = The communication protocol.
 * binding    = The binding (protocol specific, may be NULL, "").
 * interfaces = The interfaces list string (may be NULL).
 *              Individual interface names are delimited by ';'.
 * interfaces = The interfaces list (NULL terminated or NULL).
 * location   = The location (protocol specific).
 * objectid   = The object id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_create3 (rmix_remoteref_t **remoteref,
                            const char        *protocol,
                            const char        *binding,
                            const char        *interfaces,
                            const char        *location,
                            unsigned int       objectid);

/*
 * Creates a remote object reference using the appropiate provider plug-in,
 * while forcing a specific object id. This is very useful for accessing
 * standard object exports at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter. May only be NULL or "" if additional parameters
 *              contains PROTOCOL.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may be NULL).
 * objectid   = The object id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_create4 (rmix_remoteref_t        **remoteref,
                            const char               *protocol,
                            const char               *binding,
                            const rmix_parameters_t  *parameters,
                            unsigned int              objectid);

/*
 * Creates a remote object reference without loading the provider plug-in,
 * while forcing a specific object id. This is very useful for accessing
 * standard object exports at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * provider   = The preloaded provider plug-in handle.
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter. May only be NULL or "" if additional parameters
 *              contains PROTOCOL.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may be NULL).
 * objectid   = The object id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_create4_p (rmix_remoteref_t        **remoteref,
                              unsigned int              provider,
                              const char               *protocol,
                              const char               *binding,
                              const rmix_parameters_t  *parameters,
                              unsigned int              objectid);

/*
 * Creates a remote object reference using the appropiate provider plug-in and
 * without un loading the provider plug-in, while forcing a specific object
 * id. This is very useful for accessing standard object exports at the
 * client side.
 *
 * remoteref  = The remote object reference to create (return).
 * provider   = The provider plug-in handle (return).
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter. May only be NULL or "" if additional parameters
 *              contains PROTOCOL.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may be NULL).
 * objectid   = The object id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_create4_p2 (rmix_remoteref_t        **remoteref,
                               unsigned int             *provider,
                               const char               *protocol,
                               const char               *binding,
                               const rmix_parameters_t  *parameters,
                               unsigned int              objectid);

/*
 * Creates a remote object reference using the appropiate provider plug-in,
 * while forcing a specific object id. This is very useful for accessing
 * standard object exports at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * parameters = The parameters must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_create5 (rmix_remoteref_t        **remoteref,
                            const rmix_parameters_t  *parameters);

/*
 * Creates a remote object reference without loading the provider plug-in,
 * while forcing a specific object id. This is very useful for accessing
 * standard object exports at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * provider   = The preloaded provider plug-in handle.
 * parameters = The parameters must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_create5_p (rmix_remoteref_t        **remoteref,
                              unsigned int              provider,
                              const rmix_parameters_t  *parameters);

/*
 * Creates a remote object reference using the appropiate provider plug-in and
 * without un loading the provider plug-in, while forcing a specific object
 * id. This is very useful for accessing standard object exports at the
 * client side.
 *
 * remoteref  = The remote object reference to create (return).
 * provider   = The provider plug-in handle (return).
 * parameters = The parameters must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_create5_p2 (rmix_remoteref_t        **remoteref,
                               unsigned int             *provider,
                               const rmix_parameters_t  *parameters);

/*
 * Creates a remote object reference using the appropiate provider plug-in,
 * while forcing a specific object id. This is very useful for accessing
 * standard object exports at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * parameters = The parameter pairs must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_create6 (rmix_remoteref_t **remoteref,
                            const char        *parameters);

/*
 * Creates a remote object reference without loading the provider plug-in,
 * while forcing a specific object id. This is very useful for accessing
 * standard object exports at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * provider   = The preloaded provider plug-in handle.
 * parameters = The parameter pairs must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_create6_p (rmix_remoteref_t **remoteref,
                              unsigned int       provider,
                              const char        *parameters);

/*
 * Creates a remote object reference using the appropiate provider plug-in and
 * without un loading the provider plug-in, while forcing a specific object
 * id. This is very useful for accessing standard object exports at the
 * client side.
 *
 * remoteref  = The remote object reference to create (return).
 * provider   = The provider plug-in handle (return).
 * parameters = The parameter pairs must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_create6_p2 (rmix_remoteref_t **remoteref,
                               unsigned int      *provider,
                               const char        *parameters);

/*
 * Returs a remote object reference in parameters form using the appropiate
 * provider plug-in.
 *
 * parameters = The parameters form to create.
 * remoteref  = The remote object reference to convert.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_parameters (rmix_parameters_t      **parameters,
                               const rmix_remoteref_t  *remoteref);

/*
 * Returs a remote object reference in parameters form without loading the
 * provider plug-in.
 *
 * parameters = The parameters form to create.
 * provider   = The preloaded provider plug-in handle.
 * remoteref  = The remote object reference to convert.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_parameters_p (rmix_parameters_t      **parameters,
                                 unsigned int             provider,
                                 const rmix_remoteref_t  *remoteref);

/*
 * Returs a remote object reference in parameters form using the appropiate
 * provider plug-in.
 *
 * parameters = The parameters form to create.
 * remoteref  = The remote object reference to convert.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_parameters2 (char                   **parameters,
                                const rmix_remoteref_t  *remoteref);

/*
 * Returs a remote object reference in parameters form without loading the
 * provider plug-in.
 *
 * parameters = The parameters form to create.
 * provider   = The preloaded provider plug-in handle.
 * remoteref  = The remote object reference to convert.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_parameters2_p (char                   **parameters,
                                  unsigned int             provider,
                                  const rmix_remoteref_t  *remoteref);

/*
 * Duplicates a remote object reference.
 *
 * remoteref1 = The remote object reference to create (return).
 * remoteref2 = The remote object reference to duplicate.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_duplicate (rmix_remoteref_t       **remoteref1,
                              const rmix_remoteref_t  *remoteref2);

/*
 * Destroys a remote object reference.
 *
 * remoteref = The remote object reference to destroy (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_destroy (rmix_remoteref_t **remoteref);

/*
 * Gets the object id from a remote object reference.
 *
 * remoteref = The remote object reference.
 * objectid  = The object id (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_objectid (const rmix_remoteref_t *remoteref,
                             unsigned int           *objectid);

/*
 * Compares two remote object references.
 *
 * diff       = The difference between both remote object references (return).
 *              diff > 0 if remoteref1 < remoteref2,
 *              diff < 0 if remoteref1 > remoteref2, and
 *              diff = 0 if remoteref1 = remoteref2
 * remoteref1 = The remote object reference to compare against.
 * remoteref2 = The remote object reference to compare with.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_compare (int                    *diff,
                            const rmix_remoteref_t *remoteref1,
                            const rmix_remoteref_t *remoteref2);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

#endif /* RMIX_REMOTEREF_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
