/******************************************************************************
 *
 * Header file for rmix tcpip4 (TCP/IPv4 pool) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef RMIX_TCPIP4_H
#define RMIX_TCPIP4_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/*
 * TCP/IPv4 pool module defaults.
 */
#define RMIX_TCPIP4_BACKLOG 100 /* listen backlog  */

/* Server job function signature. */
/*
 * Serves a connection.
 *
 * server = The TCP/IPv4 server.
 * return = The server job function pointer on success or
 *          NULL on any error with errno set appropriately.
 */
#define RMIX_TCPIP4_SERVERJOB(name)\
  void* name (rmix_tcpip4_server_t *server)

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* TCP/IPv4 pool listener data type. */
typedef struct {
  int                socket;     /* socket         */
  struct sockaddr_in localaddr;  /* local address  */
} rmix_tcpip4_listener_t;

/* TCP/IPv4 pool server data type. */
typedef struct {
  int                socket;     /* socket         */
  struct sockaddr_in localaddr;  /* local address  */
  struct sockaddr_in remoteaddr; /* remote address */
} rmix_tcpip4_server_t;

/* TCP/IPv4 pool client data type. */
typedef struct {
  int                socket;     /* socket         */
  struct sockaddr_in localaddr;  /* local address  */
  struct sockaddr_in remoteaddr; /* remote address */
} rmix_tcpip4_client_t;

/* Thread pool job function data type. */
typedef RMIX_TCPIP4_SERVERJOB((*rmix_tcpip4_serverjob_t));

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Returns the first local non-loopback address, if any, or the loopback
 * address. This function may be called before initializing.
 *
 * localaddr = The first local non-loopback address, if any, or the loopback
 *             address (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_address (struct sockaddr_in *localaddr);

/*
 * Resolves a name/port pair to address/port. If name is NULL, the first local
 * non-loopback address, if any, or the loopback address is assigned.
 * This function may be called before initializing.
 *
 * localaddr = The local address/port to resolve (return).
 * localname = The local name to resolve (may be NULL).
 * localport = The local port to resolve.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_resolve (struct sockaddr_in *localaddr,
                         const char         *localname,
                         unsigned int        localport);

/*
 * Opens a TCP/IPv4 listener and binds it to all local interfaces.
 * This function may be called before initializing.
 *
 * listener = The TCP/IPv4 listener (return).
 * port     = The port to bind to.
 * backlog  = The listen backlog.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_listener_open (rmix_tcpip4_listener_t **listener,
                               unsigned int             localport,
                               unsigned int             backlog);

/*
 * Opens a TCP/IPv4 listener and binds it to the addressed local interface.
 * This function may be called before initializing.
 *
 * listener  = The TCP/IPv4 listener (return).
 * localaddr = The local address/port to bind to.
 * backlog   = The listen backlog.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_listener_open2 (rmix_tcpip4_listener_t   **listener,
                                const struct sockaddr_in  *localaddr,
                                unsigned int               backlog);

/*
 * Opens a TCP/IPv4 listener and binds it to the addressed local interface.
 * This function may be called before initializing.
 *
 * listener   = The TCP/IPv4 listener (return).
 * localname  = The local name to bind to.
 * localport  = The local port to bind to.
 * backlog    = The listen backlog.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_listener_open3 (rmix_tcpip4_listener_t **listener,
                                const char              *localname,
                                unsigned int             localport,
                                unsigned int             backlog);

/*
 * Opens a TCP/IPv4 listener and binds it to the addressed local interface.
 * This function may be called before initializing.
 *
 * listener = The TCP/IPv4 listener (return).
 * localurl = The local URL to bind to.
 * backlog  = The listen backlog.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_listener_open4 (rmix_tcpip4_listener_t **listener,
                                const char              *localurl,
                                unsigned int             backlog);

/*
 * Closes a TCP/IPv4 listener.
 * This function may be called before initializing.
 *
 * listener = The TCP/IPv4 listener.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_listener_close (rmix_tcpip4_listener_t **listener);

/*
 * TCP/IPv4 listener cleanup.
 * This function may be called before initializing.
 *
 * listener = The TCP/IPv4 listener.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
void rmix_tcpip4_listener_close2 (rmix_tcpip4_listener_t *listener);

/*
 * Opens a TCP/IPv4 server by accepting a connection to a TCP/IPv4 listener.
 * This function may be called before initializing.
 *
 * server   = The TCP/IPv4 server (return).
 * listener = The TCP/IPv4 listener.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_sever_open (rmix_tcpip4_server_t   **server,
                            rmix_tcpip4_listener_t  *listener);

/*
 * Closes a TCP/IPv4 server previously accepted by a TCP/IPv4 listener.
 * This function may be called before initializing.
 *
 * server = The TCP/IPv4 server.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_server_close (rmix_tcpip4_server_t **server);

/*
 * TCP/IPv4 server cleanup.
 * This function may be called before initializing.
 *
 * server = The TCP/IPv4 server.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
void rmix_tcpip4_server_close2 (void *server);

/*
 * Reads from a TCP/IPv4 server.
 * This function may be called before initializing.
 *
 * server = The TCP/IPv4 server.
 * size   = The number of bytes to read.
 * data   = The byte buffer to read into.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          ECONNABORTED = Connection closed before data transfer.
 *          EPIPE        = Connection closed during data transfer.
 */
int rmix_tcpip4_server_read (rmix_tcpip4_server_t *server,
                             unsigned int          size,
                             void                 *data);

/*
 * Reads from a TCP/IPv4 server (stream read function).
 * This function may be called before initializing.
 *
 * incoming = The TCP/IPv4 server.
 * size     = The number of bytes to read.
 * data     = The byte buffer to read into.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            ECONNABORTED = Connection closed before data transfer.
 *            EPIPE        = Connection closed during data transfer.
 */
RMIX_STREAM_READ(rmix_tcpip4_server_istream_read);

/*
 * Writes to a TCP/IPv4 server.
 * This function may be called before initializing.
 *
 * server = The TCP/IPv4 server.
 * size   = The number of bytes to write.
 * data   = The byte buffer to write from.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          ECONNABORTED = Connection closed before data transfer.
 *          EPIPE        = Connection closed during data transfer.
 */
int rmix_tcpip4_server_write (rmix_tcpip4_server_t *server,
                              unsigned int          size,
                              const void           *data);

/*
 * Writes to a TCP/IPv4 server (stream write function).
 * This function may be called before initializing.
 *
 * outgoing = The TCP/IPv4 server.
 * size     = The number of bytes to write.
 * data     = The byte buffer to write from.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            ECONNABORTED = Connection closed before data transfer.
 *            EPIPE        = Connection closed during data transfer.
 */
RMIX_STREAM_WRITE(rmix_tcpip4_server_ostream_write);

/*
 * Closes a TCP/IPv4 servers incoming stream.
 * This function may be called before initializing.
 *
 * stream = The TCP/IPv4 server.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
RMIX_STREAM_CLOSE(rmix_tcpip4_server_istream_close);

/*
 * Closes a TCP/IPv4 servers outgoing stream.
 * This function may be called before initializing.
 *
 * stream = The TCP/IPv4 server.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
RMIX_STREAM_CLOSE(rmix_tcpip4_server_ostream_close);

/*
 * Opens a TCP/IPv4 client and connects to the addressed remote TCP/IPv4 server.
 * This function may be called before initializing.
 *
 * client     = The TCP/IPv4 client (return).
 * remoteaddr = The remote address/port to connect to.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_client_open (rmix_tcpip4_client_t     **client,
                             const struct sockaddr_in  *remoteaddr);

/*
 * Opens a TCP/IPv4 client and connects to the addressed remote TCP/IPv4 server.
 * This function may be called before initializing.
 *
 * client     = The TCP/IPv4 client (return).
 * remotename = The remote name to connect to.
 * remoteport = The remote port to connect to.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_client_open2 (rmix_tcpip4_client_t **client,
                              const char            *remotename,
                              unsigned int           remoteport);

/*
 * Opens a TCP/IPv4 client and connects to the addressed remote TCP/IPv4 server.
 * This function may be called before initializing.
 *
 * client    = The TCP/IPv4 client (return).
 * remoteurl = The remote URL to connect to.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_client_open3 (rmix_tcpip4_client_t **client,
                              const char            *remoteurl);

/*
 * Disconnects and closes a TCP/IPv4 client.
 * This function may be called before initializing.
 *
 * client = The TCP/IPv4 client.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_client_close (rmix_tcpip4_client_t **client);

/*
 * TCP/IPv4 client cleanup.
 * This function may be called before initializing.
 *
 * client = The TCP/IPv4 client.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
void rmix_tcpip4_client_close2 (void *client);

/*
 * Reads from a TCP/IPv4 client.
 * This function may be called before initializing.
 *
 * client = The TCP/IPv4 client.
 * size   = The number of bytes to read.
 * data   = The byte buffer to read into.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          ECONNABORTED = Connection closed before data transfer.
 *          EPIPE        = Connection closed during data transfer.
 */
int rmix_tcpip4_client_read (rmix_tcpip4_client_t *client,
                             unsigned int          size,
                             void                 *data);

/*
 * Reads from a TCP/IPv4 client (stream read function).
 * This function may be called before initializing.
 *
 * incoming = The TCP/IPv4 client.
 * size     = The number of bytes to read.
 * data     = The byte buffer to read into.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            ECONNABORTED = Connection closed before data transfer.
 *            EPIPE        = Connection closed during data transfer.
 */
RMIX_STREAM_READ(rmix_tcpip4_client_istream_read);

/*
 * Writes to a TCP/IPv4 client.
 * This function may be called before initializing.
 *
 * client = The TCP/IPv4 client.
 * size   = The number of bytes to write.
 * data   = The byte buffer to write from.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          ECONNABORTED = Connection closed before data transfer.
 *          EPIPE        = Connection closed during data transfer.
 */
int rmix_tcpip4_client_write (rmix_tcpip4_client_t *client,
                              unsigned int          size,
                              const void           *data);

/*
 * Writes to a TCP/IPv4 client (stream write function).
 * This function may be called before initializing.
 *
 * outgoing = The TCP/IPv4 client.
 * size     = The number of bytes to write.
 * data     = The byte buffer to write from.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            ECONNABORTED = Connection closed before data transfer.
 *            EPIPE        = Connection closed during data transfer.
 */
RMIX_STREAM_WRITE(rmix_tcpip4_client_ostream_write);

/*
 * Closes a TCP/IPv4 clients incoming stream.
 * This function may be called before initializing.
 *
 * stream = The TCP/IPv4 client.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
RMIX_STREAM_CLOSE(rmix_tcpip4_client_istream_close);

/*
 * Closes a TCP/IPv4 clients outgoing stream.
 * This function may be called before initializing.
 *
 * stream = The TCP/IPv4 client.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
RMIX_STREAM_CLOSE(rmix_tcpip4_client_ostream_close);

/*
 * Opens a TCP/IPv4 server and binds it to all local interfaces.
 *
 * handle    = The TCP/IPv4 server handle (return).
 * port      = The port to bind to.
 * backlog   = The listen backlog.
 * serverjob = The server job.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_open (unsigned int            *handle,
                      unsigned int             localport,
                      unsigned int             backlog,
                      rmix_tcpip4_serverjob_t  serverjob);

/*
 * Opens a TCP/IPv4 server and binds it to the addressed local interface.
 *
 * handle    = The TCP/IPv4 server handle (return).
 * localaddr = The local address/port to bind to.
 * backlog   = The listen backlog.
 * serverjob = The server job.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_open2 (unsigned int             *handle,
                       const struct sockaddr_in *localaddr,
                       unsigned int              backlog,
                       rmix_tcpip4_serverjob_t   serverjob);

/*
 * Opens a TCP/IPv4 server and binds it to the addressed local interface.
 *
 * handle    = The TCP/IPv4 server handle (return).
 * localname = The local name to bind to.
 * localport = The local port to bind to.
 * backlog   = The listen backlog.
 * serverjob = The server job.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_open3 (unsigned int            *handle,
                       char                    *localname,
                       unsigned int             localport,
                       unsigned int             backlog,
                       rmix_tcpip4_serverjob_t  serverjob);

/*
 * Opens a TCP/IPv4 server and binds it to the addressed local interface.
 *
 * handle    = The TCP/IPv4 server handle (return).
 * localurl  = The local URL to bind to.
 * backlog   = The listen backlog.
 * serverjob = The server job.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_open4 (unsigned int            *handle,
                       char                    *localurl,
                       unsigned int             backlog,
                       rmix_tcpip4_serverjob_t  serverjob);

/*
 * Closes a TCP/IPv4 server.
 *
 * handle = The TCP/IPv4 server handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_tcpip4_close (unsigned int handle);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

#endif /* RMIX_TCPIP4_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
