/******************************************************************************
 *
 * Source file for rmixregistry (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

#define RMIXREGISTRY_OPTIONS_VERSION RMIX_STRING(\
rmixregistry (RMIX registry client) %s-%s.%s\n\
Copyright (c) 2002-2006 Oak Ridge National Laboratory.\n\
See COPYING in the source distribution top-level directory or package\n\
data directory (%s) for license information.\n\
Report bugs to %s.\n)

#define RMIXREGISTRY_OPTIONS_HELP "\
Usage:\n\
  rmixregistry[.debug] option\n\
Options:\n\
  -h|--help                           Print out command line help.\n\
  -v|--version                        Print out version information.\n\
  -b|--bind   [REGISTRY] OBJECT NAME  Bind OBJECT to NAME at REGISTRY.\n\
  -b|--rebind [REGISTRY] OBJECT NAME  Rebind OBJECT to NAME at REGISTRY.\n\
  -u|--unbind [REGISTRY] NAME         Unbind an object from NAME at REGISTRY.\n\
  -L|--lookup [REGISTRY] NAME         Lookup an object with NAME at REGISTRY.\n\
  -l|--list [REGISTRY]                List all names in REGISTRY.\n\
Option arguments:\n\
  REGISTRY and OBJECT are descriptors for remote object references in the\n\
  form: 'PROTOCOL=<potocol> OBJECTID=<object id>' with optional name/value\n\
  pairs for 'HOST=<IP address or name>', 'PORT=<IP port>', 'BINDING=<protocol\n\
  specific binding>' and 'INTERFACES=<interfaces>'. The default for REGISTRY\n\
  is 'PROTOCOL=RPC OBJECTID=42'. Examples are:\n\
    -  'PROTOCOL=RPC OBJECTID=21'\n\
    -  'PROTOCOL=RPC OBJECTID=21 HOST=192.168.1.1'\n\
    -  'PROTOCOL=RPC OBJECTID=21 HOST=192.168.1.1 PORT=5555'\n\
    -  'PROTOCOL=RPC OBJECTID=21 HOST=192.168.1.1 PORT=5555 BINDING=X'\n\
    -  'PROTOCOL=RPC OBJECTID=21 INTERFACES=MyClassA;MyClassB'\n\
    -  'HOST=192.168.1.1 PORT=5555'\n\
    -  'HOST=192.168.1.1 PORT=5555 BINDING=X'\n\
"

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include <rmixregistry.h>

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Parses the command line options and looks up a remote object reference at a
 * RMIX registry.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = EXIT_SUCCESS on success or EXIT_FAILURE on any error.
 */
int main (int    argc,
          char **argv,
          char **envp);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Parses the command line options and looks up a remote object reference at a
 * RMIX registry.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = EXIT_SUCCESS on success or EXIT_FAILURE on any error.
 */
#undef  __FUNC__
#define __FUNC__ "main"
int main (int    argc,
          char **argv,
          char **envp) {
  char              *rparametersstr, *oparametersstr, *name, **names;
  unsigned int       rprovider, oprovider, count, index;
  rmix_remoteref_t  *registry, *object;
  rmix_parameters_t *rparameters;
  /* Check for missing arguments. */
  if (1 == argc) {
      fprintf(stderr, "rmixregistry: missing option\n");
      fprintf(stderr, "rmixregistry: Try `%s --help` for more information.\n",
              argv[0]);
      return EXIT_FAILURE;
  /* Check for -h|--help option. */
  } else if ((0 == strcmp("-h",     argv[1]))||
             (0 == strcmp("--help", argv[1]))) {
    fprintf(stdout, RMIXREGISTRY_OPTIONS_HELP);
    return EXIT_SUCCESS;
  /* Check for -v|--version option. */
  } else if ((0 == strcmp("-v",        argv[1]))||
             (0 == strcmp("--version", argv[1]))) {
    fprintf(stdout, RMIXREGISTRY_OPTIONS_VERSION,
            RMIX_STRING(RMIXREGISTRY_PACKAGE_NAME),
            RMIX_STRING(RMIXREGISTRY_PACKAGE_VERSION),
            RMIX_STRING(RMIXREGISTRY_PACKAGE_RELEASE),
            RMIX_STRING(RMIXREGISTRY_PACKAGE_DATADIR),
            RMIX_STRING(RMIXREGISTRY_PACKAGE_BUGREPORT));
    return EXIT_SUCCESS;
  /* Check for -b|--bind option. */
  } else if ((0 == strcmp("-b",     argv[1]))||
             (0 == strcmp("--bind", argv[1]))) {
    /* Check for missing option arguments. */
    if (4 > argc) {
      fprintf(stderr, "rmixregistry: missing option arguments\n");
      fprintf(stderr, "rmixregistry: Try `%s --help` for more information.\n",
              argv[0]);
      return EXIT_FAILURE;
    }
    /* Check for optional option arguments. */
    if (5 > argc) {
      rparametersstr = RMIXREGISTRY_EXPORT_RPC;
      oparametersstr = argv[2];
      name           = argv[3];
    } else {
      rparametersstr = argv[2];
      oparametersstr = argv[3];
      name           = argv[4];
    }
    /* Check for too many option arguments. */
    if (5 < argc) {
      fprintf(stderr, "rmixregistry: too many option arguments\n");
      fprintf(stderr, "rmixregistry: Try `%s --help` for more information.\n",
              argv[0]);
      return EXIT_FAILURE;
    }
    /* Initialize rmix. */
    if (0 != rmix_init(RMIX_THREADS_POOL_MINIMUM, RMIX_THREADS_POOL_MAXIMUM, 0,
                       RMIX_THREADS_JOBS_MAXIMUM, RMIX_THREADS_JOBS_TIMEOUT)) {
      fprintf(stderr, "rmixregistry: unable to initialize rmix\n");
      return EXIT_FAILURE;
    }
    /* Create remote registry parameters. */
    if (0 != rmix_parameters_create4(&rparameters, rparametersstr)) {
      fprintf(stderr,
              "rmixregistry: unable to create remote registry parameters\n");
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Insert default protocol parameter. */
    if (0 != rmix_parameters_insert2(rparameters, "PROTOCOL",
               RMIX_STRING(RMIXREGISTRY_EXPORT_RPC_PROTOCOL))) {
      fprintf(stderr,
              "rmixregistry: unable to insert default protocol parameter\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Insert default object id parameter. */
    if (0 != rmix_parameters_insert2(rparameters, "OBJECTID",
               RMIX_STRING(RMIXREGISTRY_EXPORT_RPC_OBJECTID))) {
      fprintf(stderr,
              "rmixregistry: unable to insert default object id parameter\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Create remote registry reference. */
    if (0 != rmix_remoteref_create5_p2(&registry, &rprovider, rparameters)) {
      fprintf(stderr,
              "rmixregistry: unable to create remote registry reference\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Destroy remote registry parameters. */
    if (0 != rmix_parameters_destroy(&rparameters)) {
      fprintf(stderr,
              "rmixregistry: unable to destroy remote registry parameters\n");
      RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      /* Destroy remote registry reference. */
      if (0 != rmix_remoteref_destroy(&registry)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
      }
      /* Unload registry provider. */
      if (0 != rmix_providers_unload(rprovider)) {
        RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Create remote object reference. */
    if (0 != rmix_remoteref_create6_p2(&object, &oprovider, oparametersstr)) {
      fprintf(stderr,
              "rmixregistry: unable to create remote object reference\n");
      /* Destroy remote registry reference. */
      if (0 != rmix_remoteref_destroy(&registry)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
      }
      /* Unload registry provider. */
      if (0 != rmix_providers_unload(rprovider)) {
        RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Bind object. */
    if (0 != rmixregistryclient_bind_p(registry, rprovider, name, object)) {
      fprintf(stderr, "rmixregistry: unable to bind object\n");
      /* Destroy remote object reference. */
      if (0 != rmix_remoteref_destroy(&object)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote object reference)))
      }
      /* Unload object provider. */
      if (0 != rmix_providers_unload(oprovider)) {
        RMIX_LOG((RMIX_WARN(unable to unload object provider.)))
      }
      /* Destroy remote registry reference. */
      if (0 != rmix_remoteref_destroy(&registry)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
      }
      /* Unload registry provider. */
      if (0 != rmix_providers_unload(rprovider)) {
        RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Destroy remote object reference. */
    if (0 != rmix_remoteref_destroy(&object)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference)))
    }
    /* Unload object provider. */
    if (0 != rmix_providers_unload(oprovider)) {
      RMIX_LOG((RMIX_WARN(unable to unload object provider.)))
    }
    /* Destroy remote registry reference. */
    if (0 != rmix_remoteref_destroy(&registry)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
    }
    /* Unload registry provider. */
    if (0 != rmix_providers_unload(rprovider)) {
      RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
    }
    /* Finalize rmix. */
    if (0 != rmix_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
    }
    return EXIT_SUCCESS;
  /* Check for -b|--bind option. */
  } else if ((0 == strcmp("-r",       argv[1]))||
             (0 == strcmp("--rebind", argv[1]))) {
    /* Check for missing option arguments. */
    if (4 > argc) {
      fprintf(stderr, "rmixregistry: missing option arguments\n");
      fprintf(stderr, "rmixregistry: Try `%s --help` for more information.\n",
              argv[0]);
      return EXIT_FAILURE;
    }
    /* Check for optional option arguments. */
    if (5 > argc) {
      rparametersstr = RMIXREGISTRY_EXPORT_RPC;
      oparametersstr = argv[2];
      name           = argv[3];
    } else {
      rparametersstr = argv[2];
      oparametersstr = argv[3];
      name           = argv[4];
    }
    /* Check for too many option arguments. */
    if (5 < argc) {
      fprintf(stderr, "rmixregistry: too many option arguments\n");
      fprintf(stderr, "rmixregistry: Try `%s --help` for more information.\n",
              argv[0]);
      return EXIT_FAILURE;
    }
    /* Initialize rmix. */
    if (0 != rmix_init(RMIX_THREADS_POOL_MINIMUM, RMIX_THREADS_POOL_MAXIMUM, 0,
                       RMIX_THREADS_JOBS_MAXIMUM, RMIX_THREADS_JOBS_TIMEOUT)) {
      fprintf(stderr, "rmixregistry: unable to initialize rmix\n");
      return EXIT_FAILURE;
    }
    /* Create remote registry parameters. */
    if (0 != rmix_parameters_create4(&rparameters, rparametersstr)) {
      fprintf(stderr,
              "rmixregistry: unable to create remote registry parameters\n");
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Insert default protocol parameter. */
    if (0 != rmix_parameters_insert2(rparameters, "PROTOCOL",
               RMIX_STRING(RMIXREGISTRY_EXPORT_RPC_PROTOCOL))) {
      fprintf(stderr,
              "rmixregistry: unable to insert default protocol parameter\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Insert default object id parameter. */
    if (0 != rmix_parameters_insert2(rparameters, "OBJECTID",
               RMIX_STRING(RMIXREGISTRY_EXPORT_RPC_OBJECTID))) {
      fprintf(stderr,
              "rmixregistry: unable to insert default object id parameter\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Create remote registry reference. */
    if (0 != rmix_remoteref_create5_p2(&registry, &rprovider, rparameters)) {
      fprintf(stderr,
              "rmixregistry: unable to create remote registry reference\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Create remote object reference. */
    if (0 != rmix_remoteref_create6_p2(&object, &oprovider, oparametersstr)) {
      fprintf(stderr,
              "rmixregistry: unable to create remote object reference\n");
      /* Destroy remote registry reference. */
      if (0 != rmix_remoteref_destroy(&registry)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
      }
      /* Unload registry provider. */
      if (0 != rmix_providers_unload(rprovider)) {
        RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Rebind object. */
    if (0 != rmixregistryclient_rebind_p(registry, rprovider, name, object)) {
      fprintf(stderr, "rmixregistry: unable to rebind object\n");
      /* Destroy remote object reference. */
      if (0 != rmix_remoteref_destroy(&object)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote object reference)))
      }
      /* Unload object provider. */
      if (0 != rmix_providers_unload(oprovider)) {
        RMIX_LOG((RMIX_WARN(unable to unload object provider.)))
      }
      /* Destroy remote registry reference. */
      if (0 != rmix_remoteref_destroy(&registry)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
      }
      /* Unload registry provider. */
      if (0 != rmix_providers_unload(rprovider)) {
        RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Destroy remote object reference. */
    if (0 != rmix_remoteref_destroy(&object)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference)))
    }
    /* Unload object provider. */
    if (0 != rmix_providers_unload(oprovider)) {
      RMIX_LOG((RMIX_WARN(unable to unload object provider.)))
    }
    /* Destroy remote registry reference. */
    if (0 != rmix_remoteref_destroy(&registry)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
    }
    /* Unload registry provider. */
    if (0 != rmix_providers_unload(rprovider)) {
      RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
    }
    /* Finalize rmix. */
    if (0 != rmix_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
    }
    return EXIT_SUCCESS;
  /* Check for -u|--unbind option. */
  } else if ((0 == strcmp("-u",       argv[1]))||
             (0 == strcmp("--unbind", argv[1]))) {
    /* Check for missing option arguments. */
    if (3 > argc) {
      fprintf(stderr, "rmixregistry: missing option arguments\n");
      fprintf(stderr, "rmixregistry: Try `%s --help` for more information.\n",
              argv[0]);
      return EXIT_FAILURE;
    }
    /* Check for optional option arguments. */
    if (4 > argc) {
      rparametersstr = RMIXREGISTRY_EXPORT_RPC;
      name           = argv[2];
    } else {
      rparametersstr = argv[2];
      name           = argv[3];
    }
    /* Check for too many option arguments. */
    if (4 < argc) {
      fprintf(stderr, "rmixregistry: too many option arguments\n");
      fprintf(stderr, "rmixregistry: Try `%s --help` for more information.\n",
              argv[0]);
      return EXIT_FAILURE;
    }
    /* Initialize rmix. */
    if (0 != rmix_init(RMIX_THREADS_POOL_MINIMUM, RMIX_THREADS_POOL_MAXIMUM, 0,
                       RMIX_THREADS_JOBS_MAXIMUM, RMIX_THREADS_JOBS_TIMEOUT)) {
      fprintf(stderr, "rmixregistry: unable to initialize rmix\n");
      return EXIT_FAILURE;
    }
    /* Create remote registry parameters. */
    if (0 != rmix_parameters_create4(&rparameters, rparametersstr)) {
      fprintf(stderr,
              "rmixregistry: unable to create remote registry parameters\n");
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Insert default protocol parameter. */
    if (0 != rmix_parameters_insert2(rparameters, "PROTOCOL",
               RMIX_STRING(RMIXREGISTRY_EXPORT_RPC_PROTOCOL))) {
      fprintf(stderr,
              "rmixregistry: unable to insert default protocol parameter\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Insert default object id parameter. */
    if (0 != rmix_parameters_insert2(rparameters, "OBJECTID",
               RMIX_STRING(RMIXREGISTRY_EXPORT_RPC_OBJECTID))) {
      fprintf(stderr,
              "rmixregistry: unable to insert default object id parameter\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Create remote registry reference. */
    if (0 != rmix_remoteref_create5_p2(&registry, &rprovider, rparameters)) {
      fprintf(stderr,
              "rmixregistry: unable to create remote registry reference\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Unbind object. */
    if (0 != rmixregistryclient_unbind_p(registry, rprovider, name)) {
      fprintf(stderr, "rmixregistry: unable to unbind object\n");
      /* Destroy remote registry reference. */
      if (0 != rmix_remoteref_destroy(&registry)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
      }
      /* Unload registry provider. */
      if (0 != rmix_providers_unload(rprovider)) {
        RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Destroy remote registry reference. */
    if (0 != rmix_remoteref_destroy(&registry)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
    }
    /* Unload registry provider. */
    if (0 != rmix_providers_unload(rprovider)) {
      RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
    }
    /* Finalize rmix. */
    if (0 != rmix_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
    }
    return EXIT_SUCCESS;
  /* Check for -L|--lookup option. */
  } else if ((0 == strcmp("-L",       argv[1]))||
             (0 == strcmp("--lookup", argv[1]))) {
    /* Check for missing option arguments. */
    if (3 > argc) {
      fprintf(stderr, "rmixregistry: missing option arguments\n");
      fprintf(stderr, "rmixregistry: Try `%s --help` for more information.\n",
              argv[0]);
      return EXIT_FAILURE;
    }
    /* Check for optional option arguments. */
    if (4 > argc) {
      rparametersstr = RMIXREGISTRY_EXPORT_RPC;
      name           = argv[2];
    } else {
      rparametersstr = argv[2];
      name           = argv[3];
    }
    /* Check for too many option arguments. */
    if (4 < argc) {
      fprintf(stderr, "rmixregistry: too many option arguments\n");
      fprintf(stderr, "rmixregistry: Try `%s --help` for more information.\n",
              argv[0]);
      return EXIT_FAILURE;
    }
    /* Initialize rmix. */
    if (0 != rmix_init(RMIX_THREADS_POOL_MINIMUM, RMIX_THREADS_POOL_MAXIMUM, 0,
                       RMIX_THREADS_JOBS_MAXIMUM, RMIX_THREADS_JOBS_TIMEOUT)) {
      fprintf(stderr, "rmixregistry: unable to initialize rmix\n");
      return EXIT_FAILURE;
    }
    /* Create remote registry parameters. */
    if (0 != rmix_parameters_create4(&rparameters, rparametersstr)) {
      fprintf(stderr,
              "rmixregistry: unable to create remote registry parameters\n");
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Insert default protocol parameter. */
    if (0 != rmix_parameters_insert2(rparameters, "PROTOCOL",
               RMIX_STRING(RMIXREGISTRY_EXPORT_RPC_PROTOCOL))) {
      fprintf(stderr,
              "rmixregistry: unable to insert default protocol parameter\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Insert default object id parameter. */
    if (0 != rmix_parameters_insert2(rparameters, "OBJECTID",
               RMIX_STRING(RMIXREGISTRY_EXPORT_RPC_OBJECTID))) {
      fprintf(stderr,
              "rmixregistry: unable to insert default object id parameter\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Create remote registry reference. */
    if (0 != rmix_remoteref_create5_p2(&registry, &rprovider, rparameters)) {
      fprintf(stderr,
              "rmixregistry: unable to create remote registry reference\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Lookup object. */
    if (0 != rmixregistryclient_lookup_p(registry, rprovider, name, &object)) {
      fprintf(stderr, "rmixregistry: unable to lookup object\n");
      /* Destroy remote registry reference. */
      if (0 != rmix_remoteref_destroy(&registry)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
      }
      /* Unload registry provider. */
      if (0 != rmix_providers_unload(rprovider)) {
        RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Convert remote object reference. */
    if (0 != rmix_remoteref_parameters2(&oparametersstr, object)) {
      fprintf(stderr,
              "rmixregistry: unable to convert remote object reference\n");
      /* Destroy remote object reference. */
      if (0 != rmix_remoteref_destroy(&object)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote object reference)))
      }
      /* Destroy remote registry reference. */
      if (0 != rmix_remoteref_destroy(&registry)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
      }
      /* Unload registry provider. */
      if (0 != rmix_providers_unload(rprovider)) {
        RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Print out parameters. */
    fprintf(stdout, "%s\n", oparametersstr);
    /* Deallocate parameters. */
    free(oparametersstr);
    /* Destroy remote object reference. */
    if (0 != rmix_remoteref_destroy(&object)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference)))
    }
    /* Destroy remote registry reference. */
    if (0 != rmix_remoteref_destroy(&registry)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
    }
    /* Unload registry provider. */
    if (0 != rmix_providers_unload(rprovider)) {
      RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
    }
    /* Finalize rmix. */
    if (0 != rmix_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
    }
    return EXIT_SUCCESS;
  /* Check for -l|--list option. */
  } else if ((0 == strcmp("-l",     argv[1]))||
             (0 == strcmp("--list", argv[1]))) {
    /* Check for optional option arguments. */
    if (3 > argc) {
      rparametersstr = RMIXREGISTRY_EXPORT_RPC;
    } else {
      rparametersstr = argv[2];
    }
    /* Check for too many option arguments. */
    if (3 < argc) {
      fprintf(stderr, "rmixregistry: too many option arguments\n");
      fprintf(stderr, "rmixregistry: Try `%s --help` for more information.\n",
              argv[0]);
      return EXIT_FAILURE;
    }
    /* Initialize rmix. */
    if (0 != rmix_init(RMIX_THREADS_POOL_MINIMUM, RMIX_THREADS_POOL_MAXIMUM, 0,
                       RMIX_THREADS_JOBS_MAXIMUM, RMIX_THREADS_JOBS_TIMEOUT)) {
      fprintf(stderr, "rmixregistry: unable to initialize rmix\n");
      return EXIT_FAILURE;
    }
    /* Create remote registry parameters. */
    if (0 != rmix_parameters_create4(&rparameters, rparametersstr)) {
      fprintf(stderr,
              "rmixregistry: unable to create remote registry parameters\n");
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Insert default protocol parameter. */
    if (0 != rmix_parameters_insert2(rparameters, "PROTOCOL",
               RMIX_STRING(RMIXREGISTRY_EXPORT_RPC_PROTOCOL))) {
      fprintf(stderr,
              "rmixregistry: unable to insert default protocol parameter\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Insert default object id parameter. */
    if (0 != rmix_parameters_insert2(rparameters, "OBJECTID",
               RMIX_STRING(RMIXREGISTRY_EXPORT_RPC_OBJECTID))) {
      fprintf(stderr,
              "rmixregistry: unable to insert default object id parameter\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Create remote registry reference. */
    if (0 != rmix_remoteref_create5_p2(&registry, &rprovider, rparameters)) {
      fprintf(stderr,
              "rmixregistry: unable to create remote registry reference\n");
      /* Destroy remote registry parameters. */
      if (0 != rmix_parameters_destroy(&rparameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry parameters)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* List names. */
    if (0 != rmixregistryclient_list_p(registry, rprovider, &count, &names)) {
      fprintf(stderr, "rmixregistry: unable to list names\n");
      /* Destroy remote registry reference. */
      if (0 != rmix_remoteref_destroy(&registry)) {
        RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
      }
      /* Unload registry provider. */
      if (0 != rmix_providers_unload(rprovider)) {
        RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
      }
      return EXIT_FAILURE;
    }
    /* Print out and deallocate names. */
    for (index = 0; index < count; index++) {
      /* Print out and deallocate name. */
      fprintf(stdout, "%s\n", names[index]);
      free(names[index]);
    }
    if (NULL != names) {
      free(names);
    }
    /* Destroy remote registry reference. */
    if (0 != rmix_remoteref_destroy(&registry)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote registry reference)))
    }
    /* Unload registry provider. */
    if (0 != rmix_providers_unload(rprovider)) {
      RMIX_LOG((RMIX_WARN(unable to unload registry provider.)))
    }
    /* Finalize rmix. */
    if (0 != rmix_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
    }
    return EXIT_SUCCESS;
  /* Check for unknown option. */
  } else {
      fprintf(stderr, "rmixregistry: unrecognized option '%s'\n", argv[1]);
      fprintf(stderr, "rmixregistry: Try `%s --help` for more information.\n",
              argv[0]);
    return EXIT_FAILURE;
  }
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
