/******************************************************************************
 *
 * Source file for rmix-rpcx (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmix-rpcx.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Rmix-rpcx plug-in module data type. */
typedef struct {
  pthread_mutex_t            mutex;       /* mutex                 */
  unsigned int               ready;       /* ready flag            */
  struct {
    unsigned int             count;       /* instance count        */
    unsigned int            *array;       /* instance array        */
  }                          instances;   /* instances             */
  struct {
    unsigned int             count;       /* server count          */
    struct {
      unsigned int           server;      /* server id             */
      struct sockaddr_in     address;     /* server address        */
      struct {
        unsigned int         count;       /* objectid count        */
        unsigned int        *array;       /* objectid array        */
      }                      exports;     /* exports               */
    }                       *array;       /* server array          */
  }                          servers;     /* servers               */
  struct {
    pthread_cond_t           ready;       /* invocation ready      */
    unsigned int             count;       /* invocation count      */
    struct {
      unsigned int           invocation;  /* invokation id         */
      rmix_tcpip4_client_t  *client;      /* TCP/IPv4 client       */
      char                  *sigstr;      /* output type signature */
      unsigned int           outcnt;      /* output values count   */
      void                 **outary;      /* output values array   */
      int                    error;       /* error number          */
    }                       *array;       /* invokation array      */
  }                          invocations; /* invocations           */
  struct {
    pthread_cond_t           ready;       /* reading ready         */
    unsigned int             count;       /* reading count         */
    struct {
      unsigned int           callid;      /* call id               */
      unsigned int           objectid;    /* object id             */
    }                       *array;       /* reading array         */
  }                          reading;     /* reading               */
  struct {
    pthread_cond_t           ready;       /* executing ready       */
    unsigned int             count;       /* executing count       */
    struct {
      unsigned int           callid;      /* call id               */
      unsigned int           objectid;    /* object id             */
    }                       *array;       /* executing array       */
  }                          executing;   /* executing             */
} rmix_rpcx_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Initializes the rmix-rpcx plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
RMIX_PLUGINS_INIT(rmix_rpcx_init);

/*
 * Finalizes the rmix-rpcx plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
RMIX_PLUGINS_FINI(rmix_rpcx_fini);

/*
 * Retrieves the output (if any) of a one-way invocation in a separate thread
 * using an invokation id previously issued by the provider.
 *
 * cleanup  = NULL.
 * argument = The TCP/IPv4 client.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
RMIX_THREADS_JOB(rmix_rpcx_oneway_retrieve_job);

/*
 * Retrieves the output (if any) of an asynchronous invocation in a separate
 * thread using an invokation id previously issued by the provider.
 *
 * cleanup  = NULL.
 * argument = The TCP/IPv4 client.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
RMIX_THREADS_JOB(rmix_rpcx_async_retrieve_job);

/*
 * Serves a connection.
 *
 * server = The TCP/IPv4 server.
 * return = The server job function pointer on success or
 *          NULL on any error with errno set appropriately.
 */
RMIX_TCPIP4_SERVERJOB(rmix_rpcx_server);

/*
 * Moves an incoming call into the reading state.
 *
 * callid   = The call id.
 * objectid = The object id.
 * return = 0 on success, 1 on closed connection, or -1 on
 *          any other error with errno set appropriately.
 */
int rmix_rpcx_reading (unsigned int callid,
                       unsigned int objectid);

/*
 * Moves an incoming call into the executing state.
 *
 * callid = The call id.
 * return = 0 on success, 1 on closed connection, or -1 on
 *          any other error with errno set appropriately.
 */
int rmix_rpcx_executing (unsigned int callid);

/*
 * Moves an incoming call into the executed state.
 *
 * callid = The call id.
 * return = 0 on success, 1 on closed connection, or -1 on
 *          any other error with errno set appropriately.
 */
int rmix_rpcx_executed (unsigned int callid);

/*
 * The rmix-rpcx plug-in mutex unlock cleanup.
 *
 * argument = NULL.
 */
void rmix_rpcx_unlock (void *argument);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Rmix-rpcx plug-in module data. */
rmix_rpcx_data_t rmix_rpcx_data = {
  PTHREAD_MUTEX_INITIALIZER,  /* mutex            */
  0,                          /* ready flag       */
  {
    0,                        /* instance count   */
    NULL                      /* instance array   */
  },                          /* instances        */
  {
    0,                        /* server count     */
    NULL                      /* server array     */
  },                          /* servers          */
  {
    PTHREAD_COND_INITIALIZER, /* invocation ready */
    0,                        /* invokation count */
    NULL                      /* invokation array */
  },                          /* invokations      */
  {
    PTHREAD_COND_INITIALIZER, /* reading ready    */
    0,                        /* reading count    */
    NULL                      /* reading array    */
  },                          /* reading          */
  {
    PTHREAD_COND_INITIALIZER, /* executing ready  */
    0,                        /* executing count  */
    NULL                      /* executing array  */
  }                           /* executing        */
};

/* Rmix-rpcx plug-in "object". */
rmix_rpcx_t rmix_rpcx = {
  {
    LIBRMIX_RPCX_VERSION_CURRENT,  /* current version         */
    LIBRMIX_RPCX_VERSION_REVISION, /* current revision        */
    LIBRMIX_RPCX_VERSION_AGE,      /* version age             */
    LIBRMIX_RPCX_VERSION_FIRST     /* first supported version */
  },                               /* plug-in version         */
  /**********************************************/
  /* PUT YOUR PUBLIC PLUG-IN DATA POINTERS HERE */
  /**********************************************/
  /**************************************************/
  /* PUT YOUR PUBLIC PLUG-IN FUNCTION POINTERS HERE */
  /**************************************************/
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes the rmix-rpcx plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_init"
RMIX_PLUGINS_INIT(rmix_rpcx_init) {
  int           error;
  void         *array;
  unsigned int  index;
  RMIX_LOG((RMIX_INFO(rmix-rpcx startup)))
  /* Lock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Search for handle in handle array. */
  for (index = 0; index < rmix_rpcx_data.instances.count; index++) {
    if (rmix_rpcx_data.instances.array[index] == handle) {
      errno = EALREADY;
      RMIX_LOG((RMIX_WARN(handle is already in instance array)))
      errno = EALREADY;
      rmix_syserr();
      /* Unlock rmix-rpcx plug-in mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
        errno = error;
        rmix_syserr();
      }
      errno = EALREADY;
      return -1;
    }
  }
  /* Increase instance count. */
  index = rmix_rpcx_data.instances.count;
  rmix_rpcx_data.instances.count++;
  /* Reallocate instance array. */
  if (NULL == (array = realloc(rmix_rpcx_data.instances.array,
                               rmix_rpcx_data.instances.count *
                        sizeof(rmix_rpcx_data.instances.array[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate instance array)))
    /* Decrease handle array. */
    rmix_rpcx_data.instances.count--;
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    errno2 = errno;
    return -1;
  }
  rmix_rpcx_data.instances.array = array;
  /* Save instance array entry. */
  rmix_rpcx_data.instances.array[index] = handle;
  /* Check for first initialization. */
  if (1 == rmix_rpcx_data.instances.count) {
    /* Set ready flag. */
    rmix_rpcx_data.ready = 1;
  }
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    /* Reset ready flag. */
    rmix_rpcx_data.ready = 0;
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Finalizes the rmix-rpcx plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_fini"
RMIX_PLUGINS_FINI(rmix_rpcx_fini) {
  int           error, error2, result, errno2;
  void         *array;
  unsigned int  index;
  result = 0;
  errno2 = 0;
  RMIX_LOG((RMIX_INFO(rmix-rpcx shutdown)))
  /* Lock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Search for handle in instance array. */
  for (index = 0; index < rmix_rpcx_data.instances.count; index++) {
    if (rmix_rpcx_data.instances.array[index] == handle) {
      break;
    }
  }
  /* Check if handle is not in instance array. */
  if (index == rmix_rpcx_data.instances.count) {
    errno = ESRCH;
    RMIX_LOG((RMIX_WARN(handle is not in instance array)))
    errno = ESRCH;
    rmix_syserr();
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Decrease instance count. */
  rmix_rpcx_data.instances.count--;
  memmove(rmix_rpcx_data.instances.array + index,
          rmix_rpcx_data.instances.array + index + 1,
         (rmix_rpcx_data.instances.count - index) *
   sizeof(rmix_rpcx_data.instances.array[0]));
  /* Reallocate instance array. */
  if (0 == rmix_rpcx_data.instances.count) {
    free(rmix_rpcx_data.instances.array);
    rmix_rpcx_data.instances.array = NULL;
  } else if (NULL == (array = realloc(rmix_rpcx_data.instances.array,
                                      rmix_rpcx_data.instances.count *
                               sizeof(rmix_rpcx_data.instances.array[0])))) {
    RMIX_LOG((RMIX_WARN(unable to reallocate instance array)))
    errno = 0;
  } else {
    rmix_rpcx_data.instances.array = array;
  }
  /* Check for last finalization. */
  if (0 == rmix_rpcx_data.instances.count) {
    /* Reset ready flag. */
    rmix_rpcx_data.ready = 0;
    /* Close all servers. */
    for (index = 0; index < rmix_rpcx_data.servers.count; index++) {
      /* Close server. */
      if (0 != rmix_tcpip4_close(rmix_rpcx_data.servers.array[index].server)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to close server)))
        result = -1;
      }
    }
    /* Wait for all reading calls.*/
    while (0 != rmix_rpcx_data.reading.count) {
      /* Push unlock rmix-rpcx plug-in mutex to pthread cleanup stack. */
      pthread_cleanup_push(rmix_rpcx_unlock, &rmix_rpcx_data.mutex);
      /* Wait for reading ready condition. */
      if (0 != (error = pthread_cond_wait(&rmix_rpcx_data.reading.ready,
                                          &rmix_rpcx_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to wait for reading ready condition)))
        errno = error;
        rmix_syserr();
        /* Unlock rmix-rpcx plug-in mutex. */
        if (0 != (error2 = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
          errno = error2;
          RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
          errno = error2;
          rmix_syserr();
        }
        if (-1 == result) {
          errno = errno2;
        } else {
          errno = error;
        }
        return -1;
      }
      /* Pop unlock rmix-rpcx plug-in mutex from pthread cleanup stack. */
      pthread_cleanup_pop(0);
    }
    /* Wait for all executing calls.*/
    while (0 != rmix_rpcx_data.executing.count) {
      /* Push unlock rmix-rpcx plug-in mutex to pthread cleanup stack. */
      pthread_cleanup_push(rmix_rpcx_unlock, &rmix_rpcx_data.mutex);
      /* Wait for executing ready condition. */
      if (0 != (error = pthread_cond_wait(&rmix_rpcx_data.executing.ready,
                                          &rmix_rpcx_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to wait for executing ready condition)))
        errno = error;
        rmix_syserr();
        /* Unlock rmix-rpcx plug-in mutex. */
        if (0 != (error2 = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
          errno = error2;
          RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
          errno = error2;
          rmix_syserr();
        }
        if (-1 == result) {
          errno = errno2;
        } else {
          errno = error;
        }
        return -1;
      }
      /* Pop unlock rmix-rpcx plug-in mutex from pthread cleanup stack. */
      pthread_cleanup_pop(0);
    }
    /* Unexport all exports. */
    for (index = 0; index < rmix_rpcx_data.servers.count; index++) {
      /* Deallocate export array. */
      if (NULL != rmix_rpcx_data.servers.array[index].exports.array) {
        free(rmix_rpcx_data.servers.array[index].exports.array);
      }
    }
    /* Deallocate server array. */
    if (NULL != rmix_rpcx_data.servers.array) {
      free(rmix_rpcx_data.servers.array);
      rmix_rpcx_data.servers.array = NULL;
    }
    rmix_rpcx_data.servers.count = 0;
    /* Close all invocations. */
    for (index = 0; index < rmix_rpcx_data.invocations.count; index++) {
      /* Close client. */
      if (NULL != rmix_rpcx_data.invocations.array[index].client) {
        if (0 != rmix_tcpip4_client_close(&rmix_rpcx_data.invocations.
                                           array[index].client)) {
          errno2 = (0 == errno2)?errno:errno2;
          RMIX_LOG((RMIX_WARN(unable to close client)))
          result = -1;
        }
      }
      /* Destroy output array. */
      if (0 != rmix_valary_destroy(&rmix_rpcx_data.invocations.
                                    array[index].outary,
                                   &rmix_rpcx_data.invocations.
                                    array[index].outcnt,
                                   rmix_rpcx_data.invocations.
                                    array[index].sigstr)) {
        errno2 = (0 == errno2)?errno:errno2;
        RMIX_LOG((RMIX_WARN(unable to destroy output array)))
        result = -1;
      }
      /* Deallocate output array type signature duplicate. */
      free(rmix_rpcx_data.invocations.array[index].sigstr);
    }
    /* Deallocate invocations array. */
    if (NULL != rmix_rpcx_data.invocations.array) {
      free(rmix_rpcx_data.invocations.array);
      rmix_rpcx_data.invocations.array = NULL;
    }
    rmix_rpcx_data.invocations.count = 0;
  }
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    if (-1 == result) {
      errno = errno2;
    } else {
      errno = error;
    }
    return -1;
  }
  if (-1 == result) {
    errno = errno2;
  }
  return result;
}

/*
 * Invokes a method at a remote object.
 *
 * remoteref = The remote object reference.
 * interface = The remote object interface.
 * method    = The index of the method to invoke.
 * outary    = The output values array.
 * outcnt    = The output values count.
 * inary     = The input values array.
 * incnt     = The input values count.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 *             ECONNABORTED = Connection closed before data transfer.
 *             EPIPE        = Connection closed during data transfer.
 *             ESRCH        = Remote prognum unavailable.
 *             ENOTSUP      = Remote progver mismatch.
 *             EOPNOTSUPP   = Remote procnum unavailable.
 *             EBADMSG      = Remote garbled args.
 *             ENOSYS       = Remote system error.
 *             ENOPROTOOPT  = Remote RPC version mismatch.
 *             EACCES       = Remote authentication error.
 *             EPROTO       = Protocol error.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_invoke"
RMIX_PROVIDERS_INVOKE(rmix_rpcx_invoke) {
  int                          low, high;
  char                        *crebody, *verbody, *verbody2;
  unsigned int                 prognum, cresize,  versize,  versize2, callid,
                               objectid;
  rmix_istream_t               istream;
  rmix_ostream_t               ostream;
  rmix_tcpip4_client_t        *client;
  rmix_rpcx_stream_authflav_t  creflav,  verflav,  verflav2;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Get thread id and set it as call id. */
  if (0 != rmix_threadids_get(&callid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get thread id and to set it as call id)))
    errno = errno2;
    return -1;
  }
  /* Get object id from remote reference. */
  if (0 != rmix_remoteref_objectid(remoteref, &objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get object id from remote reference)))
    errno = errno2;
    return -1;
  }
  /* Compute program number. */
  prognum = objectid | 0x20000000;
  /* Create authentication credentials/verifier. */
  if (0 != rmix_rpcx_auth_create(&creflav, &cresize, &crebody,
                                 &verflav, &versize, &verbody)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create authentication credentials/verifier)))
    errno = errno2;
    return -1;
  }
  /* Open TCP/IPv4 client. */
  if (0 != rmix_tcpip4_client_open3(&client, remoteref->location)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open TCP/IPv4 client)))
    errno = errno2;
    /* Deallocate client authentication credentials. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate client authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    return -1;
  }
  /* Prepare output stream. */
  ostream.outgoing = client;
  ostream.write    = rmix_tcpip4_client_ostream_write;
  ostream.close    = rmix_tcpip4_client_ostream_close;
  /* Push close TCP/IPv4 client to pthread cleanup stack. */
  pthread_cleanup_push(rmix_tcpip4_client_close2, client);
  /* Push deallocate authentication credential to pthread cleanup stack. */
  pthread_cleanup_push(free, crebody);
  /* Push deallocate authentication verifier to pthread cleanup stack. */
  pthread_cleanup_push(free, verbody);
  /* Write call. */
  if (0 != rmix_rpcx_stream_write_call(&ostream, callid, prognum, 0, method,
             creflav, cresize, crebody, verflav, versize, verbody,
             interface->methods[method].input, inary, incnt)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to write call)))
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    /* Deallocate client authentication credentials. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate client authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    errno = errno2;
    return -1;
  }
  /* Prepare iutput stream. */
  istream.incoming = client;
  istream.read     = rmix_tcpip4_client_istream_read;
  istream.close    = rmix_tcpip4_client_istream_close;
  /* Read reply w/o parameters. */
  if (0 != rmix_rpcx_stream_read_reply(&istream, callid, &verflav2, &versize2,
                                       &verbody2, &low, &high)) {
    int errno2 = errno;
    /* Check error. */
    switch (errno2) {
      /* Check for RPC version mismatch. */
      case ENOPROTOOPT: {
        RMIX_LOG((
          RMIX_WARN(RPC version mismatch (%d <= version <= %d)), low, high))
      }
      /* Check for program version mismatch. */
      case ENOTSUP: {
        RMIX_LOG((
          RMIX_WARN(program version mismatch (%d <= version <= %d)), low, high))
      }
      /* Catch all other errors. */
      default: {
        RMIX_LOG((RMIX_WARN(unable to read reply w/o parameters)))
      }
    }
    /* Deallocate client authentication credentials. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate client authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    errno = errno2;
    return -1;
  }
  /* Pop deallocate authentication verifier from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate authentication credential from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Verify server. */
  if (0 != rmix_rpcx_auth_verify(creflav,  &cresize,  &crebody,
                                 verflav,  &versize,  &verbody,
                                 verflav2, &versize2, &verbody2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to verify server)))
    /* Deallocate client authentication verifier. */
    if (NULL != verbody2) {
      free(verbody2);
    }
    /* Deallocate client authentication credentials. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate client authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    errno = errno2;
    return -1;
  }
  /* Read procedure return. */
  if (0 != rmix_rpcx_stream_read_reply2(&istream,
             interface->methods[method].output, outary, outcnt)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to read procedure return)))
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    errno = errno2;
    return -1;
  }
  /* Pop close TCP/IPv4 client from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Close TCP/IPv4 client. */
  if (0 != rmix_tcpip4_client_close(&client)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Invokes a method at a remote object one-way, i.e. there is no return value
 * and invocation returns immediately after sending message.
 *
 * remoteref = The remote object reference.
 * interface = The remote object interface.
 * method    = The index of the method to invoke.
 * inary     = The input values array.
 * incnt     = The input values count.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 *             ECONNABORTED = Connection closed before data transfer.
 *             EPIPE        = Connection closed during data transfer.
 *             ESRCH        = Remote prognum unavailable.
 *             ENOTSUP      = Remote progver mismatch.
 *             EOPNOTSUPP   = Remote procnum unavailable.
 *             EBADMSG      = Remote garbled args.
 *             ENOSYS       = Remote system error.
 *             ENOPROTOOPT  = Remote RPC version mismatch.
 *             EACCES       = Remote authentication error.
 *             EPROTO       = Protocol error.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_oneway"
RMIX_PROVIDERS_ONEWAY(rmix_rpcx_oneway) {
  int                           error, low, high;
  char                         *sigstr, *crebody, *verbody, *verbody2;
  void                        **outary, *array, *argument;
  unsigned int                  prognum, outcnt, cresize,  versize,  versize2,
                                callid, objectid, index;
  rmix_istream_t                istream;
  rmix_ostream_t                ostream;
  rmix_tcpip4_client_t         *client;
  rmix_valary_cleanup_t         ocleanup;
  rmix_rpcx_stream_authflav_t   creflav,  verflav,  verflav2;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Get thread id and set it as call id. */
  if (0 != rmix_threadids_get(&callid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get thread id and to set it as call id)))
    errno = errno2;
    return -1;
  }
  /* Get object id from remote reference. */
  if (0 != rmix_remoteref_objectid(remoteref, &objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get object id from remote reference)))
    errno = errno2;
    return -1;
  }
  /* Compute program number. */
  prognum = objectid | 0x20000000;
  /* Duplicate output array type signature. */
  if (NULL == (sigstr = strdup(interface->methods[method].output))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate output array type signature)))
    errno = errno2;
    return -1;
  }
  /* Create output array. */
  if (0 != rmix_valary_create(&outary, &outcnt, sigstr)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create output array)))
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = errno2;
    return -1;
  }
  /* Create authentication credentials/verifier. */
  if (0 != rmix_rpcx_auth_create(&creflav, &cresize, &crebody,
                                 &verflav, &versize, &verbody)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create authentication credentials/verifier)))
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = errno2;
    return -1;
  }
  /* Open TCP/IPv4 client. */
  if (0 != rmix_tcpip4_client_open3(&client, remoteref->location)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open TCP/IPv4 client)))
    /* Deallocate client authentication credentials. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate client authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = errno2;
    return -1;
  }
  /* Prepare output stream. */
  ostream.outgoing = client;
  ostream.write    = rmix_tcpip4_client_ostream_write;
  ostream.close    = rmix_tcpip4_client_ostream_close;
  /* Push deallocate output array type signature to pthread cleanup stack. */
  pthread_cleanup_push(free, sigstr);
  /* Push deallocate call output values to pthread cleanup stack. */
  ocleanup.valary = outary;
  ocleanup.valcnt = outcnt;
  ocleanup.sigstr = sigstr;
  pthread_cleanup_push(rmix_valary_cleanup, &ocleanup);
  /* Push deallocate authentication credential to pthread cleanup stack. */
  pthread_cleanup_push(free, crebody);
  /* Push deallocate authentication verifier to pthread cleanup stack. */
  pthread_cleanup_push(free, verbody);
  /* Push close TCP/IPv4 client to pthread cleanup stack. */
  pthread_cleanup_push(rmix_tcpip4_client_close2, client);
  /* Write call. */
  if (0 != rmix_rpcx_stream_write_call(&ostream, callid, prognum, 0, method,
             creflav, cresize, crebody, verflav, versize, verbody,
             interface->methods[method].input, inary, incnt)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to write call)))
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    /* Deallocate client authentication credentials. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate client authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = errno2;
    return -1;
  }
  /* Prepare iutput stream. */
  istream.incoming = client;
  istream.read     = rmix_tcpip4_client_istream_read;
  istream.close    = rmix_tcpip4_client_istream_close;
  /* Read reply w/o parameters. */
  if (0 != rmix_rpcx_stream_read_reply(&istream, callid, &verflav2, &versize2,
                                       &verbody2, &low, &high)) {
    int errno2 = errno;
    /* Check error. */
    switch (errno2) {
      /* Check for RPC version mismatch. */
      case ENOPROTOOPT: {
        RMIX_LOG((
          RMIX_WARN(RPC version mismatch (%d <= version <= %d)), low, high))
      }
      /* Check for program version mismatch. */
      case ENOTSUP: {
        RMIX_LOG((
          RMIX_WARN(program version mismatch (%d <= version <= %d)), low, high))
      }
      /* Catch all other errors. */
      default: {
        RMIX_LOG((RMIX_WARN(unable to read reply w/o parameters)))
      }
    }
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    /* Deallocate client authentication credentials. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate client authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = errno2;
    return -1;
  }
  /* Pop close TCP/IPv4 client from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate authentication verifier from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate authentication credential from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate call output values from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate output array type signature from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Verify server. */
  if (0 != rmix_rpcx_auth_verify(creflav,  &cresize,  &crebody,
                                 verflav,  &versize,  &verbody,
                                 verflav2, &versize2, &verbody2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to verify server)))
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    /* Deallocate client authentication credentials. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate client authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = errno2;
    return -1;
  }
  /* Lock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = error;
    return -1;
  }
  /* Increase invocation count. */
  index = rmix_rpcx_data.invocations.count;
  rmix_rpcx_data.invocations.count++;
  /* Reallocate invocation array. */
  if (NULL == (array = realloc(rmix_rpcx_data.invocations.array,
                               rmix_rpcx_data.invocations.count *
                        sizeof(rmix_rpcx_data.invocations.array[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate invocation array)))
    /* Decrease invocation count. */
    rmix_rpcx_data.invocations.count--;
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno2 = errno;
    return -1;
  }
  rmix_rpcx_data.invocations.array = array;
  /* Initialize invocation array entry.*/
  rmix_rpcx_data.invocations.array[index].invocation = 0;
  rmix_rpcx_data.invocations.array[index].client     = client;
  rmix_rpcx_data.invocations.array[index].sigstr     = sigstr;
  rmix_rpcx_data.invocations.array[index].outary     = outary;
  rmix_rpcx_data.invocations.array[index].outcnt     = outcnt;
  rmix_rpcx_data.invocations.array[index].error      = 0;
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Submit job to retrieve output. */
  argument = client;
  if (0 != rmix_threads_submit(rmix_rpcx_oneway_retrieve_job,
                               NULL, &argument)) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to submit job to retrieve output)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Retrieves the output (if any) of a one-way invocation in a separate thread
 * using an invokation id previously issued by the provider.
 *
 * cleanup  = NULL.
 * argument = The TCP/IPv4 client.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_oneway_retrieve_job"
RMIX_THREADS_JOB(rmix_rpcx_oneway_retrieve_job) {
  int                     error;
  char                   *sigstr;
  void                  **outary, *array;
  unsigned int            index, outcnt;
  rmix_istream_t          istream;
  rmix_valary_cleanup_t   ocleanup;
  rmix_tcpip4_client_t   *client;
#ifdef DEBUG
  /* Check argument parameter. */
  if (NULL == argument) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argument parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Get the TCP/IPv4 client. */
  client = (rmix_tcpip4_client_t*)argument;
  /* Lock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Search for TCP/IPv4 client in invocation array. */
  for (index = 0; index < rmix_rpcx_data.invocations.count; index++) {
    if (rmix_rpcx_data.invocations.array[index].client == client) {
      break;
    }
  }
  /* Check if TCP/IPv4 client is not in invocation array. */
  if (index == rmix_rpcx_data.invocations.count) {
    errno = ENOENT;
    RMIX_LOG((RMIX_WARN(TCP/IPv4 client is not in invocation array)))
    errno = ENOENT;
    rmix_syserr();
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = ENOENT;
    return -1;
  }
  /* Restore invocation array entry.*/
  sigstr = rmix_rpcx_data.invocations.array[index].sigstr;
  outary = rmix_rpcx_data.invocations.array[index].outary;
  outcnt = rmix_rpcx_data.invocations.array[index].outcnt;
  /* Remove invocation data from invocation array. */
  rmix_rpcx_data.invocations.count--;
  memmove(rmix_rpcx_data.invocations.array + index,
          rmix_rpcx_data.invocations.array + index + 1,
         (rmix_rpcx_data.invocations.count - index) *
   sizeof(rmix_rpcx_data.invocations.array[0]));
  /* Reallocate invocation array. */
  if (0 == rmix_rpcx_data.invocations.count) {
    free(rmix_rpcx_data.invocations.array);
    rmix_rpcx_data.invocations.array = NULL;
  } else if (NULL ==
             (array = realloc(rmix_rpcx_data.invocations.array,
                              rmix_rpcx_data.invocations.count *
                       sizeof(rmix_rpcx_data.invocations.array[0])))) {
    RMIX_LOG((RMIX_WARN(unable to reallocate invocations array)))
    errno = 0;
  } else {
    rmix_rpcx_data.invocations.array = array;
  }
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = error;
    return -1;
  }
  /* Prepare iutput stream. */
  istream.incoming = client;
  istream.read     = rmix_tcpip4_client_istream_read;
  istream.close    = rmix_tcpip4_client_istream_close;
  /* Push deallocate output array type signature to pthread cleanup stack. */
  pthread_cleanup_push(free, sigstr);
  /* Push deallocate call output values to pthread cleanup stack. */
  ocleanup.valary = outary;
  ocleanup.valcnt = outcnt;
  ocleanup.sigstr = sigstr;
  pthread_cleanup_push(rmix_valary_cleanup, &ocleanup);
  /* Push close TCP/IPv4 client to pthread cleanup stack. */
  pthread_cleanup_push(rmix_tcpip4_client_close2, client);
  /* Read procedure return. */
  if (0 != rmix_rpcx_stream_read_reply2(&istream, sigstr, outary, outcnt)) {
    RMIX_LOG((RMIX_WARN(unable to read procedure return)))
  }
  /* Pop close TCP/IPv4 client from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate call output values from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate output array type signature from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Close TCP/IPv4 client. */
  if (0 != rmix_tcpip4_client_close(&client)) {
    RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
  }
  /* Destroy output array. */
  if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
    RMIX_LOG((RMIX_WARN(unable to destroy output array)))
  }
  /* Deallocate output array type signature duplicate. */
  free(sigstr);
  return 0;
}

/*
 * Asynchronously invokes a method at a remote object, i.e. invocation returns
 * immediately after sending message. The output may be retrieved later using
 * the invocation id. The provider gets not unloaded until the output was
 * retrieved.
 *
 * remoteref  = The remote object reference.
 * interface  = The remote object interface.
 * method     = The index of the method to invoke.
 * inary      = The input values array.
 * incnt      = The input values count.
 * invocation = The invocation id (return).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 *              ECONNABORTED = Connection closed before data transfer.
 *              EPIPE        = Connection closed during data transfer.
 *              ESRCH        = Remote prognum unavailable.
 *              ENOTSUP      = Remote progver mismatch.
 *              EOPNOTSUPP   = Remote procnum unavailable.
 *              EBADMSG      = Remote garbled args.
 *              ENOSYS       = Remote system error.
 *              ENOPROTOOPT  = Remote RPC version mismatch.
 *              EACCES       = Remote authentication error.
 *              EPROTO       = Protocol error.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_send"
RMIX_PROVIDERS_SEND(rmix_rpcx_send) {
  int                           error, low, high;
  char                         *sigstr, *crebody, *verbody, *verbody2;
  void                        **outary, *array, *argument;
  unsigned int                  prognum, outcnt, cresize,  versize,  versize2,
                                callid, objectid, index,
                                index2;
  rmix_istream_t                istream;
  rmix_ostream_t                ostream;
  rmix_tcpip4_client_t         *client;
  rmix_valary_cleanup_t         ocleanup;
  rmix_rpcx_stream_authflav_t   creflav,  verflav,  verflav2;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check invocation parameter. */
  if (NULL == invocation) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(invocation parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Get thread id and set it as call id. */
  if (0 != rmix_threadids_get(&callid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get thread id and to set it as call id)))
    errno = errno2;
    return -1;
  }
  /* Get object id from remote reference. */
  if (0 != rmix_remoteref_objectid(remoteref, &objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get object id from remote reference)))
    errno = errno2;
    return -1;
  }
  /* Compute program number. */
  prognum = objectid | 0x20000000;
  /* Duplicate output array type signature. */
  if (NULL == (sigstr = strdup(interface->methods[method].output))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate output array type signature)))
    errno = errno2;
    return -1;
  }
  /* Create output array. */
  if (0 != rmix_valary_create(&outary, &outcnt, sigstr)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create output array)))
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = errno2;
    return -1;
  }
  /* Create authentication credentials/verifier. */
  if (0 != rmix_rpcx_auth_create(&creflav, &cresize, &crebody,
                                 &verflav, &versize, &verbody)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create authentication credentials/verifier)))
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = errno2;
    return -1;
  }
  /* Open TCP/IPv4 client. */
  if (0 != rmix_tcpip4_client_open3(&client, remoteref->location)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open TCP/IPv4 client)))
    /* Deallocate client authentication credentials. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate client authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = errno2;
    return -1;
  }
  /* Prepare output stream. */
  ostream.outgoing = client;
  ostream.write    = rmix_tcpip4_client_ostream_write;
  ostream.close    = rmix_tcpip4_client_ostream_close;
  /* Push deallocate output array type signature to pthread cleanup stack. */
  pthread_cleanup_push(free, sigstr);
  /* Push deallocate call output values to pthread cleanup stack. */
  ocleanup.valary = outary;
  ocleanup.valcnt = outcnt;
  ocleanup.sigstr = sigstr;
  pthread_cleanup_push(rmix_valary_cleanup, &ocleanup);
  /* Push deallocate authentication credential to pthread cleanup stack. */
  pthread_cleanup_push(free, crebody);
  /* Push deallocate authentication verifier to pthread cleanup stack. */
  pthread_cleanup_push(free, verbody);
  /* Push close TCP/IPv4 client to pthread cleanup stack. */
  pthread_cleanup_push(rmix_tcpip4_client_close2, client);
  /* Write call. */
  if (0 != rmix_rpcx_stream_write_call(&ostream, callid, prognum, 0, method,
             creflav, cresize, crebody, verflav, versize, verbody,
             interface->methods[method].input, inary, incnt)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to write call)))
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    /* Deallocate client authentication credentials. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate client authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = errno2;
    return -1;
  }
  /* Prepare iutput stream. */
  istream.incoming = client;
  istream.read     = rmix_tcpip4_client_istream_read;
  istream.close    = rmix_tcpip4_client_istream_close;
  /* Read reply w/o parameters. */
  if (0 != rmix_rpcx_stream_read_reply(&istream, callid, &verflav2, &versize2,
                                       &verbody2, &low, &high)) {
    int errno2 = errno;
    /* Check error. */
    switch (errno2) {
      /* Check for RPC version mismatch. */
      case ENOPROTOOPT: {
        RMIX_LOG((
          RMIX_WARN(RPC version mismatch (%d <= version <= %d)), low, high))
      }
      /* Check for program version mismatch. */
      case ENOTSUP: {
        RMIX_LOG((
          RMIX_WARN(program version mismatch (%d <= version <= %d)), low, high))
      }
      /* Catch all other errors. */
      default: {
        RMIX_LOG((RMIX_WARN(unable to read reply w/o parameters)))
      }
    }
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    /* Deallocate client authentication credentials. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate client authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = errno2;
    return -1;
  }
  /* Pop close TCP/IPv4 client from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate authentication verifier from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate authentication credential from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate call output values from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate output array type signature from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Verify server. */
  if (0 != rmix_rpcx_auth_verify(creflav,  &cresize,  &crebody,
                                 verflav,  &versize,  &verbody,
                                 verflav2, &versize2, &verbody2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to verify server)))
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    /* Deallocate client authentication credentials. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate client authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = errno2;
    return -1;
  }
  /* Lock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno = error;
    return -1;
  }
  /* Increase invocation count. */
  index = rmix_rpcx_data.invocations.count;
  rmix_rpcx_data.invocations.count++;
  /* Reallocate invocation array. */
  if (NULL == (array = realloc(rmix_rpcx_data.invocations.array,
                               rmix_rpcx_data.invocations.count *
                        sizeof(rmix_rpcx_data.invocations.array[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate invocation array)))
    /* Decrease invocation count. */
    rmix_rpcx_data.invocations.count--;
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    /* Destroy output array. */
    if (0 != rmix_valary_destroy(&outary, &outcnt, sigstr)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output array)))
    }
    /* Deallocate output array type signature duplicate. */
    free(sigstr);
    errno2 = errno;
    return -1;
  }
  rmix_rpcx_data.invocations.array = array;
  /* Initialize invocation array entry.*/
  rmix_rpcx_data.invocations.array[index].invocation = 0;
  rmix_rpcx_data.invocations.array[index].client     = client;
  rmix_rpcx_data.invocations.array[index].sigstr     = sigstr;
  rmix_rpcx_data.invocations.array[index].outary     = outary;
  rmix_rpcx_data.invocations.array[index].outcnt     = outcnt;
  rmix_rpcx_data.invocations.array[index].error      = 0;
  /* Randomly generate a unique non-zero 32-bit invocation id. */
  do {
    *invocation = random() & 0xFFFFFFFF;
    for (index2 = 0; index2 < rmix_rpcx_data.invocations.count; index2++) {
      if (*invocation == rmix_rpcx_data.invocations.array[index2].invocation) {
          break;
      }
    }
  } while (index2 < rmix_rpcx_data.invocations.count);
  /* Save invocation id. */
  rmix_rpcx_data.invocations.array[index].invocation = *invocation;
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Submit job to retrieve output. */
  argument = client;
  if (0 != rmix_threads_submit(rmix_rpcx_async_retrieve_job, NULL, &argument)) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to submit job to retrieve output)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Retrieves the output (if any) of an asynchronous invocation in a separate
 * thread using an invokation id previously issued by the provider.
 *
 * cleanup  = NULL.
 * argument = The TCP/IPv4 client.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_async_retrieve_job"
RMIX_THREADS_JOB(rmix_rpcx_async_retrieve_job) {
  int                     error, error2;
  char                   *sigstr;
  void                  **outary;
  unsigned int            index, outcnt;
  rmix_istream_t          istream;
  rmix_valary_cleanup_t   ocleanup;
  rmix_tcpip4_client_t   *client;
#ifdef DEBUG
  /* Check argument parameter. */
  if (NULL == argument) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argument parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Get the TCP/IPv4 client. */
  client = (rmix_tcpip4_client_t*)argument;
  /* Lock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Search for TCP/IPv4 client in invocation array. */
  for (index = 0; index < rmix_rpcx_data.invocations.count; index++) {
    if (rmix_rpcx_data.invocations.array[index].client == client) {
      break;
    }
  }
  /* Check if TCP/IPv4 client is not in invocation array. */
  if (index == rmix_rpcx_data.invocations.count) {
    errno = ENOENT;
    RMIX_LOG((RMIX_WARN(TCP/IPv4 client is not in invocation array)))
    errno = ENOENT;
    rmix_syserr();
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = ENOENT;
    return -1;
  }
  /* Restore invocation array entry.*/
  sigstr = rmix_rpcx_data.invocations.array[index].sigstr;
  outary = rmix_rpcx_data.invocations.array[index].outary;
  outcnt = rmix_rpcx_data.invocations.array[index].outcnt;
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Prepare iutput stream. */
  istream.incoming = client;
  istream.read     = rmix_tcpip4_client_istream_read;
  istream.close    = rmix_tcpip4_client_istream_close;
  /* Push deallocate output array type signature to pthread cleanup stack. */
  pthread_cleanup_push(free, sigstr);
  /* Push deallocate call output values to pthread cleanup stack. */
  ocleanup.valary = outary;
  ocleanup.valcnt = outcnt;
  ocleanup.sigstr = sigstr;
  pthread_cleanup_push(rmix_valary_cleanup, &ocleanup);
  /* Push close TCP/IPv4 client to pthread cleanup stack. */
  pthread_cleanup_push(rmix_tcpip4_client_close2, client);
  /* Read procedure return. */
  if (0 != rmix_rpcx_stream_read_reply2(&istream, sigstr, outary, outcnt)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to read procedure return)))
    error2 = errno2;
  } else {
    error2 = 0;
  }
  /* Pop close TCP/IPv4 client from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate call output values from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate output array type signature from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Lock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Search for TCP/IPv4 client in invocation array. */
  for (index = 0; index < rmix_rpcx_data.invocations.count; index++) {
    if (rmix_rpcx_data.invocations.array[index].client == client) {
      break;
    }
  }
  /* Check if TCP/IPv4 client is not in invocation array. */
  if (index == rmix_rpcx_data.invocations.count) {
    errno = ENOENT;
    RMIX_LOG((RMIX_WARN(TCP/IPv4 client is not in invocation array)))
    errno = ENOENT;
    rmix_syserr();
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = ENOENT;
    return -1;
  }
  /* Modify invocation array entry.*/
  rmix_rpcx_data.invocations.array[index].client = NULL;
  rmix_rpcx_data.invocations.array[index].error  = error2;
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    /* Broadcast invocations ready condition. */
    if (0 != (error =
              pthread_cond_broadcast(&rmix_rpcx_data.invocations.ready))) {
      RMIX_LOG((RMIX_WARN(unable to broadcast invocations ready condition)))
    }
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    errno = error;
    return -1;
  }
  /* Broadcast invocations ready condition. */
  if (0 != (error =
            pthread_cond_broadcast(&rmix_rpcx_data.invocations.ready))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to broadcast invocations ready condition)))
    errno = error;
    rmix_syserr();
    /* Close TCP/IPv4 client. */
    if (0 != rmix_tcpip4_client_close(&client)) {
      RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    }
    errno = error;
    return -1;
  }
  /* Close TCP/IPv4 client. */
  if (0 != rmix_tcpip4_client_close(&client)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to close TCP/IPv4 client)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Retrieves the output (if any) of an asynchronous invocation using an
 * invokation id previously issued by the provider.
 *
 * remoteref  = The remote object reference.
 * interface  = The remote object interface.
 * method     = The index of the method to invoke.
 * outary     = The output values array.
 * outcnt     = The output values count.
 * invocation = The invocation id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 *              ECONNABORTED = Connection closed before data transfer.
 *              EPIPE        = Connection closed during data transfer.
 *              ENOENT       = Invocation not found.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_retrieve"
RMIX_PROVIDERS_RETRIEVE(rmix_rpcx_retrieve) {
  int            error, error2;
  char          *sigstr2;
  void         **outary2, *array;
  unsigned int   index, outcnt2;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Wait until TCP/IPv4 client is NULL. */
  while (1) {
    /* Search for invocation id in invocation array. */
    for (index = 0; index < rmix_rpcx_data.invocations.count; index++) {
      if (rmix_rpcx_data.invocations.array[index].invocation == invocation) {
        break;
      }
    }
    /* Check if invocation id is not in invocation array. */
    if (index == rmix_rpcx_data.invocations.count) {
      errno = ENOENT;
      RMIX_LOG((RMIX_WARN(invocation id is not in invocation array)))
      /* Unlock rmix-rpcx plug-in mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
        errno = error;
        rmix_syserr();
      }
      errno = ENOENT;
      return -1;
    }
    /* Check if TCP/IPv4 client is NULL. */
    if (NULL == rmix_rpcx_data.invocations.array[index].client) {
      break;
    }
    /* Wait for invocations ready condition. */
    if (0 != (error = pthread_cond_wait(&rmix_rpcx_data.invocations.ready,
                                        &rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to wait for invocations ready condition)))
      errno = error;
      rmix_syserr();
      /* Unlock rmix-rpcx plug-in mutex. */
      if (0 != (error2 = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
        errno = error2;
        RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
        errno = error2;
        rmix_syserr();
      }
      errno = error;
      return -1;
    }
  }
  /* Restore invocation array entry.*/
  sigstr2 = rmix_rpcx_data.invocations.array[index].sigstr;
  outary2 = rmix_rpcx_data.invocations.array[index].outary;
  outcnt2 = rmix_rpcx_data.invocations.array[index].outcnt;
  error2  = rmix_rpcx_data.invocations.array[index].error;
  /* Remove invocation data from invocation array. */
  rmix_rpcx_data.invocations.count--;
  memmove(rmix_rpcx_data.invocations.array + index,
          rmix_rpcx_data.invocations.array + index + 1,
         (rmix_rpcx_data.invocations.count - index) *
   sizeof(rmix_rpcx_data.invocations.array[0]));
  /* Reallocate invocation array. */
  if (0 == rmix_rpcx_data.invocations.count) {
    free(rmix_rpcx_data.invocations.array);
    rmix_rpcx_data.invocations.array = NULL;
  } else if (NULL ==
             (array = realloc(rmix_rpcx_data.invocations.array,
                              rmix_rpcx_data.invocations.count *
                       sizeof(rmix_rpcx_data.invocations.array[0])))) {
    RMIX_LOG((RMIX_WARN(unable to reallocate invocations array)))
    errno = 0;
  } else {
    rmix_rpcx_data.invocations.array = array;
  }
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    /* Destroy output values array. */
    if (0 != rmix_valary_destroy(&outary2, &outcnt2, sigstr2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output values array)))
    }
    /* Destroy output type signature. */
    free(sigstr2);
    errno = error;
    return -1;
  }
  /* Compare output type signatures. */
  if (0 != strcmp(interface->methods[method].output, sigstr2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(different output type signature)))
    /* Destroy output values array. */
    if (0 != rmix_valary_destroy(&outary2, &outcnt2, sigstr2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output values array)))
    }
    /* Destroy output type signature. */
    free(sigstr2);
    errno = errno2;
    return -1;
  }
  /* Check for error. */
  if (0 != error2) {
    errno = error2;
    RMIX_LOG((RMIX_WARN(unable to retrieve call return)))
    /* Destroy output values array. */
    if (0 != rmix_valary_destroy(&outary2, &outcnt2, sigstr2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output values array)))
    }
    /* Destroy output type signature. */
    free(sigstr2);
    errno = error2;
    return -1;
  }
  /* Moves all values. */
  if (0 != rmix_valary_move(outary, outary2, outcnt, sigstr2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to moves all values)))
    /* Destroy output values array. */
    if (0 != rmix_valary_destroy(&outary2, &outcnt2, sigstr2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy output values array)))
    }
    /* Destroy moved output values. */
    if (0 != rmix_valary_destroy2(outary, outcnt, sigstr2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy moved output values)))
    }
    /* Destroy output type signature. */
    free(sigstr2);
    errno = errno2;
    return -1;
  }
  /* Destroy output values array. */
  if (0 != rmix_valary_destroy(&outary2, &outcnt2, sigstr2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy output values array)))
    /* Destroy moved output values. */
    if (0 != rmix_valary_destroy2(outary, outcnt, sigstr2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy moved output values)))
    }
    /* Destroy output type signature. */
    free(sigstr2);
    errno = errno2;
    return -1;
  }
  /* Destroy output type signature. */
  free(sigstr2);
  return 0;
}

/*
 * Exports a local object.
 *
 * localref   = The local object reference of the exported object (return).
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may be NULL, "").
 *              Ignores any PROTOCOL or BINDING parameters.
 * provider   = The provider plug-in handle.
 * objectid   = The object handle.
 * object     = The local object (may be NULL).
 * interface  = The local object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_export"
RMIX_PROVIDERS_EXPORT(rmix_rpcx_export) {
  int                 error;
  char               *location;
  void               *array;
  const char         *interfaces, *host, *portstr, *backlogstr;
  unsigned int        port, backlog, index1, index2, server, prognum;
  struct sockaddr_in  address;
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check protocol parameter. */
  if (NULL == protocol) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(protocol parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check parameters parameter. */
  if (NULL == parameters) {
    /* Search for additional interfaces parameter. */
    if (0 != rmix_parameters_search(parameters, &interfaces, "INTERFACES")) {
      if (ESRCH != errno) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to search for interfaces parameter)))
        errno = errno2;
        return -1;
      }
      /* Set default interface. */
      interfaces = NULL;
    }
    /* Search for host parameter. */
    if (0 != rmix_parameters_search(parameters, &host, "HOST")) {
      if (ESRCH != errno) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to search for host parameter)))
        errno = errno2;
        return -1;
      }
      /* Set default host. */
      host = NULL;
    }
    /* Search for port parameter. */
    if (0 != rmix_parameters_search(parameters, &portstr, "PORT")) {
      if (ESRCH != errno) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to search for port parameter)))
        errno = errno2;
        return -1;
      }
      /* Set default port. */
      port = RMIX_RPCX_PORT;
    } else {
      /* Convert port parameter to port. */
      port = (unsigned int)atoi(portstr);
    }
    /* Search for backlog parameter. */
    if (0 != rmix_parameters_search(parameters, &backlogstr, "BACKLOG")) {
      if (ESRCH != errno) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to search for backlog parameter)))
        errno = errno2;
        return -1;
      }
      /* Set default backlog. */
      backlog = RMIX_RPCX_BACKLOG;
    } else {
      /* Convert backlog parameter to backlog. */
      backlog = (unsigned int)atoi(backlogstr);
    }
  } else {
    /* Set default interface. */
    interfaces = NULL;
    /* Set default host. */
    host = NULL;
    /* Set default port. */
    port = RMIX_RPCX_PORT;
    /* Set default port. */
    backlog = RMIX_RPCX_BACKLOG;
  }
  /* Resolve address. */
  if (0 != rmix_tcpip4_resolve(&address, host, port)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve address)))
    errno = errno2;
    return -1;
  }
  /* Compute program number. */
  prognum = objectid | 0x20000000;
  /* Allocate and compose object location. */
  if (-1 == asprintf(&location, "rpc://%hhu.%hhu.%hhu.%hhu:%u/%x:%x",
                     ((char*)&(address.sin_addr))[0],
                     ((char*)&(address.sin_addr))[1],
                     ((char*)&(address.sin_addr))[2],
                     ((char*)&(address.sin_addr))[3],
                     port, prognum, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate and compose remote object location)))
    errno = errno2;
    return -1;
  }
  /* Create local object reference. */
  if (0 != rmix_localref_create3(localref, protocol, binding, interfaces,
                                 location, provider, objectid, object,
                                 interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create local object reference)))
    /* Deallocate object location. */
    free(location);
    errno = errno2;
    return -1;
  }
  /* Deallocate object location. */
  free(location);
  /* Lock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    /* Destroy local object reference. */
    if (0 != rmix_localref_destroy(localref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
    }
    errno = error;
    return -1;
  }
  /* Search for object id in servers. */
  for (index1 = 0; index1 < rmix_rpcx_data.servers.count; index1++) {
    /* Search for object id in exports. */
    for (index2 = 0;
         index2 < rmix_rpcx_data.servers.array[index1].exports.count;
         index2++) {
      /* Check for object id in exports. */
      if (objectid ==
          rmix_rpcx_data.servers.array[index1].exports.array[index2]) {
        errno = EALREADY;
        RMIX_LOG((RMIX_WARN(object id is already in use)))
        errno = EALREADY;
        rmix_syserr();
        /* Unlock rmix-rpcx plug-in mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
          errno = error;
          rmix_syserr();
        }
        /* Destroy local object reference. */
        if (0 != rmix_localref_destroy(localref)) {
          RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
        }
        errno = EALREADY;
        return -1;
      }
    }
  }
  /* Search for address/port in servers. */
  for (index1 = 0; index1 < rmix_rpcx_data.servers.count; index1++) {
    /* Check for address/port in server. */
    if ((((char*)&address.sin_addr)[0] ==
         ((char*)&rmix_rpcx_data.servers.array[index1].address.sin_addr)[0])&&
        (((char*)&address.sin_addr)[1] ==
         ((char*)&rmix_rpcx_data.servers.array[index1].address.sin_addr)[1])&&
        (((char*)&address.sin_addr)[2] ==
         ((char*)&rmix_rpcx_data.servers.array[index1].address.sin_addr)[2])&&
        (((char*)&address.sin_addr)[3] ==
         ((char*)&rmix_rpcx_data.servers.array[index1].address.sin_addr)[3])&&
        (address.sin_port ==
         rmix_rpcx_data.servers.array[index1].address.sin_port)) {
      /* Increase exports array. */
      index2 = rmix_rpcx_data.servers.array[index1].exports.count;
      rmix_rpcx_data.servers.array[index1].exports.count++;
      /* Reallocate exports array. */
      if (NULL ==
          (array =
           realloc(rmix_rpcx_data.servers.array[index1].exports.array,
                   rmix_rpcx_data.servers.array[index1].exports.count *
            sizeof(rmix_rpcx_data.servers.array[index1].exports.array[0])))) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to reallocate exports array)))
        /* Decrease exports array. */
        rmix_rpcx_data.servers.array[index1].exports.count++;
        /* Unlock rmix-rpcx plug-in mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
          errno = error;
          rmix_syserr();
        }
        /* Destroy local object reference. */
        if (0 != rmix_localref_destroy(localref)) {
          RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
        }
        errno2 = errno;
        return -1;
      }
      rmix_rpcx_data.servers.array[index1].exports.array = array;
      /* Save exports array entry. */
      rmix_rpcx_data.servers.array[index1].exports.array[index2] = objectid;
      /* Unlock rmix-rpcx plug-in mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
        errno = error;
        rmix_syserr();
        /* Destroy local object reference. */
        if (0 != rmix_localref_destroy(localref)) {
          RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
        }
        errno = error;
        return -1;
      }
      return 0;
    }
  }
  /* Open server. */
  if (0 != rmix_tcpip4_open2(&server, &address, backlog, rmix_rpcx_server)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to open server)))
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Destroy local object reference. */
    if (0 != rmix_localref_destroy(localref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
    }
    errno2 = errno;
    return -1;
  }
  /* Increase server array. */
  index1 = rmix_rpcx_data.servers.count;
  rmix_rpcx_data.servers.count++;
  /* Reallocate server array. */
  if (NULL == (array = realloc(rmix_rpcx_data.servers.array,
                               rmix_rpcx_data.servers.count *
                        sizeof(rmix_rpcx_data.servers.array[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate server array)))
    /* Decrease server array. */
    rmix_rpcx_data.servers.count--;
    /* Close server. */
    if (0 != rmix_tcpip4_close(server)) {
      RMIX_LOG((RMIX_WARN(unable to close server)))
    }
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Destroy local object reference. */
    if (0 != rmix_localref_destroy(localref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
    }
    errno2 = errno;
    return -1;
  }
  rmix_rpcx_data.servers.array = array;
  /* Allocate exports array. */
  if (NULL ==
      (array =
       malloc(sizeof(rmix_rpcx_data.servers.array[index1].exports.array[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate exports array)))
    /* Decrease server array. */
    rmix_rpcx_data.servers.count--;
    /* Reallocate server array. */
    if (0 == rmix_rpcx_data.servers.count) {
      free(rmix_rpcx_data.servers.array);
      rmix_rpcx_data.servers.array = NULL;
    } else if (NULL == (array = realloc(rmix_rpcx_data.servers.array,
                                        rmix_rpcx_data.servers.count *
                                 sizeof(rmix_rpcx_data.servers.array[0])))) {
      RMIX_LOG((RMIX_WARN(unable to reallocate server array)))
    } else {
      rmix_rpcx_data.servers.array = array;
    }
    /* Close server. */
    if (0 != rmix_tcpip4_close(server)) {
      RMIX_LOG((RMIX_WARN(unable to close server)))
    }
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Destroy local object reference. */
    if (0 != rmix_localref_destroy(localref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
    }
    errno2 = errno;
    return -1;
  }
  /* Save server array entry. */
  rmix_rpcx_data.servers.array[index1].server           = server;
  rmix_rpcx_data.servers.array[index1].address          = address;
  rmix_rpcx_data.servers.array[index1].exports.count    = 1;
  rmix_rpcx_data.servers.array[index1].exports.array    = array;
  rmix_rpcx_data.servers.array[index1].exports.array[0] = objectid;
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    /* Destroy local object reference. */
    if (0 != rmix_localref_destroy(localref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
    }
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Unxports a local object.
 *
 * localref = The local object reference of the exported object (also return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_unexport"
RMIX_PROVIDERS_UNEXPORT(rmix_rpcx_unexport) {
  int           error, error2;
  void         *array;
  unsigned int  index1, index2, server, index;
  /* Lock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Search for object id in servers. */
  for (index2 = 0,
       index1 = 0; index1 < rmix_rpcx_data.servers.count; index1++) {
    /* Search for object id in exports. */
    for (index2 = 0;
         index2 < rmix_rpcx_data.servers.array[index1].exports.count;
         index2++) {
      /* Check for object id in exports. */
      if ((*localref)->objectid ==
          rmix_rpcx_data.servers.array[index1].exports.array[index2]) {
        break;
      }
    }
    if (index2 < rmix_rpcx_data.servers.array[index1].exports.count) {
      break;
    }
  }
  /* Check if object id is not in servers. */
  if (index1 == rmix_rpcx_data.servers.count) {
    errno = ESRCH;
    RMIX_LOG((RMIX_WARN(object id is not in use)))
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Decrease exports array. */
  rmix_rpcx_data.servers.array[index1].exports.count--;
  memmove(rmix_rpcx_data.servers.array[index1].exports.array + index2,
          rmix_rpcx_data.servers.array[index1].exports.array + index2 + 1,
         (rmix_rpcx_data.servers.array[index1].exports.count - index2) *
   sizeof(rmix_rpcx_data.servers.array[index1].exports.array[0]));
  /* Reallocate exports array. */
  if (0 == rmix_rpcx_data.servers.array[index1].exports.count) {
    free(rmix_rpcx_data.servers.array[index1].exports.array);
    rmix_rpcx_data.servers.array[index1].exports.array = NULL;
    /* Save server array entry. */
    server = rmix_rpcx_data.servers.array[index1].server;
    /* Decrease server array. */
    rmix_rpcx_data.servers.count--;
    memmove(rmix_rpcx_data.servers.array + index1,
            rmix_rpcx_data.servers.array + index1 + 1,
           (rmix_rpcx_data.servers.count - index1) *
     sizeof(rmix_rpcx_data.servers.array[0]));
    /* Reallocate server array. */
    if (0 == rmix_rpcx_data.servers.count) {
      free(rmix_rpcx_data.servers.array);
      rmix_rpcx_data.servers.array = NULL;
    } else if (NULL == (array = realloc(rmix_rpcx_data.servers.array,
                                        rmix_rpcx_data.servers.count *
                                 sizeof(rmix_rpcx_data.servers.array[0])))) {
      RMIX_LOG((RMIX_WARN(unable to reallocate server array)))
      errno = 0;
    } else {
      rmix_rpcx_data.servers.array = array;
    }
    /* Close server. */
    if (0 != rmix_tcpip4_close(server)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to close server)))
      /* Unlock rmix-rpcx plug-in mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
        errno = error;
        rmix_syserr();
      }
      /* Destroy local object reference. */
      if (0 != rmix_localref_destroy(localref)) {
        RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
      }
      errno = errno2;
      return -1;
    }
  } else if (NULL ==
             (array =
              realloc(
               rmix_rpcx_data.servers.array[index1].exports.array,
               rmix_rpcx_data.servers.array[index1].exports.count *
        sizeof(rmix_rpcx_data.servers.array[index1].exports.array[0])))) {
    RMIX_LOG((RMIX_WARN(unable to reallocate exports array)))
    errno = 0;
  } else {
    rmix_rpcx_data.servers.array[index1].exports.array = array;
  }
  /* Search for object id in reading array. */
  for (index = 0; index < rmix_rpcx_data.reading.count; index++) {
    /* Check for for object id in reading array. */
    if (rmix_rpcx_data.reading.array[index].objectid == (*localref)->objectid) {
      /* Push unlock rmix-rpcx plug-in mutex to pthread cleanup stack. */
      pthread_cleanup_push(rmix_rpcx_unlock, &rmix_rpcx_data.mutex);
      /* Wait for reading ready condition. */
      if (0 != (error = pthread_cond_wait(&rmix_rpcx_data.reading.ready,
                                          &rmix_rpcx_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to wait for reading ready condition)))
        errno = error;
        rmix_syserr();
        /* Unlock rmix-rpcx plug-in mutex. */
        if (0 != (error2 = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
          errno = error2;
          RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
          errno = error2;
          rmix_syserr();
        }
        /* Destroy local object reference. */
        if (0 != rmix_localref_destroy(localref)) {
          RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
        }
        errno = error;
        return -1;
      }
      /* Pop unlock rmix-rpcx plug-in mutex from pthread cleanup stack. */
      pthread_cleanup_pop(0);
      /* Check for finalized rmix-rpcx plug-in. */
      if (0 == rmix_rpcx_data.ready) {
        errno = EACCES;
        RMIX_LOG((RMIX_WARN(rmix-rpcx plug-in finalized)))
        /* Unlock rmix-rpcx plug-in mutex. */
        if (0 != (error2 = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
          errno = error2;
          RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
          errno = error2;
          rmix_syserr();
        }
        /* Destroy local object reference. */
        if (0 != rmix_localref_destroy(localref)) {
          RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
        }
        errno = EACCES;
        return -1;
      }
      /* Reset index. */
      index = 0;
      index--;
    }
  }
  /* Search for object id in executing array. */
  for (index = 0; index < rmix_rpcx_data.executing.count; index++) {
    /* Check for for object id in executing array. */
    if (rmix_rpcx_data.executing.array[index].objectid ==
        (*localref)->objectid) {
      /* Push unlock rmix-rpcx plug-in mutex to pthread cleanup stack. */
      pthread_cleanup_push(rmix_rpcx_unlock, &rmix_rpcx_data.mutex);
      /* Wait for executing ready condition. */
      if (0 != (error = pthread_cond_wait(&rmix_rpcx_data.executing.ready,
                                          &rmix_rpcx_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to wait for executing ready condition)))
        errno = error;
        rmix_syserr();
        /* Unlock rmix-rpcx plug-in mutex. */
        if (0 != (error2 = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
          errno = error2;
          RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
          errno = error2;
          rmix_syserr();
        }
        /* Destroy local object reference. */
        if (0 != rmix_localref_destroy(localref)) {
          RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
        }
        errno = error;
        return -1;
      }
      /* Pop unlock rmix-rpcx plug-in mutex from pthread cleanup stack. */
      pthread_cleanup_pop(0);
      /* Check for finalized rmix-rpcx plug-in. */
      if (0 == rmix_rpcx_data.ready) {
        errno = EACCES;
        RMIX_LOG((RMIX_WARN(rmix-rpcx plug-in finalized)))
        /* Unlock rmix-rpcx plug-in mutex. */
        if (0 != (error2 = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
          errno = error2;
          RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
          errno = error2;
          rmix_syserr();
        }
        /* Destroy local object reference. */
        if (0 != rmix_localref_destroy(localref)) {
          RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
        }
        errno = EACCES;
        return -1;
      }
      /* Reset index. */
      index = 0;
      index--;
    }
  }
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    /* Destroy local object reference. */
    if (0 != rmix_localref_destroy(localref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
    }
    errno = error;
    return -1;
  }
  /* Destroy local object reference. */
  if (0 != rmix_localref_destroy(localref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy local object reference)))
    errno2 = errno;
    return -1;
  }
  return 0;
}

/*
 * Creates a remote object reference, while forcing a specific object id.
 * This is very useful for accessing standard object servers at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may not be NULL, "").
 *              Ignores any PROTOCOL or BINDING parameters. Requires HOST and
 *              PORT parameters.
 * objectid   = The object id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_remoteref_create"
RMIX_PROVIDERS_REMOTEREF_CREATE(rmix_rpcx_remoteref_create) {
  char               *location;
  const char         *host, *portstr, *interfaces;
  unsigned int        port, prognum;
  struct sockaddr_in  address;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check protocol parameter. */
  if (NULL == protocol) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(protocol parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check objectid parameter. */
  if (0xFFFF < objectid) {
    errno = E2BIG;
    RMIX_LOG((RMIX_WARN(objectid parameter is too big)))
    errno = E2BIG;
    return -1;
  }
  /* Search for host parameter. */
  if (0 != rmix_parameters_search(parameters, &host, "HOST")) {
    if (ESRCH == errno) {
      /* Set first available non-loopback interface. */
      host = NULL;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for host parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Search for port parameter. */
  if (0 != rmix_parameters_search(parameters, &portstr, "PORT")) {
    if (ESRCH == errno) {
      /* Set default port. */
      port = RMIX_RPCX_PORT;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for port parameter)))
      errno = errno2;
      return -1;
    }
  } else {
    /* Convert port parameter to port. */
    port = (unsigned int)atoi(portstr);
  }
  /* Search for additional binding parameter. */
  if (0 != rmix_parameters_search(parameters, &binding, "BINDING")) {
    if (ESRCH != errno) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for binding parameter)))
      errno = errno2;
      return -1;
    }
    binding = NULL;
  }
  /* Search for additional interfaces parameter. */
  if (0 != rmix_parameters_search(parameters, &interfaces, "INTERFACES")) {
    if (ESRCH != errno) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for interfaces parameter)))
      errno = errno2;
      return -1;
    }
    interfaces = NULL;
  }
  /* Resolve address. */
  if (0 != rmix_tcpip4_resolve(&address, host, port)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve address)))
    errno = errno2;
    return -1;
  }
  /* Compute program number. */
  prognum = objectid | 0x20000000;
  /* Allocate and compose object location. */
  if (-1 == asprintf(&location, "rpc://%hhu.%hhu.%hhu.%hhu:%u/%x:%x",
                     ((char*)&(address.sin_addr))[0],
                     ((char*)&(address.sin_addr))[1],
                     ((char*)&(address.sin_addr))[2],
                     ((char*)&(address.sin_addr))[3],
                     port, prognum, 0)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate and compose remote object location)))
    errno = errno2;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != rmix_remoteref_create3(remoteref, protocol, binding,
                                  interfaces, location, objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create remote object reference)))
    /* Deallocate remote object location. */
    free(location);
    errno = errno2;
    return -1;
  }
  /* Deallocate remote object location. */
  free(location);
  return 0;
}

/*
 * Returs a remote object reference in parameters form.
 *
 * parameters = The parameters form to create.
 * remoteref  = The remote object reference to convert.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_remoteref_parameters"
RMIX_PROVIDERS_REMOTEREF_PARAMETERS(rmix_rpcx_remoteref_parameters) {
  unsigned int  count, length, index;
  char         *names[6], *values[6], host[256], port[5], objectid[10],
               *interfaces;
#ifdef DEBUG
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  count = 0;
  /* Add protocol name/value pair. */
  names [count] = "PROTOCOL";
  values[count] = "RPC";
  count++;
  /* Reconstruct host and port. */
  if (2 != sscanf(remoteref->location, "rpc://%255[0-9.]:%5[0-9.]/%*x:%*x",
                  host, port)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reconstruct host and port)))
    errno = errno2;
    return -1;
  }
  /* Add host name/value pair. */
  names [count] = "HOST";
  values[count] = host;
  count++;
  /* Add port name/value pair. */
  names [count] = "PORT";
  values[count] = port;
  count++;
  /* Reconstruct object id. */
  if (1 != sscanf(remoteref->object, "%*u:%*u:%10[0-9]:%*u", objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reconstruct object id)))
    errno = errno2;
    return -1;
  }
  /* Add object id name/value pair. */
  names [count] = "OBJECTID";
  values[count] = objectid;
  count++;
  /* Check for interfaces. */
  if (0 != remoteref->interfaces.count) {
    /* Calculate interfaces string length. */
    for (length = 0, index = 0; index < remoteref->interfaces.count; index++) {
      /* Calculate interface string length and
         increase interfaces string length. */
      length += strlen(remoteref->interfaces.array[index]) + 1;
    }
    /* Allocate interfaces string. */
    if (NULL == (interfaces = malloc(length * sizeof(char*)))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to allocate interfaces string)))
      errno = errno2;
      return -1;
    }
    /* Construct interfaces string. */
    for (length = 0, index = 0; index < remoteref->interfaces.count; index++) {
      unsigned int length2;
      /* Calculate interface string length. */
      length2 = strlen(remoteref->interfaces.array[index]);
      /* Copy interface string. */
      memcpy(interfaces + length, remoteref->interfaces.array[index], length2);
      /* Increase interface string length. */
      length += length2;
      /* Add delimiter. */
      interfaces[length] = ';';
      length++;
    }
    /* Terminate interfaces string. */
    interfaces[length - 1] = '\0';
    /* Add interfaces name/value pair. */
    names [count] = "INTERFACES";
    values[count] = interfaces;
    count++;
  } else {
    interfaces = NULL;
  }
  /* Check for binding. */
  if (NULL != remoteref->binding) {
    /* Add binding name/value pair. */
    names [count] = "BINDING";
    values[count] = remoteref->binding;
    count++;
  }
  /* Create parameters */
  if (0 != rmix_parameters_create2(parameters, count,
                                   (const char**)names,
                                   (const char**)values)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create parameters)))
    /* Deallocate interfaces string. */
    if (NULL != interfaces) {
      free(interfaces);
    }
    errno = errno2;
    return -1;
  }
  /* Deallocate interfaces string. */
  if (NULL != interfaces) {
    free(interfaces);
  }
  return 0;
}

/*
 * Serves a connection.
 *
 * server = The TCP/IPv4 server.
 * return = The server job function pointer on success or
 *          NULL on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_server"
RMIX_TCPIP4_SERVERJOB(rmix_rpcx_server) {
  char                        *crebody, *verbody;
  void                        *object, **inary, **outary;
  unsigned int                 objectid, callid, prognum, progver, procnum,
                               cresize, versize, incnt, outcnt;
  rmix_istream_t               istream;
  rmix_ostream_t               ostream;
  rmix_valary_cleanup_t        icleanup, ocleanup;
  const rmix_interface_t      *interface;
  rmix_rpcx_stream_authstat_t  authstat;
  rmix_rpcx_stream_authflav_t  creflav, verflav;
#ifdef DEBUG
  /* Check server parameter. */
  if (NULL == server) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(server parameter is null)))
    errno = EINVAL;
    return NULL;
  }
#endif /* DEBUG */
  /* Prepare input/output streams. */
  istream.incoming = server;
  istream.read     = rmix_tcpip4_server_istream_read;
  istream.close    = rmix_tcpip4_server_istream_close;
  ostream.outgoing = server;
  ostream.write    = rmix_tcpip4_server_ostream_write;
  ostream.close    = rmix_tcpip4_server_ostream_close;
  /* Read call header. */
  if (0 != rmix_rpcx_stream_read_call(&istream, &ostream, &callid,  &prognum,
                                      &progver, &procnum, &creflav, &cresize,
                                      &crebody, &verflav, &versize, &verbody)) {
    /* Check for other errors than connection closed before transfer. */
    if (ECONNABORTED != errno) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to read call header)))
      errno = errno2;
    }
    return NULL;
  }
  /* Push deallocate authentication credential to pthread cleanup stack. */
  pthread_cleanup_push(free, crebody);
  /* Push deallocate authentication verifier to pthread cleanup stack. */
  pthread_cleanup_push(free, verbody);
  /* Compute object id. */
  objectid = prognum & 0xFFFF;
  /* Enter reading state. */
  if (0 != rmix_rpcx_reading(callid, objectid)) {
    int errno2 = errno;
    /* Check for finalized rmix-rpcx plug-in. */
    if (EACCES == errno) {
      /* Close the incoming XDR data stream. */
      if (0 != istream.close(istream.incoming)) {
        RMIX_LOG((RMIX_WARN(unable to close the incoming XDR data stream)))
      } else {
        /* Write program unavailable message. */
        if (0 != rmix_rpcx_stream_write_error(&ostream, callid, verflav,
                   versize, verbody,
                   RMIX_RPCX_STREAM_ACCEPTSTAT_PROG_UNAVAIL, 0, 0)) {
          /* Check for other errors than connection closed before transfer. */
          if (ECONNABORTED != errno) {
            RMIX_LOG((RMIX_WARN(unable to write program unavailable message)))
          }
        }
      }
    } else {
      RMIX_LOG((RMIX_WARN(unable to enter reading state)))
    }
    /* Deallocate authentication credential. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    errno = errno2;
    return NULL;
  }
  /* Pop deallocate authentication verifier from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate authentication credential from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Authenticate client. */
  if (0 != rmix_rpcx_auth_authenticate(creflav, &cresize, &crebody, &verflav,
                                       &versize, &verbody, &authstat)) {
    /* Close the incoming XDR data stream. */
    if (0 != istream.close(istream.incoming)) {
      RMIX_LOG((RMIX_WARN(unable to close the incoming XDR data stream)))
    } else {
      /* Write authentication error message. */
      if (0 != rmix_rpcx_stream_write_deny(&ostream, callid, authstat)) {
        /* Check for other errors than connection closed before transfer. */
        if (ECONNABORTED != errno) {
          RMIX_LOG((RMIX_WARN(unable to write authentication error message)))
        }
      }
    }
    /* Deallocate authentication credential. */
    if (NULL != crebody) {
      free(crebody);
    }
    /* Deallocate authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    /* Enter executing state. */
    if (0 != rmix_rpcx_executing(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executing state)))
    }
    /* Enter executed state. */
    if (0 != rmix_rpcx_executed(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executed state)))
    }
    errno = EPROTO;
    return NULL;
  }
  /* Push deallocate authentication verifier to pthread cleanup stack. */
  pthread_cleanup_push(free, verbody);
  /* Lookup object at object registry. */
  if (0 != rmix_objects_get(objectid, &object, &interface)) {
    /* Check if object is not in registry. */
    if (ESRCH == errno) {
      /* Close the incoming XDR data stream. */
      if (0 != istream.close(istream.incoming)) {
        RMIX_LOG((RMIX_WARN(unable to close the incoming XDR data stream)))
      } else {
        /* Write program unavailable message. */
        if (0 != rmix_rpcx_stream_write_error(&ostream, callid, verflav,
                   versize, verbody,
                   RMIX_RPCX_STREAM_ACCEPTSTAT_PROG_UNAVAIL, 0, 0)) {
          /* Check for other errors than connection closed before transfer. */
          if (ECONNABORTED != errno) {
            RMIX_LOG((RMIX_WARN(unable to write program unavailable message)))
          }
        }
      }
      /* Enter executing state. */
      if (0 != rmix_rpcx_executing(callid)) {
        RMIX_LOG((RMIX_WARN(unable to enter executing state)))
      }
      /* Enter executed state. */
      if (0 != rmix_rpcx_executed(callid)) {
        RMIX_LOG((RMIX_WARN(unable to enter executed state)))
      }
      /* Deallocate authentication verifier. */
      if (NULL != verbody) {
        free(verbody);
      }
      errno = EPROTO;
      return NULL;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to lookup object at object registry)))
      /* Close the incoming XDR data stream. */
      if (0 != istream.close(istream.incoming)) {
        RMIX_LOG((RMIX_WARN(unable to close the incoming XDR data stream)))
      } else {
        /* Write system error message. */
        if (0 != rmix_rpcx_stream_write_error(&ostream, callid, verflav,
                   versize, verbody,
                   RMIX_RPCX_STREAM_ACCEPTSTAT_SYSTEM_ERR, 0, 0)) {
          /* Check for other errors than connection closed before transfer. */
          if (ECONNABORTED != errno) {
            RMIX_LOG((RMIX_WARN(unable to write system error message)))
          }
        }
      }
      /* Enter executing state. */
      if (0 != rmix_rpcx_executing(callid)) {
        RMIX_LOG((RMIX_WARN(unable to enter executing state)))
      }
      /* Enter executed state. */
      if (0 != rmix_rpcx_executed(callid)) {
        RMIX_LOG((RMIX_WARN(unable to enter executed state)))
      }
      /* Deallocate authentication verifier. */
      if (NULL != verbody) {
        free(verbody);
      }
      errno = errno2;
      return NULL;
    }
  }
  /* Verify program version. */
  if (0 != progver) {
    /* Close the incoming XDR data stream. */
    if (0 != istream.close(istream.incoming)) {
      RMIX_LOG((RMIX_WARN(unable to close the incoming XDR data stream)))
    } else {
      /* Write program version mismatch message. */
      if (0 != rmix_rpcx_stream_write_error(&ostream, callid, verflav,
                 versize, verbody,
                 RMIX_RPCX_STREAM_ACCEPTSTAT_PROG_MISMATCH, 0, 0)) {
        /* Check for other errors than connection closed before transfer. */
        if (ECONNABORTED != errno) {
          RMIX_LOG((
            RMIX_WARN(unable to write program version mismatch message)))
        }
      }
    }
    /* Enter executing state. */
    if (0 != rmix_rpcx_executing(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executing state)))
    }
    /* Enter executed state. */
    if (0 != rmix_rpcx_executed(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executed state)))
    }
    /* Deallocate authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    errno = EPROTO;
    return NULL;
  }
  /* Verify procedure number. */
  if (procnum >= interface->count) {
    /* Close the incoming XDR data stream. */
    if (0 != istream.close(istream.incoming)) {
      RMIX_LOG((RMIX_WARN(unable to close the incoming XDR data stream)))
    } else {
      /* Write procedure unavailable message. */
      if (0 != rmix_rpcx_stream_write_error(&ostream, callid, verflav,
                 versize, verbody,
                 RMIX_RPCX_STREAM_ACCEPTSTAT_PROC_UNAVAIL, 0, 0)) {
        /* Check for other errors than connection closed before transfer. */
        if (ECONNABORTED != errno) {
          RMIX_LOG((RMIX_WARN(unable to write procedure unavailable message)))
        }
      }
    }
    /* Enter executing state. */
    if (0 != rmix_rpcx_executing(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executing state)))
    }
    /* Enter executed state. */
    if (0 != rmix_rpcx_executed(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executed state)))
    }
    /* Deallocate authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    errno = EPROTO;
    return NULL;
  }
  /* Create and read call input values. */
  if (0 != rmix_rpcx_stream_read_call2(&istream, &ostream,
             interface->methods[procnum].input, &inary, &incnt,
             callid, verflav, versize, verbody)) {
    int errno2 = (ECONNABORTED == errno)?(errno = EPIPE):errno;
    RMIX_LOG((RMIX_WARN(unable to create and read call input values)))
    /* Close the incoming XDR data stream. */
    if (0 != istream.close(istream.incoming)) {
      RMIX_LOG((RMIX_WARN(unable to close the incoming XDR data stream)))
    } else {
      /* Write garbled args message. */
      if (0 != rmix_rpcx_stream_write_error(&ostream, callid, verflav,
                 versize, verbody,
                 RMIX_RPCX_STREAM_ACCEPTSTAT_GARBAGE_ARGS, 0, 0)) {
        /* Check for other errors than connection closed before transfer. */
        if (ECONNABORTED != errno) {
          RMIX_LOG((RMIX_WARN(unable to write garbled args message)))
        }
      }
    }
    /* Enter executing state. */
    if (0 != rmix_rpcx_executing(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executing state)))
    }
    /* Enter executed state. */
    if (0 != rmix_rpcx_executed(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executed state)))
    }
    /* Deallocate authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    errno = errno2;
    return NULL;
  }
  /* Push deallocate call input values to pthread cleanup stack. */
  icleanup.valary = inary;
  icleanup.valcnt = incnt;
  icleanup.sigstr = interface->methods[procnum].input;
  pthread_cleanup_push(rmix_valary_cleanup, &icleanup);
  /* Create call output values. */
  if (0 != rmix_valary_create(&outary, &outcnt,
                              interface->methods[procnum].output)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create call output values)))
    /* Close the incoming XDR data stream. */
    if (0 != istream.close(istream.incoming)) {
      RMIX_LOG((RMIX_WARN(unable to close the incoming XDR data stream)))
    } else {
      /* Write system error message. */
      if (0 != rmix_rpcx_stream_write_error(&ostream, callid, verflav,
                 versize, verbody,
                 RMIX_RPCX_STREAM_ACCEPTSTAT_SYSTEM_ERR, 0, 0)) {
        /* Check for other errors than connection closed before transfer. */
        if (ECONNABORTED != errno) {
          RMIX_LOG((RMIX_WARN(unable to write system error message)))
        }
      }
    }
    /* Enter executing state. */
    if (0 != rmix_rpcx_executing(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executing state)))
    }
    /* Enter executed state. */
    if (0 != rmix_rpcx_executed(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executed state)))
    }
    /* Destroy call input values. */
    if (0 != rmix_valary_destroy(&inary, &incnt,
                                 interface->methods[procnum].input)) {
      RMIX_LOG((RMIX_WARN(unable to destroy call input values)))
    }
    /* Deallocate authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    errno = errno2;
    return NULL;
  }
  /* Push deallocate call output values to pthread cleanup stack. */
  ocleanup.valary = outary;
  ocleanup.valcnt = outcnt;
  ocleanup.sigstr = interface->methods[procnum].output;
  pthread_cleanup_push(rmix_valary_cleanup, &ocleanup);
  /* Write reply message w/o procedure return. */
  if (0 != rmix_rpcx_stream_write_reply(&ostream, callid,
                                        verflav, versize, verbody)) {
    int errno2 = errno;
    /* Check for other errors than connection closed before transfer. */
    if (ECONNABORTED != errno) {
      RMIX_LOG((RMIX_WARN(unable to write reply message w/o procedure return)))
    }
    /* Enter executing state. */
    if (0 != rmix_rpcx_executing(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executing state)))
    }
    /* Enter executed state. */
    if (0 != rmix_rpcx_executed(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executed state)))
    }
    /* Destroy call output values. */
    if (0 != rmix_valary_destroy(&outary, &outcnt,
                                 interface->methods[procnum].output)) {
      RMIX_LOG((RMIX_WARN(unable to destroy call output values)))
    }
    /* Destroy call input values. */
    if (0 != rmix_valary_destroy(&inary, &incnt,
                                 interface->methods[procnum].input)) {
      RMIX_LOG((RMIX_WARN(unable to destroy call input values)))
    }
    /* Deallocate authentication verifier. */
    if (NULL != verbody) {
      free(verbody);
    }
    errno = errno2;
    return NULL;
  }
  /* Pop deallocate call output values from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate call input values from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Pop deallocate authentication verifier from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Deallocate authentication verifier. */
  if (NULL != verbody) {
    free(verbody);
  }
  /* Push deallocate call output values to pthread cleanup stack. */
  ocleanup.valary = outary;
  ocleanup.valcnt = outcnt;
  ocleanup.sigstr = interface->methods[procnum].output;
  pthread_cleanup_push(rmix_valary_cleanup, &ocleanup);
  /* Push deallocate call input values to pthread cleanup stack. */
  icleanup.valary = inary;
  icleanup.valcnt = incnt;
  icleanup.sigstr = interface->methods[procnum].input;
  pthread_cleanup_push(rmix_valary_cleanup, &icleanup);
  /* Enter executing state. */
  if (0 != rmix_rpcx_executing(callid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to enter executing state)))
    /* Destroy call output values. */
    if (0 != rmix_valary_destroy(&outary, &outcnt,
                                 interface->methods[procnum].output)) {
      RMIX_LOG((RMIX_WARN(unable to destroy call output values)))
    }
    /* Destroy call input values. */
    if (0 != rmix_valary_destroy(&inary, &incnt,
                                 interface->methods[procnum].input)) {
      RMIX_LOG((RMIX_WARN(unable to destroy call input values)))
    }
    errno = errno2;
    return NULL;
  }
  /* Call method. */
  if (0 != interface->methods[procnum].call(object, outary, outcnt,
                                            (const void**)inary, incnt)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to call method)))
    /* Enter executed state. */
    if (0 != rmix_rpcx_executed(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executed state)))
    }
    /* Destroy call output values. */
    if (0 != rmix_valary_destroy(&outary, &outcnt,
                                 interface->methods[procnum].output)) {
      RMIX_LOG((RMIX_WARN(unable to destroy call output values)))
    }
    /* Destroy call input values. */
    if (0 != rmix_valary_destroy(&inary, &incnt,
                                 interface->methods[procnum].input)) {
      RMIX_LOG((RMIX_WARN(unable to destroy call input values)))
    }
    errno = errno2;
    return NULL;
  }
  /* Pop deallocate call input values from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Write and destroy the procedure return. */
  if (0 != rmix_rpcx_stream_write_reply2(&ostream,
                                         interface->methods[procnum].output,
                                         &outary, &outcnt)) {
    int errno2 = errno;
    /* Check for other errors than connection closed before transfer. */
    if (ECONNABORTED != errno) {
      RMIX_LOG((RMIX_WARN(unable to write and destroy the procedure return)))
    }
    /* Enter executed state. */
    if (0 != rmix_rpcx_executed(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executed state)))
    }
    /* Destroy call output values. */
    if (0 != rmix_valary_destroy(&outary, &outcnt,
                                 interface->methods[procnum].output)) {
      RMIX_LOG((RMIX_WARN(unable to destroy call output values)))
    }
    /* Destroy call input values. */
    if (0 != rmix_valary_destroy(&inary, &incnt,
                                 interface->methods[procnum].input)) {
      RMIX_LOG((RMIX_WARN(unable to destroy call input values)))
    }
    errno = errno2;
    return NULL;
  }
  /* Pop deallocate call output values from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Destroy call input values. */
  if (0 != rmix_valary_destroy(&inary, &incnt,
                               interface->methods[procnum].input)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy call input values)))
    /* Enter executed state. */
    if (0 != rmix_rpcx_executed(callid)) {
      RMIX_LOG((RMIX_WARN(unable to enter executed state)))
    }
    errno = errno2;
    return NULL;
  }
  /* Enter executed state. */
  if (0 != rmix_rpcx_executed(callid)) {
    RMIX_LOG((RMIX_WARN(unable to enter executed state)))
    return NULL;
  }
  return rmix_rpcx_server;
}

/*
 * Moves an incoming call into the reading state.
 *
 * callid   = The call id.
 * objectid = The object id.
 * return = 0 on success, 1 on closed connection, or -1 on
 *          any other error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_reading"
int rmix_rpcx_reading (unsigned int callid,
                       unsigned int objectid) {
  void         *array;
  unsigned int  index, error, error2;
  /* Lock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Search for call id in reading array. */
  for (index = 0; index < rmix_rpcx_data.reading.count; index++) {
    /* Check for finalized rmix-rpcx plug-in. */
    if (0 == rmix_rpcx_data.ready) {
      errno = EACCES;
      RMIX_LOG((RMIX_WARN(rmix-rpcx plug-in finalized)))
      /* Unlock rmix-rpcx plug-in mutex. */
      if (0 != (error2 = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
        errno = error2;
        RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
        errno = error2;
        rmix_syserr();
      }
      errno = EACCES;
      return -1;
    }
    /* Check for for call id in reading array. */
    if (rmix_rpcx_data.reading.array[index].callid == callid) {
      /* Push unlock rmix-rpcx plug-in mutex to pthread cleanup stack. */
      pthread_cleanup_push(rmix_rpcx_unlock, &rmix_rpcx_data.mutex);
      /* Wait for reading ready condition. */
      if (0 != (error = pthread_cond_wait(&rmix_rpcx_data.reading.ready,
                                          &rmix_rpcx_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to wait for reading ready condition)))
        errno = error;
        rmix_syserr();
        /* Unlock rmix-rpcx plug-in mutex. */
        if (0 != (error2 = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
          errno = error2;
          RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
          errno = error2;
          rmix_syserr();
        }
        errno = error;
        return -1;
      }
      /* Pop unlock rmix-rpcx plug-in mutex from pthread cleanup stack. */
      pthread_cleanup_pop(0);
      /* Reset index. */
      index = 0;
      index--;
    }
  }
  /* Increase reading count. */
  index = rmix_rpcx_data.reading.count;
  rmix_rpcx_data.reading.count++;
  /* Reallocate reading array. */
  if (NULL == (array = realloc(rmix_rpcx_data.reading.array,
                               rmix_rpcx_data.reading.count *
                        sizeof(rmix_rpcx_data.reading.array[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate reading array)))
    /* Decrease reading array. */
    rmix_rpcx_data.reading.count--;
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    errno2 = errno;
    return -1;
  }
  rmix_rpcx_data.reading.array = array;
  /* Save reading array entry. */
  rmix_rpcx_data.reading.array[index].callid   = callid;
  rmix_rpcx_data.reading.array[index].objectid = objectid;
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Moves an incoming call into the executing state.
 *
 * callid = The call id.
 * return = 0 on success, 1 on closed connection, or -1 on
 *          any other error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_executing"
int rmix_rpcx_executing (unsigned int callid) {
  int           error, error2;
  void         *array;
  unsigned int  index, objectid;
  /* Lock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Search for call id in reading array. */
  for (index = 0; index < rmix_rpcx_data.reading.count; index++) {
    /* Check for for call id in reading array. */
    if (rmix_rpcx_data.reading.array[index].callid == callid) {
      /* Save object id. */
      objectid = rmix_rpcx_data.reading.array[index].objectid;
      break;
    }
  }
  /* Check if call id is not in reading array. */
  if (index == rmix_rpcx_data.reading.count) {
    errno = ESRCH;
    RMIX_LOG((RMIX_WARN(call id is not in reading array)))
    errno = ESRCH;
    rmix_syserr();
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Decrease reading count. */
  rmix_rpcx_data.reading.count--;
  memmove(rmix_rpcx_data.reading.array + index,
          rmix_rpcx_data.reading.array + index + 1,
         (rmix_rpcx_data.reading.count - index) *
   sizeof(rmix_rpcx_data.reading.array[0]));
  /* Reallocate reading array. */
  if (0 == rmix_rpcx_data.reading.count) {
    free(rmix_rpcx_data.reading.array);
    rmix_rpcx_data.reading.array = NULL;
  } else if (NULL == (array = realloc(rmix_rpcx_data.reading.array,
                                      rmix_rpcx_data.reading.count *
                               sizeof(rmix_rpcx_data.reading.array[0])))) {
    RMIX_LOG((RMIX_WARN(unable to reallocate reading array)))
    errno = 0;
  } else {
    rmix_rpcx_data.reading.array = array;
  }
  /* Broadcast reading ready condition. */
  if (0 != (error = pthread_cond_broadcast(&rmix_rpcx_data.reading.ready))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to broadcast reading ready condition)))
    errno = error;
    rmix_syserr();
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error2 = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error2;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error2;
      rmix_syserr();
    }
    errno = error;
    return -1;
  }
  /* Search for call id in executing array. */
  for (index = 0; index < rmix_rpcx_data.executing.count; index++) {
    /* Check for for call id in executing array. */
    if (rmix_rpcx_data.executing.array[index].callid == callid) {
      /* Push unlock rmix-rpcx plug-in mutex to pthread cleanup stack. */
      pthread_cleanup_push(rmix_rpcx_unlock, &rmix_rpcx_data.mutex);
      /* Wait for executing ready condition. */
      if (0 != (error = pthread_cond_wait(&rmix_rpcx_data.executing.ready,
                                          &rmix_rpcx_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to wait for executing ready condition)))
        errno = error;
        rmix_syserr();
        /* Unlock rmix-rpcx plug-in mutex. */
        if (0 != (error2 = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
          errno = error2;
          RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
          errno = error2;
          rmix_syserr();
        }
        errno = error;
        return -1;
      }
      /* Pop unlock rmix-rpcx plug-in mutex from pthread cleanup stack. */
      pthread_cleanup_pop(0);
      /* Reset index. */
      index = 0;
      index--;
    }
  }
  /* Increase reading count. */
  index = rmix_rpcx_data.executing.count;
  rmix_rpcx_data.executing.count++;
  /* Reallocate reading array. */
  if (NULL == (array = realloc(rmix_rpcx_data.executing.array,
                               rmix_rpcx_data.executing.count *
                        sizeof(rmix_rpcx_data.executing.array[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate executing array)))
    /* Decrease executing array. */
    rmix_rpcx_data.reading.count--;
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    errno2 = errno;
    return -1;
  }
  rmix_rpcx_data.executing.array = array;
  /* Save executing array entry. */
  rmix_rpcx_data.executing.array[index].callid   = callid;
  rmix_rpcx_data.executing.array[index].objectid = objectid;
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Moves an incoming call into the executed state.
 *
 * callid = The call id.
 * return = 0 on success, 1 on closed connection, or -1 on
 *          any other error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_executed"
int rmix_rpcx_executed (unsigned int callid) {
  int           error;
  void         *array;
  unsigned int  index;
  /* Lock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Search for call id in executing array. */
  for (index = 0; index < rmix_rpcx_data.executing.count; index++) {
    /* Check for for call id in executing array. */
    if (rmix_rpcx_data.executing.array[index].callid == callid) {
      break;
    }
  }
  /* Check if call id is not in executing array. */
  if (index == rmix_rpcx_data.executing.count) {
    errno = ESRCH;
    RMIX_LOG((RMIX_WARN(call id is not in executing array)))
    errno = ESRCH;
    rmix_syserr();
    /* Unlock rmix-rpcx plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Decrease executing count. */
  rmix_rpcx_data.executing.count--;
  memmove(rmix_rpcx_data.executing.array + index,
          rmix_rpcx_data.executing.array + index + 1,
         (rmix_rpcx_data.executing.count - index) *
   sizeof(rmix_rpcx_data.executing.array[0]));
  /* Reallocate executing array. */
  if (0 == rmix_rpcx_data.executing.count) {
    free(rmix_rpcx_data.executing.array);
    rmix_rpcx_data.executing.array = NULL;
  } else if (NULL == (array = realloc(rmix_rpcx_data.executing.array,
                                      rmix_rpcx_data.executing.count *
                               sizeof(rmix_rpcx_data.executing.array[0])))) {
    RMIX_LOG((RMIX_WARN(unable to reallocate executing array)))
    errno = 0;
  } else {
    rmix_rpcx_data.executing.array = array;
  }
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Broadcast executing ready condition. */
  if (0 != (error = pthread_cond_broadcast(&rmix_rpcx_data.executing.ready))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to broadcast executing ready condition)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * The rmix-rpcx plug-in mutex unlock cleanup.
 *
 * argument = NULL.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_rpcx_unlock"
void rmix_rpcx_unlock (void *argument) {
  int error;
  /* Unlock rmix-rpcx plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_rpcx_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmix-rpcx plug-in mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
  }
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
